module lib_array_sort
  use lib_log
  implicit none
  private
  !-------------------------------------------------------------
  ! Public procedures
  !-------------------------------------------------------------
  public :: sort
  public :: sorted
  public :: argsort
  public :: search
  public :: search_nearest
  public :: search_linear
  !-------------------------------------------------------------
  ! Interfaces
  !-------------------------------------------------------------
  interface sort
    module procedure sort_int1
    module procedure sort_int2
    module procedure sort_int4
    module procedure sort_int8
    module procedure sort_real
    module procedure sort_dble
    module procedure sort_arg_int1
    module procedure sort_arg_int2
    module procedure sort_arg_int4
    module procedure sort_arg_int8
    module procedure sort_arg_real
    module procedure sort_arg_dble
  end interface

  interface argsort
    module procedure argsort_int1
    module procedure argsort_int2
    module procedure argsort_int4
    module procedure argsort_int8
    module procedure argsort_real
    module procedure argsort_dble
  end interface

  interface sorted
    module procedure sorted_int1
    module procedure sorted_int2
    module procedure sorted_int4
    module procedure sorted_int8
    module procedure sorted_real
    module procedure sorted_dble
  end interface

  interface search
    module procedure search_binary_sorted_int4
    module procedure search_binary_sorted_int8
    module procedure search_binary_sorted_real
    module procedure search_binary_sorted_dble
    module procedure search_binary_arg_int8
  end interface

  interface search_nearest
    module procedure search_binary_nearest_sorted_int4
    module procedure search_binary_nearest_sorted_dble
    module procedure search_binary_nearest_arg_dble
  end interface

  interface search_linear
    module procedure search_linear_int4
    module procedure search_linear_int8
    module procedure search_linear_int4__lap__index_int4
    module procedure search_linear_int8__lap__index_int4
  end interface
  !-------------------------------------------------------------
contains
!==============================================================
!
!==============================================================
subroutine sort_arg_int1(array, arg)
  implicit none
  integer(1), intent(inout) :: array(:)
  integer(8), intent(in)    :: arg(:)
  integer(1), allocatable :: arr(:)
  integer(8) :: imax, i

  imax = size(arg,kind=8)
  allocate(arr(imax))

  do i = 1_8, imax
    arr(i) = array(arg(i))
  enddo

  array(:) = arr(:)

  deallocate(arr)
end subroutine sort_arg_int1
!==============================================================
!
!==============================================================
subroutine sort_arg_int2(array, arg)
  implicit none
  integer(2), intent(inout) :: array(:)
  integer(8), intent(in)    :: arg(:)
  integer(2), allocatable :: arr(:)
  integer(8) :: imax, i

  imax = size(arg,kind=8)
  allocate(arr(imax))

  do i = 1_8, imax
    arr(i) = array(arg(i))
  enddo

  array(:) = arr(:)

  deallocate(arr)
end subroutine sort_arg_int2
!==============================================================
!
!==============================================================
subroutine sort_arg_int4(array, arg)
  implicit none
  integer(4), intent(inout) :: array(:)
  integer(8), intent(in)    :: arg(:)
  integer(4), allocatable :: arr(:)
  integer(8) :: imax, i

  imax = size(arg,kind=8)
  allocate(arr(imax))

  do i = 1_8, imax
    arr(i) = array(arg(i))
  enddo

  array(:) = arr(:)

  deallocate(arr)
end subroutine sort_arg_int4
!==============================================================
!
!==============================================================
subroutine sort_arg_int8(array, arg)
  implicit none
  integer(8), intent(inout) :: array(:)
  integer(8), intent(in)    :: arg(:)
  integer(8), allocatable :: arr(:)
  integer(8) :: imax, i

  imax = size(arg,kind=8)
  allocate(arr(imax))

  do i = 1_8, imax
    arr(i) = array(arg(i))
  enddo

  array(:) = arr(:)

  deallocate(arr)
end subroutine sort_arg_int8
!==============================================================
!
!==============================================================
subroutine sort_arg_real(array, arg)
  implicit none
  real(4)   , intent(inout) :: array(:)
  integer(8), intent(in)    :: arg(:)
  real(4), allocatable :: arr(:)
  integer(8) :: imax, i

  imax = size(arg,kind=8)
  allocate(arr(imax))

  do i = 1_8, imax
    arr(i) = array(arg(i))
  enddo

  array(:) = arr(:)

  deallocate(arr)
end subroutine sort_arg_real
!==============================================================
!
!==============================================================
subroutine sort_arg_dble(array, arg)
  implicit none
  real(8)   , intent(inout) :: array(:)
  integer(8), intent(in)    :: arg(:)
  real(8), allocatable :: arr(:)
  integer(8) :: imax, i

  imax = size(arg,kind=8)
  allocate(arr(imax))

  do i = 1_8, imax
    arr(i) = array(arg(i))
  enddo

  array(:) = arr(:)

  deallocate(arr)
end subroutine sort_arg_dble
!==============================================================
!
!==============================================================
!
!
!
!
!
!==============================================================
!
!==============================================================
subroutine sort_int1(array)
  implicit none
  integer(1), intent(inout) :: array(:)
  integer(1) :: t
  integer(8) :: n
  integer(8) :: i, j, k, l

  n = size(array)

  if( n == 1 ) return

  l = n/2 + 1
  k = n
  do while (k /= 1)
    if (l > 1) then
      l = l-1
      t = array(l)
    else
      t = array(k)
      array(k) = array(1)
      k = k - 1
      if (k == 1) then
         array(1) = t
         exit
      endif
    endif
    i = l
    j = l + l
    do while (j <= k)
      if (j < k) then
        if (array(j) < array(j+1)) j = j + 1
      endif
      if (t < array(j)) then
        array(i) = array(j)
        i = j
        j = j + j
      else
        j = k + 1
      endif
    enddo
    array(i) = t
  enddo
end subroutine sort_int1
!==============================================================
!
!==============================================================
subroutine sort_int2(array)
  implicit none
  integer(2), intent(inout) :: array(:)
  integer(2) :: t
  integer(8) :: n
  integer(8) :: i, j, k, l

  n = size(array)

  if( n == 1 ) return

  l = n/2 + 1
  k = n
  do while (k /= 1)
    if (l > 1) then
      l = l-1
      t = array(l)
    else
      t = array(k)
      array(k) = array(1)
      k = k - 1
      if (k == 1) then
         array(1) = t
         exit
      endif
    endif
    i = l
    j = l + l
    do while (j <= k)
      if (j < k) then
        if (array(j) < array(j+1)) j = j + 1
      endif
      if (t < array(j)) then
        array(i) = array(j)
        i = j
        j = j + j
      else
        j = k + 1
      endif
    enddo
    array(i) = t
  enddo
end subroutine sort_int2
!==============================================================
!
!==============================================================
subroutine sort_int4(array)
  implicit none
  integer(4), intent(inout) :: array(:)
  integer(4) :: t
  integer(8) :: n
  integer(8) :: i, j, k, l

  n = size(array)

  if( n == 1 ) return

  l = n/2 + 1
  k = n
  do while (k /= 1)
    if (l > 1) then
      l = l-1
      t = array(l)
    else
      t = array(k)
      array(k) = array(1)
      k = k - 1
      if (k == 1) then
         array(1) = t
         exit
      endif
    endif
    i = l
    j = l + l
    do while (j <= k)
      if (j < k) then
        if (array(j) < array(j+1)) j = j + 1
      endif
      if (t < array(j)) then
        array(i) = array(j)
        i = j
        j = j + j
      else
        j = k + 1
      endif
    enddo
    array(i) = t
  enddo
end subroutine sort_int4
!==============================================================
!
!==============================================================
subroutine sort_int8(array)
  implicit none
  integer(8), intent(inout) :: array(:)
  integer(8) :: t
  integer(8) :: n
  integer(8) :: i, j, k, l

  n = size(array)

  if( n == 1 ) return

  l = n/2 + 1
  k = n
  do while (k /= 1)
    if (l > 1) then
      l = l-1
      t = array(l)
    else
      t = array(k)
      array(k) = array(1)
      k = k - 1
      if (k == 1) then
         array(1) = t
         exit
      endif
    endif
    i = l
    j = l + l
    do while (j <= k)
      if (j < k) then
        if (array(j) < array(j+1)) j = j + 1
      endif
      if (t < array(j)) then
        array(i) = array(j)
        i = j
        j = j + j
      else
        j = k + 1
      endif
    enddo
    array(i) = t
  enddo
end subroutine sort_int8
!==============================================================
!
!==============================================================
subroutine sort_real(array)
  implicit none
  real(4), intent(inout) :: array(:)
  real(4) :: t
  integer(8) :: n
  integer(8) :: i, j, k, l

  n = size(array)

  if( n == 1 ) return

  l = n/2 + 1
  k = n
  do while (k /= 1)
    if (l > 1) then
      l = l-1
      t = array(l)
    else
      t = array(k)
      array(k) = array(1)
      k = k - 1
      if (k == 1) then
         array(1) = t
         exit
      endif
    endif
    i = l
    j = l + l
    do while (j <= k)
      if (j < k) then
        if (array(j) < array(j+1)) j = j + 1
      endif
      if (t < array(j)) then
        array(i) = array(j)
        i = j
        j = j + j
      else
        j = k + 1
      endif
    enddo
    array(i) = t
  enddo
end subroutine sort_real
!==============================================================
!
!==============================================================
subroutine sort_dble(array)
  implicit none
  real(8), intent(inout) :: array(:)
  real(8) :: t
  integer(8) :: n
  integer(8) :: i, j, k, l

  n = size(array)

  if( n == 1 ) return

  l = n/2 + 1
  k = n
  do while (k /= 1)
    if (l > 1) then
      l = l-1
      t = array(l)
    else
      t = array(k)
      array(k) = array(1)
      k = k - 1
      if (k == 1) then
         array(1) = t
         exit
      endif
    endif
    i = l
    j = l + l
    do while (j <= k)
      if (j < k) then
        if (array(j) < array(j+1)) j = j + 1
      endif
      if (t < array(j)) then
        array(i) = array(j)
        i = j
        j = j + j
      else
        j = k + 1
      endif
    enddo
    array(i) = t
  enddo
end subroutine sort_dble
!==============================================================
!
!==============================================================
!
!
!
!
!
!==============================================================
!
!==============================================================
subroutine argsort_int1(array, arg)
  implicit none
  integer, parameter :: byte = 1
  integer(byte), intent(in)  :: array(:)
  integer(8)   , intent(out) :: arg(:)
  integer(byte), allocatable :: arr(:)
  integer(byte) :: t  ! arr
  integer(8)    :: u  ! arg
  integer(8) :: n
  integer(8) :: i, j, k, l

  n = size(array)

  do i = 1_8, n
    arg(i) = i
  enddo

  if( n == 1 ) return

  allocate(arr(n))
  arr = array

  l = n/2 + 1
  k = n
  do while (k /= 1)
    if (l > 1) then
      l = l-1
      t = arr(l)
      u = arg(l)
    else
      t = arr(k)
      u = arg(k)
      arr(k) = arr(1)
      arg(k) = arg(1)
      k = k - 1
      if (k == 1) then
         arr(1) = t
         arg(1) = u
         exit
      endif
    endif
    i = l
    j = l + l
    do while (j <= k)
      if (j < k) then
        if (arr(j) < arr(j+1)) j = j + 1
      endif
      if (t < arr(j)) then
        arr(i) = arr(j)
        arg(i) = arg(j)
        i = j
        j = j + j
      else
        j = k + 1
      endif
    enddo
    arr(i) = t
    arg(i) = u
  enddo

  deallocate(arr)
end subroutine argsort_int1
!==============================================================
!
!==============================================================
subroutine argsort_int2(array, arg)
  implicit none
  integer, parameter :: byte = 2
  integer(byte), intent(in)  :: array(:)
  integer(8)   , intent(out) :: arg(:)
  integer(byte), allocatable :: arr(:)
  integer(byte) :: t  ! arr
  integer(8)    :: u  ! arg
  integer(8) :: n
  integer(8) :: i, j, k, l

  n = size(array)

  do i = 1_8, n
    arg(i) = i
  enddo

  if( n == 1 ) return

  allocate(arr(n))
  arr = array

  l = n/2 + 1
  k = n
  do while (k /= 1)
    if (l > 1) then
      l = l-1
      t = arr(l)
      u = arg(l)
    else
      t = arr(k)
      u = arg(k)
      arr(k) = arr(1)
      arg(k) = arg(1)
      k = k - 1
      if (k == 1) then
         arr(1) = t
         arg(1) = u
         exit
      endif
    endif
    i = l
    j = l + l
    do while (j <= k)
      if (j < k) then
        if (arr(j) < arr(j+1)) j = j + 1
      endif
      if (t < arr(j)) then
        arr(i) = arr(j)
        arg(i) = arg(j)
        i = j
        j = j + j
      else
        j = k + 1
      endif
    enddo
    arr(i) = t
    arg(i) = u
  enddo

  deallocate(arr)
end subroutine argsort_int2
!==============================================================
!
!==============================================================
subroutine argsort_int4(array, arg)
  implicit none
  integer, parameter :: byte = 4
  integer(byte), intent(in)  :: array(:)
  integer(8)   , intent(out) :: arg(:)
  integer(byte), allocatable :: arr(:)
  integer(byte) :: t  ! arr
  integer(8)    :: u  ! arg
  integer(8) :: n
  integer(8) :: i, j, k, l

  n = size(array)

  do i = 1_8, n
    arg(i) = i
  enddo

  if( n == 1 ) return

  allocate(arr(n))
  arr = array

  l = n/2 + 1
  k = n
  do while (k /= 1)
    if (l > 1) then
      l = l-1
      t = arr(l)
      u = arg(l)
    else
      t = arr(k)
      u = arg(k)
      arr(k) = arr(1)
      arg(k) = arg(1)
      k = k - 1
      if (k == 1) then
         arr(1) = t
         arg(1) = u
         exit
      endif
    endif
    i = l
    j = l + l
    do while (j <= k)
      if (j < k) then
        if (arr(j) < arr(j+1)) j = j + 1
      endif
      if (t < arr(j)) then
        arr(i) = arr(j)
        arg(i) = arg(j)
        i = j
        j = j + j
      else
        j = k + 1
      endif
    enddo
    arr(i) = t
    arg(i) = u
  enddo

  deallocate(arr)
end subroutine argsort_int4
!==============================================================
!
!==============================================================
subroutine argsort_int8(array, arg)
  implicit none
  integer, parameter :: byte = 8
  integer(byte), intent(in)  :: array(:)
  integer(8)   , intent(out) :: arg(:)
  integer(byte), allocatable :: arr(:)
  integer(byte) :: t  ! arr
  integer(8)    :: u  ! arg
  integer(8) :: n
  integer(8) :: i, j, k, l

  n = size(array)

  do i = 1_8, n
    arg(i) = i
  enddo

  if( n == 1 ) return

  allocate(arr(n))
  arr = array

  l = n/2 + 1
  k = n
  do while (k /= 1)
    if (l > 1) then
      l = l-1
      t = arr(l)
      u = arg(l)
    else
      t = arr(k)
      u = arg(k)
      arr(k) = arr(1)
      arg(k) = arg(1)
      k = k - 1
      if (k == 1) then
         arr(1) = t
         arg(1) = u
         exit
      endif
    endif
    i = l
    j = l + l
    do while (j <= k)
      if (j < k) then
        if (arr(j) < arr(j+1)) j = j + 1
      endif
      if (t < arr(j)) then
        arr(i) = arr(j)
        arg(i) = arg(j)
        i = j
        j = j + j
      else
        j = k + 1
      endif
    enddo
    arr(i) = t
    arg(i) = u
  enddo

  deallocate(arr)
end subroutine argsort_int8
!==============================================================
!
!==============================================================
subroutine argsort_real(array, arg)
  implicit none
  integer, parameter :: byte = 4
  real(byte), intent(in)  :: array(:)
  integer(8), intent(out) :: arg(:)
  real(byte), allocatable :: arr(:)
  real(byte) :: t  ! arr
  integer(8) :: u  ! arg
  integer(8) :: n
  integer(8) :: i, j, k, l

  n = size(array)

  do i = 1_8, n
    arg(i) = i
  enddo

  if( n == 1 ) return

  allocate(arr(n))
  arr = array

  l = n/2 + 1
  k = n
  do while (k /= 1)
    if (l > 1) then
      l = l-1
      t = arr(l)
      u = arg(l)
    else
      t = arr(k)
      u = arg(k)
      arr(k) = arr(1)
      arg(k) = arg(1)
      k = k - 1
      if (k == 1) then
         arr(1) = t
         arg(1) = u
         exit
      endif
    endif
    i = l
    j = l + l
    do while (j <= k)
      if (j < k) then
        if (arr(j) < arr(j+1)) j = j + 1
      endif
      if (t < arr(j)) then
        arr(i) = arr(j)
        arg(i) = arg(j)
        i = j
        j = j + j
      else
        j = k + 1
      endif
    enddo
    arr(i) = t
    arg(i) = u
  enddo

  deallocate(arr)
end subroutine argsort_real
!==============================================================
!
!==============================================================
subroutine argsort_dble(array, arg)
  implicit none
  integer, parameter :: byte = 8
  real(byte), intent(in)  :: array(:)
  integer(8), intent(out) :: arg(:)
  real(byte), allocatable :: arr(:)
  real(byte) :: t  ! arr
  integer(8) :: u  ! arg
  integer(8) :: n
  integer(8) :: i, j, k, l

  n = size(array)

  do i = 1_8, n
    arg(i) = i
  enddo

  if( n == 1 ) return

  allocate(arr(n))
  arr = array

  l = n/2 + 1
  k = n
  do while (k /= 1)
    if (l > 1) then
      l = l-1
      t = arr(l)
      u = arg(l)
    else
      t = arr(k)
      u = arg(k)
      arr(k) = arr(1)
      arg(k) = arg(1)
      k = k - 1
      if (k == 1) then
         arr(1) = t
         arg(1) = u
         exit
      endif
    endif
    i = l
    j = l + l
    do while (j <= k)
      if (j < k) then
        if (arr(j) < arr(j+1)) j = j + 1
      endif
      if (t < arr(j)) then
        arr(i) = arr(j)
        arg(i) = arg(j)
        i = j
        j = j + j
      else
        j = k + 1
      endif
    enddo
    arr(i) = t
    arg(i) = u
  enddo

  deallocate(arr)
end subroutine argsort_dble
!==============================================================
!
!==============================================================
!
!
!
!
!
!==============================================================
!
!==============================================================
function sorted_int1(array) result(array_sorted)
  implicit none
  integer(1), intent(in) :: array(:)
  integer(1)             :: array_sorted(size(array))

  array_sorted(:) = array(:)

  call sort(array_sorted)
end function sorted_int1
!==============================================================
!
!==============================================================
function sorted_int2(array) result(array_sorted)
  implicit none
  integer(2), intent(in) :: array(:)
  integer(2)             :: array_sorted(size(array))

  array_sorted(:) = array(:)

  call sort(array_sorted)
end function sorted_int2
!==============================================================
!
!==============================================================
function sorted_int4(array) result(array_sorted)
  implicit none
  integer(4), intent(in) :: array(:)
  integer(4)             :: array_sorted(size(array))

  array_sorted(:) = array(:)

  call sort(array_sorted)
end function sorted_int4
!==============================================================
!
!==============================================================
function sorted_int8(array) result(array_sorted)
  implicit none
  integer(8), intent(in) :: array(:)
  integer(8)             :: array_sorted(size(array))

  array_sorted(:) = array(:)

  call sort(array_sorted)
end function sorted_int8
!==============================================================
!
!==============================================================
function sorted_real(array) result(array_sorted)
  implicit none
  real(4)   , intent(in) :: array(:)
  real(4)                :: array_sorted(size(array))

  array_sorted(:) = array(:)

  call sort(array_sorted)
end function sorted_real
!==============================================================
!
!==============================================================
function sorted_dble(array) result(array_sorted)
  implicit none
  real(8), intent(in) :: array(:)
  real(8)             :: array_sorted(size(array))

  array_sorted(:) = array(:)

  call sort(array_sorted)
end function sorted_dble
!==============================================================
!
!==============================================================
!
!
!
!
!
!==============================================================
! Binary search
!==============================================================
subroutine search_binary_sorted_int4(x, arr, loc)
  implicit none
  integer(4), intent(in)  :: x
  integer(4), intent(in)  :: arr(:)
  integer(8), intent(out) :: loc
  integer(8) :: ijmax, left, right

  ijmax = size(arr)

  left = 1_8
  right = ijmax
  do while( left <= right )
    loc = (left + right) / 2_8
    if( arr(loc) == x )then
      return
    elseif( arr(loc) < x )then
      left = loc + 1
    else
      right = loc - 1
    endif
  enddo

  loc = 0_8
end subroutine search_binary_sorted_int4
!==============================================================
!
!==============================================================
subroutine search_binary_sorted_int8(x, arr, loc)
  implicit none
  integer(8), intent(in)  :: x
  integer(8), intent(in)  :: arr(:)
  integer(8), intent(out) :: loc
  integer(8) :: ijmax, left, right

  ijmax = size(arr)

  left = 1_8
  right = ijmax
  do while( left <= right )
    loc = (left + right) / 2_8
    if( arr(loc) == x )then
      return
    elseif( arr(loc) < x )then
      left = loc + 1
    else
      right = loc - 1
    endif
  enddo

  loc = 0_8
end subroutine search_binary_sorted_int8
!==============================================================
!
!==============================================================
subroutine search_binary_sorted_real(x, arr, loc)
  implicit none
  real(4)   , intent(in)  :: x
  real(4)   , intent(in)  :: arr(:)
  integer(8), intent(out) :: loc
  integer(8) :: ijmax, left, right

  ijmax = size(arr)

  left = 1_8
  right = ijmax
  do while( left <= right )
    loc = (left + right) / 2_8
    if( arr(loc) == x )then
      return
    elseif( arr(loc) < x )then
      left = loc + 1
    else
      right = loc - 1
    endif
  enddo

  loc = 0_8
end subroutine search_binary_sorted_real
!==============================================================
! Binary search
!==============================================================
subroutine search_binary_sorted_dble(x, arr, loc)
  implicit none
  real(8)   , intent(in)  :: x
  real(8)   , intent(in)  :: arr(:)
  integer(8), intent(out) :: loc
  integer(8) :: ijmax, left, right

  ijmax = size(arr)

  left = 1_8
  right = ijmax
  do while( left <= right )
    loc = (left + right) / 2_8
    if( arr(loc) == x )then
      return
    elseif( arr(loc) < x )then
      left = loc + 1
    else
      right = loc - 1
    endif
  enddo

  loc = 0_8
end subroutine search_binary_sorted_dble
!==============================================================
! Binary search
!==============================================================
subroutine search_binary_arg_int8(x, arr, arg, loc)
  implicit none
  integer(8), intent(in)  :: x
  integer(8), intent(in)  :: arr(:)
  integer(8), intent(in)  :: arg(:)
  integer(8), intent(out) :: loc
  integer(8) :: ijmax, left, right

  ijmax = size(arg)

  left = 1_8
  right = ijmax
  do while( left <= right )
    loc = (left + right) / 2_8
    if( arr(arg(loc)) == x )then
      return
    elseif( arr(arg(loc)) < x )then
      left = loc + 1
    else
      right = loc - 1
    endif
  enddo

  loc = 0_8
end subroutine search_binary_arg_int8
!==============================================================
!
!==============================================================
!
!
!
!
!
!==============================================================
! Usually arr(loc1) <= x <= arr(loc2)
! When arr(ijmax) < x, loc1=ijmax, loc2=ijmax+1
! When x < arr(1), loc1=0, loc2=1
!==============================================================
subroutine search_binary_nearest_sorted_int4(x, arr, loc1, loc2)
  implicit none
  integer(4), intent(in) :: x
  integer(4), intent(in) :: arr(:)
  integer(8), intent(out) :: loc1, loc2

  integer(8) :: ijmax, left, right
  integer(8) :: loc

  ijmax = size(arr)

  left = 1_8
  right = ijmax
  do while( left <= right )
    loc = (left + right) / 2_8
    if( arr(loc) == x )then
      loc1 = loc
      do while( loc1 > 1 )
        if( arr(loc1-1) < x ) exit
        loc1 = loc1 - 1
      enddo
      loc2 = loc
      do while( loc2 < ijmax )
        if( arr(loc2+1) > x ) exit
        loc2 = loc2 + 1
      enddo
      return
    elseif( arr(loc) < x )then
      if( loc == ijmax )then
        loc1 = loc
        loc2 = loc + 1
        return
      elseif( arr(loc+1) > x )then
        loc1 = loc
        loc2 = loc + 1
        return
      endif
      left = loc + 1
    else
      if( loc == 1 )then
        loc1 = loc - 1
        loc2 = loc
        return
      elseif( arr(loc-1) < x )then
        loc1 = loc - 1
        loc2 = loc
        return
      endif
      right = loc - 1
    endif
  enddo
end subroutine search_binary_nearest_sorted_int4
!==============================================================
!
!==============================================================
subroutine search_binary_nearest_sorted_dble(x, arr, loc1, loc2)
  implicit none
  real(8), intent(in) :: x
  real(8), intent(in) :: arr(:)
  integer(8), intent(out) :: loc1, loc2

  integer(8) :: ijmax, left, right
  integer(8) :: loc

  ijmax = size(arr)

  left = 1_8
  right = ijmax
  do while( left <= right )
    loc = (left + right) / 2_8
    if( arr(loc) == x )then
      loc1 = loc
      do while( loc1 > 1 )
        if( arr(loc1-1) < x ) exit
        loc1 = loc1 - 1
      enddo
      loc2 = loc
      do while( loc2 < ijmax )
        if( arr(loc2+1) > x ) exit
        loc2 = loc2 + 1
      enddo
      return
    elseif( arr(loc) < x )then
      if( loc == ijmax )then
        loc1 = loc
        loc2 = loc + 1
        return
      elseif( arr(loc+1) > x )then
        loc1 = loc
        loc2 = loc + 1
        return
      endif
      left = loc + 1
    else
      if( loc == 1 )then
        loc1 = loc - 1
        loc2 = loc
        return
      elseif( arr(loc-1) < x )then
        loc1 = loc - 1
        loc2 = loc
        return
      endif
      right = loc - 1
    endif
  enddo
end subroutine search_binary_nearest_sorted_dble
!==============================================================
!
!==============================================================
subroutine search_binary_nearest_arg_dble(x, arr, arg, loc1, loc2)
  implicit none
  real(8), intent(in) :: x
  real(8), intent(in) :: arr(:)
  integer(8), intent(in) :: arg(:)
  integer(8), intent(out) :: loc1, loc2

  integer(8) :: ijmax, left, right
  integer(8) :: loc

  ijmax = size(arr)

  left = 1_8
  right = ijmax
  do while( left <= right )
    loc = (left + right) / 2_8
    if( arr(arg(loc)) == x )then
      loc1 = loc
      do while( loc1 > 1 )
        if( arr(arg(loc1-1)) < x ) exit
        loc1 = loc1 - 1
      enddo
      loc2 = loc
      do while( loc2 < ijmax )
        if( arr(arg(loc2+1)) > x ) exit
        loc2 = loc2 + 1
      enddo
      return
    elseif( arr(arg(loc)) < x )then
      if( loc == ijmax )then
        loc1 = loc
        loc2 = loc + 1
        return
      elseif( arr(arg(loc+1)) > x )then
        loc1 = loc
        loc2 = loc + 1
        return
      endif
      left = loc + 1
    else
      if( loc == 1 )then
        loc1 = loc - 1
        loc2 = loc
        return
      elseif( arr(arg(loc-1)) < x )then
        loc1 = loc - 1
        loc2 = loc
        return
      endif
      right = loc - 1
    endif
  enddo
end subroutine search_binary_nearest_arg_dble
!==============================================================
!
!==============================================================
!
!
!
!
!
!==============================================================
!
!==============================================================
subroutine search_linear_int4(x, arr, ij)
  implicit none
  integer(4), intent(in) :: x
  integer(4), intent(in) :: arr(:)
  integer(8), intent(out) :: ij

  ij = 1_8
  do while( ij <= size(arr) )
    if( arr(ij) == x ) return
    ij = ij + 1_8
  enddo

  ! ERROR
  ij = 0_8
end subroutine search_linear_int4
!==============================================================
!
!==============================================================
subroutine search_linear_int8(x, arr, ij)
  implicit none
  integer(8), intent(in) :: x
  integer(8), intent(in) :: arr(:)
  integer(8), intent(out) :: ij

  ij = 1_8
  do while( ij <= size(arr) )
    if( arr(ij) == x ) return
    ij = ij + 1_8
  enddo

  ! ERROR
  ij = 0_8
end subroutine search_linear_int8
!==============================================================
!
!==============================================================
subroutine search_linear_int4__lap__index_int4(x, arr, ij)
  implicit none
  integer(4), intent(in) :: x
  integer(4), intent(in) :: arr(:)
  integer(4), intent(out) :: ij

  integer(8) :: ij8

  call search_linear_int4(x, arr, ij8)
  ij = int(ij8,4)
end subroutine search_linear_int4__lap__index_int4
!==============================================================
!
!==============================================================
subroutine search_linear_int8__lap__index_int4(x, arr, ij)
  implicit none
  integer(8), intent(in) :: x
  integer(8), intent(in) :: arr(:)
  integer(4), intent(out) :: ij

  integer(8) :: ij8

  call search_linear_int8(x, arr, ij8)
  ij = int(ij8,4)
end subroutine search_linear_int8__lap__index_int4
!==============================================================
!
!==============================================================
end module lib_array_sort
