module lib_math_base
  use lib_const
  use lib_base
  use lib_log
  implicit none
  private
  !-------------------------------------------------------------
  ! Public procedures
  !-------------------------------------------------------------
  public :: inc

  public :: swap

  public :: gcd

  public :: conv_unit

  public :: calc_inv_matrix_3x3

  public :: calc_cross_product
  !-------------------------------------------------------------
  ! Interfaces
  !-------------------------------------------------------------
  interface inc
    module procedure inc_int1_default
    module procedure inc_int2_default
    module procedure inc_int4_default
    module procedure inc_int8_default
    module procedure inc_real_default
    module procedure inc_dble_default
    module procedure inc_int1_nonDefault
    module procedure inc_int2_nonDefault
    module procedure inc_int4_nonDefault
    module procedure inc_int8_nonDefault
    module procedure inc_real_nonDefault
    module procedure inc_dble_nonDefault
  end interface

  interface swap
    module procedure swap_int4
  end interface

  interface gcd
    module procedure gcd_int4
    module procedure gcd_int8
  end interface

  interface conv_unit
    module procedure conv_unit_self_dble_1d
    module procedure conv_unit_self_dble_2d
  end interface

  interface calc_cross_product
    module procedure calc_cross_product_array
    module procedure calc_cross_product_components
  end interface
  !-------------------------------------------------------------
contains
!===============================================================
!
!===============================================================
subroutine inc_int1_default(a)
  implicit none
  integer(1), intent(inout) :: a

  a = a + 1_1
end subroutine inc_int1_default
!===============================================================
!
!===============================================================
subroutine inc_int2_default(a)
  implicit none
  integer(2), intent(inout) :: a

  a = a + 1_2
end subroutine inc_int2_default
!===============================================================
!
!===============================================================
subroutine inc_int4_default(a)
  implicit none
  integer(4), intent(inout) :: a

  a = a + 1_4
end subroutine inc_int4_default
!===============================================================
!
!===============================================================
subroutine inc_int8_default(a)
  implicit none
  integer(8), intent(inout) :: a

  a = a + 1_8
end subroutine inc_int8_default
!===============================================================
!
!===============================================================
subroutine inc_real_default(a)
  implicit none
  real(4), intent(inout) :: a

  a = a + 1.0
end subroutine inc_real_default
!===============================================================
!
!===============================================================
subroutine inc_dble_default(a)
  implicit none
  real(8), intent(inout) :: a

  a = a + 1.d0
end subroutine inc_dble_default
!===============================================================
!
!===============================================================
subroutine inc_int1_nonDefault(a, b)
  implicit none
  integer(1), intent(inout) :: a
  integer(1), intent(in)    :: b

  a = a + b
end subroutine inc_int1_nonDefault
!===============================================================
!
!===============================================================
subroutine inc_int2_nonDefault(a, b)
  implicit none
  integer(2), intent(inout) :: a
  integer(2), intent(in)    :: b

  a = a + b
end subroutine inc_int2_nonDefault
!===============================================================
!
!===============================================================
subroutine inc_int4_nonDefault(a, b)
  implicit none
  integer(4), intent(inout) :: a
  integer(4), intent(in)    :: b

  a = a + b
end subroutine inc_int4_nonDefault
!===============================================================
!
!===============================================================
subroutine inc_int8_nonDefault(a, b)
  implicit none
  integer(8), intent(inout) :: a
  integer(8), intent(in)    :: b

  a = a + b
end subroutine inc_int8_nonDefault
!===============================================================
!
!===============================================================
subroutine inc_real_nonDefault(a, b)
  implicit none
  real(4), intent(inout) :: a
  real(4), intent(in)    :: b

  a = a + b
end subroutine inc_real_nonDefault
!===============================================================
!
!===============================================================
subroutine inc_dble_nonDefault(a, b)
  implicit none
  real(8), intent(inout) :: a
  real(8), intent(in)    :: b

  a = a + b
end subroutine inc_dble_nonDefault
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine swap_int4(a, b)
  implicit none
  integer(4), intent(inout) :: a, b
  integer(4) :: c

  c = a
  a = b
  b = c
end subroutine swap_int4
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
integer(4) function gcd_int4(x, y) result(i)
  implicit none
  integer(4), intent(in) :: x, y

  i = min(x,y)
  do while( mod(x,i) /= 0_4 .and. mod(y,i) /= 0_4 )
    i = i - 1_4
  enddo
end function gcd_int4
!===============================================================
!
!===============================================================
integer(8) function gcd_int8(x, y) result(i)
  implicit none
  integer(8), intent(in) :: x, y

  i = min(x,y)
  do while( mod(x,i) /= 0_8 .and. mod(y,i) /= 0_8 )
    i = i - 1_8
  enddo
end function gcd_int8
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine conv_unit_self_dble_1d(d, uin, uout)
  implicit none
  real(8), intent(inout)  :: d(:)
  character(*), intent(in) :: uin
  character(*), intent(in) :: uout

  call echo(code%bgn, 'conv_unit__MP__conv_unit_self_dble', '-p -x2')
  !-------------------------------------------------------------
  if( uin == uout )then
    call echo(code%ret)
    return
  endif

  selectcase( trim(uin)//'_'//trim(uout) )
  case( trim(unit_square_meter)//'_'//trim(unit_square_kilometer) )
    d(:) = d(:) * 1d-6
  case( trim(unit_square_kilometer)//'_'//trim(unit_square_meter) )
    d(:) = d(:) * 1d6
  case default
    call eerr(str(msg_unexpected_condition())//&
            '\n  unit of input : '//str(uin)//&
            '\n  unit of output: '//str(uout))
  endselect
  !-------------------------------------------------------------
  call echo(code%ret)
end subroutine conv_unit_self_dble_1d
!===============================================================
!
!===============================================================
subroutine conv_unit_self_dble_2d(d, uin, uout)
  implicit none
  real(8), intent(inout)  :: d(:,:)
  character(*), intent(in) :: uin
  character(*), intent(in) :: uout

  call echo(code%bgn, 'conv_unit__MP__conv_unit_self_dble', '-p -x2')
  !-------------------------------------------------------------
  if( uin == uout )then
    call echo(code%ret)
    return
  endif

  selectcase( trim(uin)//'_'//trim(uout) )
  case( trim(unit_square_meter)//'_'//trim(unit_square_kilometer) )
    d(:,:) = d(:,:) * 1d-6
  case( trim(unit_square_kilometer)//'_'//trim(unit_square_meter) )
    d(:,:) = d(:,:) * 1d6
  case default
    call eerr(str(msg_unexpected_condition())//&
            '\n  unit of input : '//str(uin)//&
            '\n  unit of output: '//str(uout))
  endselect
  !-------------------------------------------------------------
  call echo(code%ret)
end subroutine conv_unit_self_dble_2d
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine calc_inv_matrix_3x3(a, b, coef)
  implicit none
  real(8), intent(in)  :: a(3,3)
  real(8), intent(out) :: b(3,3)
  real(8), intent(out) :: coef

  coef =  a(1,1)*a(2,2)*a(3,3) + a(2,1)*a(3,2)*a(1,3) + a(3,1)*a(1,2)*a(2,3) &
          - (a(3,1)*a(2,2)*a(1,3) + a(2,1)*a(1,2)*a(3,3) + a(1,1)*a(3,2)*a(2,3))
  b(1,1) =   a(2,2)*a(3,3) - a(3,2)*a(2,3)
  b(1,2) = - a(1,2)*a(3,3) + a(3,2)*a(1,3)
  b(1,3) =   a(1,2)*a(2,3) - a(2,2)*a(1,3)
  b(2,1) = - a(2,1)*a(3,3) + a(3,1)*a(2,3)
  b(2,2) =   a(1,1)*a(3,3) - a(3,1)*a(1,3)
  b(2,3) = - a(1,1)*a(2,3) + a(2,1)*a(1,3)
  b(3,1) =   a(2,1)*a(3,2) - a(3,1)*a(2,2)
  b(3,2) = - a(1,1)*a(3,2) + a(3,1)*a(1,2)
  b(3,3) =   a(1,1)*a(2,2) - a(2,1)*a(1,2)
end subroutine calc_inv_matrix_3x3
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine calc_cross_product_array(a, b, c)
  implicit none
  real(8), intent(in)  :: a(:), b(:)
  real(8), intent(out) :: c(:)

  c(1) = a(2)*b(3) - a(3)*b(2)
  c(2) = a(3)*b(1) - a(1)*b(3)
  c(3) = a(1)*b(2) - a(2)*b(1)
end subroutine calc_cross_product_array
!===============================================================
!
!===============================================================
subroutine calc_cross_product_components(a1, a2, a3, b1, b2, b3, c1, c2, c3)
  implicit none
  real(8), intent(in)  :: a1, a2, a3, b1, b2, b3
  real(8), intent(out) :: c1, c2, c3

  c1 = a2*b3 - a3*b2
  c2 = a3*b1 - a1*b3
  c3 = a1*b2 - a2*b1
end subroutine calc_cross_product_components
!===============================================================
!
!===============================================================
end module lib_math_base
