module lib_time
  use lib_const, only: &
    SEC_DAY
  implicit none
  private
!--------------------------------------------------------------
  public :: days
  public :: dayOfYear
  public :: second
  public :: timediff
  public :: date_and_time_values
!--------------------------------------------------------------
  interface days
    module procedure days_year
    module procedure days_month
  end interface

  interface second
    module procedure second_baseDefault
    module procedure second_baseSpecified
  end interface
!--------------------------------------------------------------
  integer, parameter :: year_base = 1900
!--------------------------------------------------------------
contains
!==============================================================
!
!==============================================================
integer function days_year(yr) result(days)
  implicit none
  integer, intent(in) :: yr

  days = 365
  if( mod(yr,4) == 0 )then
    days = 366
    if( mod(yr,100) == 0 )then
      days = 365
      if( mod(yr,400) == 0 )then
        days = 366
      endif
    endif
  endif
end function days_year
!==============================================================
!
!==============================================================
integer function days_month(yr, mn) result(days)
  implicit none
  integer, intent(in) :: yr, mn

  selectcase( mn )
  case( 2 )
    days = 28 + (days_year(yr) - 365)
  case( 4, 6, 9, 11 )
    days = 30
  case( 1, 3, 5, 7, 8, 10, 12 )
    days = 31
  case default
    days = 0
  endselect
end function days_month
!==============================================================
!
!==============================================================
integer function dayOfYear(yr, mn, dy)
  implicit none
  integer, intent(in) :: yr, mn, dy
  integer :: imn

  dayOfYear = dy
  do imn = 1, mn-1
    dayOfYear = dayOfYear + days(yr, imn)
  enddo
end function dayOfYear
!==============================================================
!
!==============================================================
real(8) function second_baseDefault(time) result(t)
  implicit none
  integer, intent(in) :: time(:)  !(yr,mn,dy,hr,mm,sc,ms)

  t = second_baseSpecified(time, YEAR_BASE)
end function second_baseDefault
!==============================================================
!
!==============================================================
real(8) function second_baseSpecified(time, yr0) result(t)
  implicit none
  integer, intent(in) :: time(:)  !(yr,mn,dy,timediff_from_UTC,hr,mm,sc,ms)
  integer, intent(in) :: yr0
  integer :: time_(8)
  integer :: d, yr

  time_(:) = (/0, 1, 1, 0, 0, 0, 0, 0/)  ! 0000/01/01 00:00:00.000 UTC
  time_(:size(time)) = time(:)

  d = 0
  do yr = yr0, time_(1)-1
    d = d + days(yr)
  enddo

  t = SEC_DAY * (d + dayOfYear(time_(1),time_(2),time_(3))) &
        + 60 * ((60 * time_(5)) + time_(6)) + time_(7) + time_(8)*1d-3
end function second_baseSpecified
!==============================================================
!
!==============================================================
real(8) function timediff(t1, t0)
  implicit none
  integer, intent(in) :: t1(:), t0(:)  !(yr,mn,dy,timediff_from_UTC,hr,mm,sc,ms)
  integer :: yr0

  yr0 = min(t1(1), t0(1))
  timediff = second(t1,yr0) - second(t0,yr0)
end function timediff
!==============================================================
!
!==============================================================
function date_and_time_values() result(values)
  implicit none
  character(8)  :: date
  character(10) :: time
  character(5)  :: zone
  integer       :: values(8)

  call date_and_time(date, time, zone, values)
end function date_and_time_values
!==============================================================
!
!==============================================================
end module lib_time
