program main
  use lib_const
  use lib_base
  use lib_log
  use lib_util
  use lib_io
  implicit none
  character(:), allocatable :: res
  type(file_) :: f
  integer(8)  :: nij
  integer :: iarg_min, iarg_fmin, iarg_cmin, &
             iarg_max, iarg_fmax, iarg_cmax
  character(clen_wfmt) :: wfmt_min, wfmt_fmin, wfmt_cmin, &
                          wfmt_max, wfmt_fmax, wfmt_cmax
  real(8)     , allocatable :: dble_1d(:)
  integer(8)  , allocatable :: int8_1d(:)
  character(clen_var) :: str_miss
  logical    :: is_miss_specified
  real(8)    :: miss_dble
  integer(8) :: miss_int8
  integer :: iarg
  character(clen_line) :: c

  character(clen_wfmt), parameter :: wfmt_int8_default = '(i0)'
  character(clen_wfmt), parameter :: wfmt_dble_default = '(es10.3)'

  call echo(code%set, '-ar')
  call echo(code%bgn, 'program get_stats', '-p')
  !-------------------------------------------------------------
  ! Read arguments
  !-------------------------------------------------------------
  call echo(code%ent, 'Reading arguments')

  call add_arg_char('file')
  call add_arg_int8('--nij'   , default=0_8)
  call add_arg_char('--dtype' , default=dtype_dble)
  call add_arg_int4('--rec'   , default=1)
  call add_arg_char('--endian', default=endian_little)
  call add_arg_char('--miss', default='')
  call add_arg_char('--min' , omittable=.true., default='')
  call add_arg_char('--fmin', omittable=.true., default='', metavar='floor_min')
  call add_arg_char('--cmin', omittable=.true., default='', metavar='ceil_min')
  call add_arg_char('--max' , omittable=.true., default='')
  call add_arg_char('--fmax', omittable=.true., default='', metavar='floor_max')
  call add_arg_char('--cmax', omittable=.true., default='', metavar='ceil_max')

  call read_arguments()

  !call print_arguments()

  call get_argument('file', f%path)
  call get_argument('--nij', nij)
  call get_argument('--dtype', f%dtype)
  call get_argument('--rec', f%rec)
  call get_argument('--endian', f%endian)
  call get_argument('--miss', str_miss)
  call get_argument('--min' , wfmt_min , iarg_min )
  call get_argument('--fmin', wfmt_fmin, iarg_fmin)
  call get_argument('--cmin', wfmt_cmin, iarg_cmin)
  call get_argument('--max' , wfmt_max , iarg_max )
  call get_argument('--fmax', wfmt_fmax, iarg_fmax)
  call get_argument('--cmax', wfmt_cmax, iarg_cmax)

  call free_arguments_list()
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  call check_permission(f%path, action_read)

  if( nij == 0_8 )then
    if( f%rec > 1 )then
      call eerr(str(msg_unexpected_condition())//&
              '\n  "nij" may not be omitted when "rec" is greater than 1')
    endif

    if( mod(filesize(f%path),byte_dtype(f%dtype)) /= 0 )then
      call eerr(str(msg_unexpected_condition())//&
              '\n  File size is indivisible by '//&
                'the byte size of data type "'//str(f%dtype)//'"')
    endif
    nij = filesize(f%path) / byte_dtype(f%dtype)
  endif

  is_miss_specified = .false.
  if( str_miss /= '' )then
    is_miss_specified = .true.
    selectcase( f%dtype )
    case( dtype_dble )
      miss_dble = dble_char(str_miss)
    case( dtype_int8 )
      miss_int8 = int8_char(str_miss)
    case default
      call eerr(str(msg_invalid_value())//&
              '\n  f%dtype: '//str(f%dtype))
    endselect
  endif

  selectcase( f%dtype )
  case( dtype_int1, &
        dtype_int2, &
        dtype_int4, &
        dtype_int8 )
    if( iarg_fmin > 0 .or. iarg_cmin > 0 .or. &
        iarg_fmax > 0 .or. iarg_cmax > 0 )then
      call eerr(str(msg_unexpected_condition())//&
              '\n  --fmin, --cmin, --fmax or --cmax may not be used for integer')
    endif
  case( dtype_real, &
        dtype_dble )
    continue
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  f%dtype: '//str(f%dtype))
  endselect

  call echo(code%ext)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  selectcase( f%dtype )
  case( dtype_int1, &
        dtype_int2, &
        dtype_int4, &
        dtype_int8 )
    if( wfmt_min == '' ) wfmt_min = wfmt_int8_default
    if( wfmt_max == '' ) wfmt_max = wfmt_int8_default
  case( dtype_real, &
        dtype_dble )
    if( wfmt_min  == '' ) wfmt_min  = wfmt_dble_default
    if( wfmt_fmin == '' ) wfmt_fmin = wfmt_int8_default
    if( wfmt_cmin == '' ) wfmt_cmin = wfmt_int8_default
    if( wfmt_max  == '' ) wfmt_max  = wfmt_dble_default
    if( wfmt_fmax == '' ) wfmt_fmax = wfmt_int8_default
    if( wfmt_cmax == '' ) wfmt_cmax = wfmt_int8_default
  endselect
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  call echo(code%ent, 'Making output')

  allocate(character(0) :: res)

  selectcase( f%dtype )
  !-------------------------------------------------------------
  ! Case: Integer
  case( dtype_int1, &
        dtype_int2, &
        dtype_int4, &
        dtype_int8 )
    allocate(int8_1d(nij))
    call read_binary(int8_1d, f%path, f%dtype, f%rec, f%endian)

    res = ''

    do iarg = 1, argnum()
      c = ''

      if( iarg == iarg_min )then
        if( is_miss_specified )then
          write(c,wfmt_min) minval(int8_1d,mask=int8_1d/=miss_int8)
        else
          write(c,wfmt_min) minval(int8_1d)
        endif
      elseif( iarg == iarg_max )then
        if( is_miss_specified )then
          write(c,wfmt_max) maxval(int8_1d,mask=int8_1d/=miss_int8)
        else
          write(c,wfmt_max) maxval(int8_1d)
        endif
      endif

      if( len_trim(c) > 0 ) res = trim(res)//','//trim(c)
    enddo  ! iarg/

    deallocate(int8_1d)
  !-------------------------------------------------------------
  ! Case: Float
  case( dtype_real, &
        dtype_dble )
    allocate(dble_1d(nij))
    call read_binary(dble_1d, f%path, f%dtype, f%rec, f%endian)

    res = ''

    do iarg = 1, argnum()
      c = ''
      !---------------------------------------------------------
      ! min
      if( iarg == iarg_min )then
        if( is_miss_specified )then
          write(c,wfmt_min) minval(dble_1d,mask=dble_1d/=miss_dble)
        else
          write(c,wfmt_min) minval(dble_1d)
        endif
      !---------------------------------------------------------
      ! fmin
      elseif( iarg == iarg_fmin )then
        if( is_miss_specified )then
          write(c,wfmt_fmin) floor(minval(dble_1d,mask=dble_1d/=miss_dble),8)
        else
          write(c,wfmt_fmin) floor(minval(dble_1d),8)
        endif
      !---------------------------------------------------------
      ! cmin
      elseif( iarg == iarg_cmin )then
        if( is_miss_specified )then
          write(c,wfmt_fmin) ceiling(minval(dble_1d,mask=dble_1d/=miss_dble),8)
        else
          write(c,wfmt_fmin) ceiling(minval(dble_1d),8)
        endif
      !---------------------------------------------------------
      ! max
      elseif( iarg == iarg_max )then
        if( is_miss_specified )then
          write(c,wfmt_max) maxval(dble_1d,mask=dble_1d/=miss_dble)
        else
          write(c,wfmt_max) maxval(dble_1d)
        endif
      !---------------------------------------------------------
      ! fmax
      elseif( iarg == iarg_fmax )then
        if( is_miss_specified )then
          write(c,wfmt_fmax) floor(maxval(dble_1d,mask=dble_1d/=miss_dble),8)
        else
          write(c,wfmt_fmax) floor(maxval(dble_1d),8)
        endif
      !---------------------------------------------------------
      ! cmax
      elseif( iarg == iarg_cmax )then
        if( is_miss_specified )then
          write(c,wfmt_fmax) ceiling(maxval(dble_1d,mask=dble_1d/=miss_dble),8)
        else
          write(c,wfmt_fmax) ceiling(maxval(dble_1d),8)
        endif
      endif
      !---------------------------------------------------------
      if( len_trim(c) > 0 ) res = trim(res)//','//trim(c)
    enddo  ! iarg/
  endselect

  if( len_trim(res) > 0 )then
    print"(a)", trim(res(2:))
  endif

  call echo(code%ext)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------

  !-------------------------------------------------------------
  call echo(code%ret)
contains
end program main
