module lib_math
  use lib_const, only: &
    pi, &
    r2d, &
    d2r, &
    rad_0deg, &
    rad_90deg, &
    rad_180deg, &
    rad_360deg, &
    polygon_position_undef, &
    polygon_position_normal, &
    polygon_position_lon0, &
    polygon_position_polar, &
    arc_type_normal, &
    arc_type_meridian, &
    arc_type_parallel, &
    arc_convex_undef, &
    arc_convex_no, &
    arc_convex_southward, &
    arc_convex_northward, &
    stat_inclusion_undef, &
    stat_inclusion_inside, &
    stat_inclusion_outside, &
    stat_inclusion_on_side, &
    stat_inclusion_invalid, &
    stat_intersection_no, &
    stat_intersection_yes
  use lib_base, only: &
    msg_invalid_value, &
    msg_unexpected_condition
  use lib_log, only: &
    CODE, &
    str, &
    dgt, &
    echo, &
    edbg, &
    ewrn, &
    eerr
  implicit none
  private
  !-------------------------------------------------------------
  ! Procedures
  !-------------------------------------------------------------
  public :: inc

  public :: get_stats

  public :: gcd

  public :: conv_spherical_to_cartesian_deg
  public :: conv_spherical_to_cartesian_rad
  public :: conv_cartesian_to_spherical_deg
  public :: conv_cartesian_to_spherical_rad

  public :: which_is_western
  public :: western
  public :: eastern

  public :: included_in_triangle_cartesian
  public :: included_in_bbox_sphere
  public :: bbox_intersect

  public :: judge_inclusion_point_plane
  public :: judge_inclusion_polygon_plane

  public :: calc_coefs_large_arc
  public :: calc_lat_range_large_arc

  public :: londiff_deg
  public :: londiff_rad

  public :: area_sphere_rect
  public :: area_ellips_rect

  public :: area_sphere_polarrect

  public :: area_sphere_polartri1

  public :: area_sphere_polartri2
  public :: calc_vals_for_polartri_from_longit
  public :: calc_vals_for_polartri_from_latit

  public :: area_sphere_tri_south_bottom
  public :: area_sphere_tri_north_bottom

  public :: area_sphere_pl_tri_6
  public :: area_sphere_pl_tri_5

  public :: area_sphere_polartri4
  public :: area_sphere_polartri3

  public :: area_sphere_tri

  public :: area_sphere_polygon

  public :: area_sphere_intersection_polygon_polygon
  public :: area_sphere_intersection_lattice_polygon
  public :: calc_intersection_sphere_normal_normal
  public :: calc_intersection_sphere_normal_meridian

  public :: calc_intersection_sphere_normal_parallel3
  public :: calc_intersection_sphere_normal_parallel2
  public :: calc_intersection_sphere_normal_parallel1

  public :: dist_sphere

  public :: calc_inv_matrix_3x3
  public :: calc_cross_product

  public :: swap
  !-------------------------------------------------------------
  ! Interfaces
  !-------------------------------------------------------------
  interface inc
    module procedure inc_int1_default
    module procedure inc_int2_default
    module procedure inc_int4_default
    module procedure inc_int8_default
    module procedure inc_real_default
    module procedure inc_dble_default
    module procedure inc_int1_nonDefault
    module procedure inc_int2_nonDefault
    module procedure inc_int4_nonDefault
    module procedure inc_int8_nonDefault
    module procedure inc_real_nonDefault
    module procedure inc_dble_nonDefault
  end interface 

  interface get_stats
    module procedure get_stats_int1_1d
    module procedure get_stats_int1_2d
    module procedure get_stats_int2_1d
    module procedure get_stats_int2_2d
    module procedure get_stats_int4_1d
    module procedure get_stats_int4_2d
    module procedure get_stats_int8_1d
    module procedure get_stats_int8_2d
    module procedure get_stats_real_1d
    module procedure get_stats_real_2d
    module procedure get_stats_dble_1d
    module procedure get_stats_dble_2d
  end interface

  interface gcd
    module procedure gcd_int4
    module procedure gcd_int8
  end interface

  interface conv_spherical_to_cartesian_deg
    module procedure conv_spherical_to_cartesian_deg_0d_miss_unspecified
    module procedure conv_spherical_to_cartesian_deg_0d_miss_specified
    module procedure conv_spherical_to_cartesian_deg_1d_miss_unspecified
    module procedure conv_spherical_to_cartesian_deg_1d_miss_specified
  end interface

  interface conv_spherical_to_cartesian_rad
    module procedure conv_spherical_to_cartesian_rad_0d_miss_unspecified
    module procedure conv_spherical_to_cartesian_rad_0d_miss_specified
    module procedure conv_spherical_to_cartesian_rad_1d_miss_unspecified
    module procedure conv_spherical_to_cartesian_rad_1d_miss_specified
  end interface

  interface conv_cartesian_to_spherical_deg
    module procedure conv_cartesian_to_spherical_deg_0d_miss_unspecified
    module procedure conv_cartesian_to_spherical_deg_0d_miss_specified
    module procedure conv_cartesian_to_spherical_deg_1d_miss_unspecified
    module procedure conv_cartesian_to_spherical_deg_1d_miss_specified
  end interface

  interface conv_cartesian_to_spherical_rad
    module procedure conv_cartesian_to_spherical_rad_0d_miss_unspecified
    module procedure conv_cartesian_to_spherical_rad_0d_miss_specified
    module procedure conv_cartesian_to_spherical_rad_1d_miss_unspecified
    module procedure conv_cartesian_to_spherical_rad_1d_miss_specified
  end interface

  interface londiff_deg
    module procedure londiff_deg_0d
    module procedure londiff_deg_1d
  end interface

  interface londiff_rad
    module procedure londiff_rad_0d
    module procedure londiff_rad_1d
  end interface

  interface calc_coefs_large_arc
    module procedure calc_coefs_large_arc_spherical
    module procedure calc_coefs_large_arc_cartesian
  end interface

  interface calc_intersection_sphere_normal_meridian
    module procedure calc_intersection_sphere_normal_meridian_0d
    module procedure calc_intersection_sphere_normal_meridian_1d
  end interface

  interface area_sphere_rect
    module procedure area_sphere_rect_lat0d
    module procedure area_sphere_rect_lat1d
  end interface

  interface area_ellips_rect
    module procedure area_ellips_rect_lat0d
    module procedure area_ellips_rect_lat1d
  end interface

  interface area_sphere_pl_tri_6
    module procedure area_sphere_pl_tri_6_spherical
  end interface area_sphere_pl_tri_6

  interface area_sphere_pl_tri_5
    module procedure area_sphere_pl_tri_5_spherical
  end interface area_sphere_pl_tri_5

  interface area_sphere_polartri4
    module procedure area_sphere_polartri4_spherical
    module procedure area_sphere_polartri4_cartesian
  end interface

  interface area_sphere_polartri1
    module procedure area_sphere_polartri1_spherical
    module procedure area_sphere_polartri1_cartesian
  end interface

  interface area_sphere_polygon
    module procedure area_sphere_polygon_spherical
    module procedure area_sphere_polygon_cartesian
  end interface

  interface calc_intersection_sphere_normal_normal
    module procedure calc_intersection_sphere_normal_normal_confirmed
  end interface

  interface dist_sphere
    module procedure dist_sphere_0d
  end interface

  interface calc_cross_product
    module procedure calc_cross_product_array
    module procedure calc_cross_product_components
  end interface

  interface swap
    module procedure swap_int4
  end interface
!---------------------------------------------------------------
contains
!===============================================================
!
!===============================================================
subroutine inc_int1_default(a)
  implicit none
  integer(1), intent(inout) :: a

  a = a + 1_1
end subroutine inc_int1_default
!===============================================================
!
!===============================================================
subroutine inc_int2_default(a)
  implicit none
  integer(2), intent(inout) :: a

  a = a + 1_2
end subroutine inc_int2_default
!===============================================================
!
!===============================================================
subroutine inc_int4_default(a)
  implicit none
  integer(4), intent(inout) :: a

  a = a + 1_4
end subroutine inc_int4_default
!===============================================================
!
!===============================================================
subroutine inc_int8_default(a)
  implicit none
  integer(8), intent(inout) :: a

  a = a + 1_8
end subroutine inc_int8_default
!===============================================================
!
!===============================================================
subroutine inc_real_default(a)
  implicit none
  real(4), intent(inout) :: a

  a = a + 1.0
end subroutine inc_real_default
!===============================================================
!
!===============================================================
subroutine inc_dble_default(a)
  implicit none
  real(8), intent(inout) :: a

  a = a + 1.d0
end subroutine inc_dble_default
!===============================================================
!
!===============================================================
subroutine inc_int1_nonDefault(a, b)
  implicit none
  integer(1), intent(inout) :: a
  integer(1), intent(in)    :: b

  a = a + b
end subroutine inc_int1_nonDefault
!===============================================================
!
!===============================================================
subroutine inc_int2_nonDefault(a, b)
  implicit none
  integer(2), intent(inout) :: a
  integer(2), intent(in)    :: b

  a = a + b
end subroutine inc_int2_nonDefault
!===============================================================
!
!===============================================================
subroutine inc_int4_nonDefault(a, b)
  implicit none
  integer(4), intent(inout) :: a
  integer(4), intent(in)    :: b

  a = a + b
end subroutine inc_int4_nonDefault
!===============================================================
!
!===============================================================
subroutine inc_int8_nonDefault(a, b)
  implicit none
  integer(8), intent(inout) :: a
  integer(8), intent(in)    :: b

  a = a + b
end subroutine inc_int8_nonDefault
!===============================================================
!
!===============================================================
subroutine inc_real_nonDefault(a, b)
  implicit none
  real(4), intent(inout) :: a
  real(4), intent(in)    :: b

  a = a + b
end subroutine inc_real_nonDefault
!===============================================================
!
!===============================================================
subroutine inc_dble_nonDefault(a, b)
  implicit none
  real(8), intent(inout) :: a
  real(8), intent(in)    :: b

  a = a + b
end subroutine inc_dble_nonDefault
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine get_stats_int1_1d(dat, vmin, vmax, vsum, imin, imax, miss, mask, stat)
  implicit none
  integer, parameter :: byte = 1
  integer(byte), intent(in)            :: dat(:)
  integer(byte), intent(out), optional :: vmin, vmax
  integer(8)   , intent(out), optional :: vsum
  integer(8)   , intent(out), optional :: imin, imax
  integer(byte), intent(in) , optional :: miss
  logical      , intent(in) , optional :: mask(:)
  integer      , intent(out), optional :: stat  ! status
  integer(byte) :: vmin_, vmax_
  integer(8)    :: vsum_
  integer(8)    :: imin_, imax_
  integer(8)    :: nx, ix
  logical       :: found

  call echo(CODE%BGN, 'get_stats__MP__get_stats_int1_1d', '-p')
  !-------------------------------------------------------------
  if( present(stat) ) stat = 0

  nx = size(dat,kind=8)
  if( present(mask) )then
    if( size(mask,kind=8) /= nx )then
      call eerr(str(msg_unexpected_condition())//&
              '\n  size(mask) /= size(dat)')
    endif
  endif

  if( present(miss) )then
    vmin_ = miss
    vmax_ = miss
    vsum_ = int(miss,8)
  else
    vmin_ = 0_1
    vmax_ = 0_1
    vsum_ = 0_8
  endif

  imin_ = 0_8
  imax_ = 0_8
  !-------------------------------------------------------------
  ! Case: Both missing value and mask were specified.
  if( present(miss) .and. present(mask) )then
    found = .false.
    ix = 1_8
    do while( ix <= nx )
      if( dat(ix) /= miss .and. mask(ix) )then
        call put_values()
        found = .true.
        exit
      endif
      ix = ix + 1_8
    enddo

    if( .not. found )then
      if( present(stat) )then
        stat = 1
      else
        call eerr('No valid value was found.')
      endif
    else
      ix = ix + 1_8
      do while( ix <= nx )
        if( dat(ix) /= miss .and. mask(ix) )then
          call update_values()
        endif
        ix = ix + 1_8
      enddo
    endif
  !-------------------------------------------------------------
  ! Case: Only missing value was specified.
  elseif( present(miss) )then
    found = .false.
    ix = 1_8
    do while( ix <= nx )
      if( dat(ix) /= miss )then
        call put_values()
        found = .true.
        exit
      endif
      ix = ix + 1_8
    enddo

    if( .not. found )then
      if( present(stat) )then
        stat = 1
      else
        call eerr('No valid value was found.')
      endif
    else
      ix = ix + 1_8
      do while( ix <= nx )
        if( dat(ix) /= miss )then
          call update_values()
        endif
        ix = ix + 1_8
      enddo
    endif
  !-------------------------------------------------------------
  ! Case: Only mask was specified.
  elseif( present(mask) )then
    if( .not. any(mask) )then
      if( present(stat) )then
        stat = 1
      else
        call eerr('No valid value was found.')
      endif
    else
      vmin_ = minval(dat,mask)
      vmax_ = maxval(dat,mask)
      imin_ = minloc(dat,1,mask)
      imax_ = maxloc(dat,1,mask)

      if( present(vsum) )then
        vsum_ = 0_8
        do ix = 1_8, nx
          if( mask(ix) ) vsum_ = vsum_ + int(dat(ix),kind=8)
        enddo
      endif
    endif
  !-------------------------------------------------------------
  ! Case: Neither missing value or mask was specified.
  else
    vmin_ = minval(dat)
    vmax_ = maxval(dat)
    imin_ = minloc(dat,1)
    imax_ = maxloc(dat,1)

    if( present(vsum) )then
      vsum_ = 0_8
      do ix = 1_8, nx
        vsum_ = vsum_ + int(dat(ix),kind=8)
      enddo
    endif
  endif
  !-------------------------------------------------------------
  if( present(vmin) ) vmin = vmin_
  if( present(vmax) ) vmax = vmax_
  if( present(vsum) ) vsum = vsum_
  if( present(imin) ) imin = imin_
  if( present(imax) ) imax = imax_
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine put_values()
  implicit none

  vmin_ = dat(ix)
  vmax_ = dat(ix)
  vsum_ = int(dat(ix),8)
  imin_ = ix
  imax_ = ix
end subroutine put_values
!---------------------------------------------------------------
subroutine update_values()
  implicit none

  if( dat(ix) < vmin_ )then
    vmin_ = dat(ix)
    imin_ = ix
  endif

  if( dat(ix) > vmax_ )then
    vmax_ = dat(ix)
    imax_ = ix
  endif

  vsum_ = vsum_ + int(dat(ix),8)
end subroutine update_values
!---------------------------------------------------------------
end subroutine get_stats_int1_1d
!===============================================================
!
!===============================================================
subroutine get_stats_int2_1d(dat, vmin, vmax, vsum, imin, imax, miss, mask, stat)
  implicit none
  integer, parameter :: byte = 2
  integer(byte), intent(in)            :: dat(:)
  integer(byte), intent(out), optional :: vmin, vmax
  integer(8)   , intent(out), optional :: vsum
  integer(8)   , intent(out), optional :: imin, imax
  integer(byte), intent(in) , optional :: miss
  logical      , intent(in) , optional :: mask(:)
  integer      , intent(out), optional :: stat  ! status
  integer(byte) :: vmin_, vmax_
  integer(8)    :: vsum_
  integer(8)    :: imin_, imax_
  integer(8)    :: nx, ix
  logical       :: found

  call echo(CODE%BGN, 'get_stats__MP__get_stats_int2_1d', '-p')
  !-------------------------------------------------------------
  if( present(stat) ) stat = 0

  nx = size(dat,kind=8)
  if( present(mask) )then
    if( size(mask,kind=8) /= nx )then
      call eerr(str(msg_unexpected_condition())//&
              '\n  size(mask) /= size(dat)')
    endif
  endif

  if( present(miss) )then
    vmin_ = miss
    vmax_ = miss
    vsum_ = int(miss,8)
  else
    vmin_ = 0_2
    vmax_ = 0_2
    vsum_ = 0_8
  endif

  imin_ = 0_8
  imax_ = 0_8
  !-------------------------------------------------------------
  ! Case: Both missing value and mask were specified.
  if( present(miss) .and. present(mask) )then
    found = .false.
    ix = 1_8
    do while( ix <= nx )
      if( dat(ix) /= miss .and. mask(ix) )then
        call put_values()
        found = .true.
        exit
      endif
      ix = ix + 1_8
    enddo

    if( .not. found )then
      if( present(stat) )then
        stat = 1
      else
        call eerr('No valid value was found.')
      endif
    else
      ix = ix + 1_8
      do while( ix <= nx )
        if( dat(ix) /= miss .and. mask(ix) )then
          call update_values()
        endif
        ix = ix + 1_8
      enddo
    endif
  !-------------------------------------------------------------
  ! Case: Only missing value was specified.
  elseif( present(miss) )then
    found = .false.
    ix = 1_8
    do while( ix <= nx )
      if( dat(ix) /= miss )then
        call put_values()
        found = .true.
        exit
      endif
      ix = ix + 1_8
    enddo

    if( .not. found )then
      if( present(stat) )then
        stat = 1
      else
        call eerr('No valid value was found.')
      endif
    else
      ix = ix + 1_8
      do while( ix <= nx )
        if( dat(ix) /= miss )then
          call update_values()
        endif
        ix = ix + 1_8
      enddo
    endif
  !-------------------------------------------------------------
  ! Case: Only mask was specified.
  elseif( present(mask) )then
    if( .not. any(mask) )then
      if( present(stat) )then
        stat = 1
      else
        call eerr('No valid value was found.')
      endif
    else
      vmin_ = minval(dat,mask)
      vmax_ = maxval(dat,mask)
      imin_ = minloc(dat,1,mask)
      imax_ = maxloc(dat,1,mask)

      if( present(vsum) )then
        vsum_ = 0_8
        do ix = 1_8, nx
          if( mask(ix) ) vsum_ = vsum_ + int(dat(ix),8)
        enddo
      endif
    endif
  !-------------------------------------------------------------
  ! Case: Neither missing value or mask was specified.
  else
    vmin_ = minval(dat)
    vmax_ = maxval(dat)
    imin_ = minloc(dat,1)
    imax_ = maxloc(dat,1)

    if( present(vsum) )then
      vsum_ = 0_8
      do ix = 1_8, nx
        vsum_ = vsum_ + int(dat(ix),8)
      enddo
    endif
  endif
  !-------------------------------------------------------------
  if( present(vmin) ) vmin = vmin_
  if( present(vmax) ) vmax = vmax_
  if( present(vsum) ) vsum = vsum_
  if( present(imin) ) imin = imin_
  if( present(imax) ) imax = imax_
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine put_values()
  implicit none

  vmin_ = dat(ix)
  vmax_ = dat(ix)
  vsum_ = int(dat(ix),8)
  imin_ = ix
  imax_ = ix
end subroutine put_values
!---------------------------------------------------------------
subroutine update_values()
  implicit none

  if( dat(ix) < vmin_ )then
    vmin_ = dat(ix)
    imin_ = ix
  endif

  if( dat(ix) > vmax_ )then
    vmax_ = dat(ix)
    imax_ = ix
  endif

  vsum_ = vsum_ + int(dat(ix),8)
end subroutine update_values
!---------------------------------------------------------------
end subroutine get_stats_int2_1d
!===============================================================
!
!===============================================================
subroutine get_stats_int4_1d(dat, vmin, vmax, vsum, imin, imax, miss, mask, stat)
  implicit none
  integer, parameter :: byte = 4
  integer(byte), intent(in)            :: dat(:)
  integer(byte), intent(out), optional :: vmin, vmax
  integer(8)   , intent(out), optional :: vsum
  integer(8)   , intent(out), optional :: imin, imax
  integer(byte), intent(in) , optional :: miss
  logical      , intent(in) , optional :: mask(:)
  integer      , intent(out), optional :: stat  ! status
  integer(byte) :: vmin_, vmax_
  integer(8)    :: vsum_
  integer(8)    :: imin_, imax_
  integer(8)    :: nx, ix
  logical       :: found

  call echo(CODE%BGN, 'get_stats__MP__get_stats_int4_1d', '-p')
  !-------------------------------------------------------------
  if( present(stat) ) stat = 0

  nx = size(dat,kind=8)
  if( present(mask) )then
    if( size(mask,kind=8) /= nx )then
      call eerr(str(msg_unexpected_condition())//&
              '\n  size(mask) /= size(dat)')
    endif
  endif

  if( present(miss) )then
    vmin_ = miss
    vmax_ = miss
    vsum_ = int(miss,8)
  else
    vmin_ = 0_1
    vmax_ = 0_1
    vsum_ = 0_8
  endif

  imin_ = 0_8
  imax_ = 0_8
  !-------------------------------------------------------------
  ! Case: Both missing value and mask were specified.
  if( present(miss) .and. present(mask) )then
    found = .false.
    ix = 1_8
    do while( ix <= nx )
      if( dat(ix) /= miss .and. mask(ix) )then
        call put_values()
        found = .true.
        exit
      endif
      ix = ix + 1_8
    enddo

    if( .not. found )then
      if( present(stat) )then
        stat = 1
      else
        call eerr('No valid value was found.')
      endif
    else
      ix = ix + 1_8
      do while( ix <= nx )
        if( dat(ix) /= miss .and. mask(ix) )then
          call update_values()
        endif
        ix = ix + 1_8
      enddo
    endif
  !-------------------------------------------------------------
  ! Case: Only missing value was specified.
  elseif( present(miss) )then
    found = .false.
    ix = 1_8
    do while( ix <= nx )
      if( dat(ix) /= miss )then
        call put_values()
        found = .true.
        exit
      endif
      ix = ix + 1_8
    enddo

    if( .not. found )then
      if( present(stat) )then
        stat = 1
      else
        call eerr('No valid value was found.')
      endif
    else
      ix = ix + 1_8
      do while( ix <= nx )
        if( dat(ix) /= miss )then
          call update_values()
        endif
        ix = ix + 1_8
      enddo
    endif
  !-------------------------------------------------------------
  ! Case: Only mask was specified.
  elseif( present(mask) )then
    if( .not. any(mask) )then
      if( present(stat) )then
        stat = 1
      else
        call eerr('No valid value was found.')
      endif
    else
      vmin_ = minval(dat,mask)
      vmax_ = maxval(dat,mask)
      imin_ = minloc(dat,1,mask)
      imax_ = maxloc(dat,1,mask)

      if( present(vsum) )then
        vsum_ = 0_8
        do ix = 1_8, nx
          if( mask(ix) ) vsum_ = vsum_ + int(dat(ix),8)
        enddo
      endif
    endif
  !-------------------------------------------------------------
  ! Case: Neither missing value or mask was specified.
  else
    vmin_ = minval(dat)
    vmax_ = maxval(dat)
    imin_ = minloc(dat,1)
    imax_ = maxloc(dat,1)

    if( present(vsum) )then
      vsum_ = 0_8
      do ix = 1_8, nx
        vsum_ = vsum_ + int(dat(ix),8)
      enddo
    endif
  endif
  !-------------------------------------------------------------
  if( present(vmin) ) vmin = vmin_
  if( present(vmax) ) vmax = vmax_
  if( present(vsum) ) vsum = vsum_
  if( present(imin) ) imin = imin_
  if( present(imax) ) imax = imax_
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine put_values()
  implicit none

  vmin_ = dat(ix)
  vmax_ = dat(ix)
  vsum_ = int(dat(ix),8)
  imin_ = ix
  imax_ = ix
end subroutine put_values
!---------------------------------------------------------------
subroutine update_values()
  implicit none

  if( dat(ix) < vmin_ )then
    vmin_ = dat(ix)
    imin_ = ix
  endif

  if( dat(ix) > vmax_ )then
    vmax_ = dat(ix)
    imax_ = ix
  endif

  vsum_ = vsum_ + int(dat(ix),8)
end subroutine update_values
!---------------------------------------------------------------
end subroutine get_stats_int4_1d
!===============================================================
!
!===============================================================
subroutine get_stats_int8_1d(dat, vmin, vmax, vsum, imin, imax, miss, mask, stat)
  implicit none
  integer, parameter :: byte = 8
  integer(byte), intent(in)            :: dat(:)
  integer(byte), intent(out), optional :: vmin, vmax
  integer(8)   , intent(out), optional :: vsum
  integer(8)   , intent(out), optional :: imin, imax
  integer(byte), intent(in) , optional :: miss
  logical      , intent(in) , optional :: mask(:)
  integer      , intent(out), optional :: stat  ! status
  integer(byte) :: vmin_, vmax_
  integer(8)    :: vsum_
  integer(8)    :: imin_, imax_
  integer(8)    :: nx, ix
  logical       :: found

  call echo(CODE%BGN, 'get_stats__MP__get_stats_int8_1d', '-p')
  !-------------------------------------------------------------
  if( present(stat) ) stat = 0

  nx = size(dat,kind=8)
  if( present(mask) )then
    if( size(mask,kind=8) /= nx )then
      call eerr(str(msg_unexpected_condition())//&
              '\n  size(mask) /= size(dat)')
    endif
  endif

  if( present(miss) )then
    vmin_ = miss
    vmax_ = miss
    vsum_ = int(miss,8)
  else
    vmin_ = 0_1
    vmax_ = 0_1
    vsum_ = 0_8
  endif

  imin_ = 0_8
  imax_ = 0_8
  !-------------------------------------------------------------
  ! Case: Both missing value and mask were specified.
  if( present(miss) .and. present(mask) )then
    found = .false.
    ix = 1_8
    do while( ix <= nx )
      if( dat(ix) /= miss .and. mask(ix) )then
        call put_values()
        found = .true.
        exit
      endif
      ix = ix + 1_8
    enddo

    if( .not. found )then
      if( present(stat) )then
        stat = 1
      else
        call eerr('No valid value was found.')
      endif
    else
      ix = ix + 1_8
      do while( ix <= nx )
        if( dat(ix) /= miss .and. mask(ix) )then
          call update_values()
        endif
        ix = ix + 1_8
      enddo
    endif
  !-------------------------------------------------------------
  ! Case: Only missing value was specified.
  elseif( present(miss) )then
    found = .false.
    ix = 1_8
    do while( ix <= nx )
      if( dat(ix) /= miss )then
        call put_values()
        found = .true.
        exit
      endif
      ix = ix + 1_8
    enddo

    if( .not. found )then
      if( present(stat) )then
        stat = 1
      else
        call eerr('No valid value was found.')
      endif
    else
      ix = ix + 1_8
      do while( ix <= nx )
        if( dat(ix) /= miss )then
          call update_values()
        endif
        ix = ix + 1_8
      enddo
    endif
  !-------------------------------------------------------------
  ! Case: Only mask was specified.
  elseif( present(mask) )then
    if( .not. any(mask) )then
      if( present(stat) )then
        stat = 1
      else
        call eerr('No valid value was found.')
      endif
    else
      vmin_ = minval(dat,mask)
      vmax_ = maxval(dat,mask)
      imin_ = minloc(dat,1,mask)
      imax_ = maxloc(dat,1,mask)

      if( present(vsum) )then
        vsum_ = 0_8
        do ix = 1_8, nx
          if( mask(ix) ) vsum_ = vsum_ + int(dat(ix),8)
        enddo
      endif
    endif
  !-------------------------------------------------------------
  ! Case: Neither missing value or mask was specified.
  else
    vmin_ = minval(dat)
    vmax_ = maxval(dat)
    imin_ = minloc(dat,1)
    imax_ = maxloc(dat,1)

    if( present(vsum) )then
      vsum_ = 0_8
      do ix = 1_8, nx
        vsum_ = vsum_ + int(dat(ix),8)
      enddo
    endif
  endif
  !-------------------------------------------------------------
  if( present(vmin) ) vmin = vmin_
  if( present(vmax) ) vmax = vmax_
  if( present(vsum) ) vsum = vsum_
  if( present(imin) ) imin = imin_
  if( present(imax) ) imax = imax_
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine put_values()
  implicit none

  vmin_ = dat(ix)
  vmax_ = dat(ix)
  vsum_ = int(dat(ix),kind=8)
  imin_ = ix
  imax_ = ix
end subroutine put_values
!---------------------------------------------------------------
subroutine update_values()
  implicit none

  if( dat(ix) < vmin_ )then
    vmin_ = dat(ix)
    imin_ = ix
  endif

  if( dat(ix) > vmax_ )then
    vmax_ = dat(ix)
    imax_ = ix
  endif

  vsum_ = vsum_ + int(dat(ix),8)
end subroutine update_values
!---------------------------------------------------------------
end subroutine get_stats_int8_1d
!===============================================================
!
!===============================================================
subroutine get_stats_real_1d(dat, vmin, vmax, vsum, imin, imax, miss, mask, stat)
  implicit none
  integer, parameter :: byte = 4
  real(byte), intent(in)            :: dat(:)
  real(byte), intent(out), optional :: vmin, vmax
  real(8)   , intent(out), optional :: vsum
  integer(8), intent(out), optional :: imin, imax
  real(byte), intent(in) , optional :: miss
  logical   , intent(in) , optional :: mask(:)
  integer   , intent(out), optional :: stat  ! status
  real(byte) :: vmin_, vmax_
  real(8)    :: vsum_
  integer(8) :: imin_, imax_
  integer(8) :: nx, ix
  logical    :: found

  call echo(CODE%BGN, 'get_stats__MP__get_stats_real_1d', '-p')
  !-------------------------------------------------------------
  if( present(stat) ) stat = 0

  nx = size(dat,kind=8)
  if( present(mask) )then
    if( size(mask,kind=8) /= nx )then
      call eerr(str(msg_unexpected_condition())//&
              '\n  size(mask) /= size(dat)')
    endif
  endif

  if( present(miss) )then
    vmin_ = miss
    vmax_ = miss
    vsum_ = real(miss,8)
  else
    vmin_ = 0.0
    vmax_ = 0.0
    vsum_ = 0.d0
  endif

  imin_ = 0_8
  imax_ = 0_8
  !-------------------------------------------------------------
  ! Case: Both missing value and mask were specified.
  if( present(miss) .and. present(mask) )then
    found = .false.
    ix = 1_8
    do while( ix <= nx )
      if( dat(ix) /= miss .and. mask(ix) )then
        call put_values()
        found = .true.
        exit
      endif
      ix = ix + 1_8
    enddo

    if( .not. found )then
      if( present(stat) )then
        stat = 1
      else
        call eerr('No valid value was found.')
      endif
    else
      ix = ix + 1_8
      do while( ix <= nx )
        if( dat(ix) /= miss .and. mask(ix) )then
          call update_values()
        endif
        ix = ix + 1_8
      enddo
    endif
  !-------------------------------------------------------------
  ! Case: Only missing value was specified.
  elseif( present(miss) )then
    found = .false.
    ix = 1_8
    do while( ix <= nx )
      if( dat(ix) /= miss )then
        call put_values()
        found = .true.
        exit
      endif
      ix = ix + 1_8
    enddo

    if( .not. found )then
      if( present(stat) )then
        stat = 1
      else
        call eerr('No valid value was found.')
      endif
    else
      ix = ix + 1_8
      do while( ix <= nx )
        if( dat(ix) /= miss )then
          call update_values()
        endif
        ix = ix + 1_8
      enddo
    endif
  !-------------------------------------------------------------
  ! Case: Only mask was specified.
  elseif( present(mask) )then
    if( .not. any(mask) )then
      if( present(stat) )then
        stat = 1
      else
        call eerr('No valid value was found.')
      endif
    else
      vmin_ = minval(dat,mask)
      vmax_ = maxval(dat,mask)
      imin_ = minloc(dat,1,mask)
      imax_ = maxloc(dat,1,mask)

      if( present(vsum) )then
        vsum_ = 0.d0
        do ix = 1_8, nx
          if( mask(ix) ) vsum_ = vsum_ + real(dat(ix),8)
        enddo
      endif
    endif
  !-------------------------------------------------------------
  ! Case: Neither missing value or mask was specified.
  else
    vmin_ = minval(dat)
    vmax_ = maxval(dat)
    imin_ = minloc(dat,1)
    imax_ = maxloc(dat,1)
    if( present(vsum) )then
      vsum_ = 0.d0
      do ix = 1_8, nx
        vsum_ = vsum_ + real(dat(ix),8)
      enddo
    endif
  endif
  !-------------------------------------------------------------
  if( present(vmin) ) vmin = vmin_
  if( present(vmax) ) vmax = vmax_
  if( present(vsum) ) vsum = vsum_
  if( present(imin) ) imin = imin_
  if( present(imax) ) imax = imax_
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine put_values()
  implicit none

  vmin_ = dat(ix)
  vmax_ = dat(ix)
  vsum_ = real(dat(ix),8)
  imin_ = ix
  imax_ = ix
end subroutine put_values
!---------------------------------------------------------------
subroutine update_values()
  implicit none

  if( dat(ix) < vmin_ )then
    vmin_ = dat(ix)
    imin_ = ix
  endif

  if( dat(ix) > vmax_ )then
    vmax_ = dat(ix)
    uloc_ = ix
  endif

  vsum_ = vsum_ + real(dat(ix),8)
end subroutine update_values
!---------------------------------------------------------------
end subroutine get_stats_real_1d
!===============================================================
!
!===============================================================
subroutine get_stats_dble_1d(dat, vmin, vmax, vsum, imin, imax, miss, mask, stat)
  implicit none
  integer, parameter :: byte = 8
  real(byte), intent(in)            :: dat(:)
  real(byte), intent(out), optional :: vmin, vmax
  real(8)   , intent(out), optional :: vsum
  integer(8), intent(out), optional :: imin, imax
  real(byte), intent(in) , optional :: miss
  logical   , intent(in) , optional :: mask(:)
  integer   , intent(out), optional :: stat  ! status
  real(byte) :: vmin_, vmax_
  real(8)    :: vsum_
  integer(8) :: imin_, imax_
  integer(8) :: nx, ix
  logical    :: found

  call echo(CODE%BGN, 'get_stats__MP__get_stats_dble_1d', '-p')
  !-------------------------------------------------------------
  if( present(stat) ) stat = 0

  nx = size(dat,kind=8)
  if( present(mask) )then
    if( size(mask,kind=8) /= nx )then
      call eerr(str(msg_unexpected_condition())//&
              '\n  size(mask) /= size(dat)')
    endif
  endif

  if( present(miss) )then
    vmin_ = miss
    vmax_ = miss
    vsum_ = real(miss,8)
  else
    vmin_ = 0.d0
    vmax_ = 0.d0
    vsum_ = 0.d0
  endif

  imin_ = 0_8
  imax_ = 0_8
  !-------------------------------------------------------------
  ! Case: Both missing value and mask were specified.
  if( present(miss) .and. present(mask) )then
    found = .false.
    ix = 1_8
    do while( ix <= nx )
      if( dat(ix) /= miss .and. mask(ix) )then
        call put_values()
        found = .true.
        exit
      endif
      ix = ix + 1_8
    enddo

    if( .not. found )then
      if( present(stat) )then
        stat = 1
      else
        call eerr('No valid value was found.')
      endif
    else
      ix = ix + 1_8
      do while( ix <= nx )
        if( dat(ix) /= miss .and. mask(ix) )then
          call update_values()
        endif
        ix = ix + 1_8
      enddo
    endif
  !-------------------------------------------------------------
  ! Case: Only missing value was specified.
  elseif( present(miss) )then
    found = .false.
    ix = 1_8
    do while( ix <= nx )
      if( dat(ix) /= miss )then
        call put_values()
        found = .true.
        exit
      endif
      ix = ix + 1_8
    enddo

    if( .not. found )then
      if( present(stat) )then
        stat = 1
      else
        call eerr('No valid value was found.')
      endif
    else
      ix = ix + 1_8
      do while( ix <= nx )
        if( dat(ix) /= miss )then
          call update_values()
        endif
        ix = ix + 1_8
      enddo
    endif
  !-------------------------------------------------------------
  ! Case: Only mask was specified.
  elseif( present(mask) )then
    if( .not. any(mask) )then
      if( present(stat) )then
        stat = 1
      else
        call eerr('No valid value was found.')
      endif
    else
      vmin_ = minval(dat,mask)
      vmax_ = maxval(dat,mask)
      imin_ = minloc(dat,1,mask)
      imax_ = maxloc(dat,1,mask)

      if( present(vsum) )then
        vsum_ = sum(dat,mask)
      endif
    endif
  !-------------------------------------------------------------
  ! Case: Neither missing value or mask was specified.
  else
    vmin_ = minval(dat)
    vmax_ = maxval(dat)
    imin_ = minloc(dat,1)
    imax_ = maxloc(dat,1)

    if( present(vsum) )then
      vsum_ = sum(dat)
    endif
  endif
  !-------------------------------------------------------------
  if( present(vmin) ) vmin = vmin_
  if( present(vmax) ) vmax = vmax_
  if( present(vsum) ) vsum = vsum_
  if( present(imin) ) imin = imin_
  if( present(imax) ) imax = imax_
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine put_values()
  implicit none

  vmin_ = dat(ix)
  vmax_ = dat(ix)
  vsum_ = real(dat(ix),8)
  imin_ = ix
  imax_ = ix
end subroutine put_values
!---------------------------------------------------------------
subroutine update_values()
  implicit none

  if( dat(ix) < vmin_ )then
    vmin_ = dat(ix)
    imin_ = ix
  endif

  if( dat(ix) > vmax_ )then
    vmax_ = dat(ix)
    uloc_ = ix
  endif

  vsum_ = vsum_ + real(dat(ix),8)
end subroutine update_values
!---------------------------------------------------------------
end subroutine get_stats_dble_1d
!===============================================================
!
!===============================================================
subroutine get_stats_int1_2d(dat, vmin, vmax, vsum, imin1, imax1, imin2, imax2, miss, stat)
  use lib_base, only: is_nan
  implicit none
  integer, parameter :: byte = 1
  integer(byte), intent(in)            :: dat(:,:)
  integer(byte), intent(out), optional :: vmin, vmax
  integer(8)   , intent(out), optional :: vsum
  integer(8)   , intent(out), optional :: imin1, uloc1, imin2, uloc2
  integer(byte), intent(in) , optional :: miss
  integer      , intent(out), optional :: stat  ! status
  integer(byte) :: vmin_, vmax_
  integer(8)    :: vsum_
  integer(8)    :: imin1_, imin2_, uloc1_, uloc2_
  integer(8)    :: nx, ix, ny, iy
  logical       :: found

  call echo(CODE%BGN, 'get_stats__MP__get_stats_int1_2d', '-p')
  !-------------------------------------------------------------
  if( present(stat) ) stat = 0

  nx = size(dat,1,kind=8)
  ny = size(dat,2,kind=8)
  !-------------------------------------------------------------
  ! Case: Missing value was specified.
  if( present(miss) )then
    found = .false.
    iy = 1_8
    loop_init: &
    do while( iy <= ny )
      ix = 1_8
      do while( ix <= nx )
        if( dat(ix,iy) /= miss )then
          vmin_ = dat(ix,iy)
          vmax_ = dat(ix,iy)
          vsum_ = dat(ix,iy)
          imin1_ = ix
          imin2_ = iy
          uloc1_ = ix
          uloc2_ = iy
          found = .true.
          exit loop_init
        endif
        ix = ix + 1_8
      enddo
      iy = iy + 1_8
    enddo &
    loop_init

    if( .not. found )then
      if( present(stat) )then
        stat = 1
        vmin_ = 0_1
        vmax_ = 0_1
        vsum_ = 0_8
        imin1_ = 0_8
        imin2_ = 0_8
        uloc1_ = 0_8
        uloc2_ = 0_8
      else
        call eerr('All elements have missing value.')
      endif
    else
      ix = ix + 1_8
      do while( ix <= nx )
        if( dat(ix,iy) /= miss ) call updateValues()
        ix = ix + 1_8
      enddo
      iy = iy + 1_8
      do while( iy <= ny )
        ix = 1_8
        do while( ix <= nx )
          if( dat(ix,iy) /= miss ) call updateValues()
          ix = ix + 1_8
        enddo
        iy = iy + 1_8
      enddo
    endif
  !-------------------------------------------------------------
  ! Case: Missing value was not specified.
  else
    vmin_ = dat(1,1)
    vmax_ = dat(1,1)
    vsum_ = dat(1,1)
    imin1_ = 1_8
    imin2_ = 1_8
    uloc1_ = 1_8
    uloc2_ = 1_8

    do iy = 1_8, ny
      do ix = 1_8, nx
        call updateValues()
      enddo
    enddo
  endif
  !-------------------------------------------------------------
  if( present(vmin) ) vmin = vmin_
  if( present(vmax) ) vmax = vmax_
  if( present(vsum) ) vsum = vsum_
  if( present(imin1) ) imin1 = imin1_
  if( present(imin2) ) imin2 = imin2_
  if( present(uloc1) ) uloc1 = uloc1_
  if( present(uloc2) ) uloc2 = uloc2_
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine updateValues()
  implicit none

  if( dat(ix,iy) < vmin_ )then
    vmin_ = dat(ix,iy)
    imin1_ = ix
    imin2_ = iy
  endif

  if( dat(ix,iy) > vmax_ )then
    vmax_ = dat(ix,iy)
    uloc1_ = ix
    uloc2_ = iy
  endif

  vsum_ = vsum_ + dat(ix,iy)
end subroutine updateValues
!---------------------------------------------------------------
end subroutine get_stats_int1_2d
!===============================================================
!
!===============================================================
subroutine get_stats_int2_2d(dat, vmin, vmax, vsum, imin1, uloc1, imin2, uloc2, miss, stat)
  use lib_base, only: is_nan
  implicit none
  integer, parameter :: byte = 2
  integer(byte), intent(in)            :: dat(:,:)
  integer(byte), intent(out), optional :: vmin, vmax
  integer(8)   , intent(out), optional :: vsum
  integer(8)   , intent(out), optional :: imin1, uloc1, imin2, uloc2
  integer(byte), intent(in) , optional :: miss
  integer      , intent(out), optional :: stat  ! status
  integer(byte) :: vmin_, vmax_
  integer(8)    :: vsum_
  integer(8)    :: imin1_, imin2_, uloc1_, uloc2_
  integer(8)    :: nx, ix, ny, iy
  logical       :: found

  call echo(CODE%BGN, 'get_stats__MP__get_stats_int2_2d', '-p')
  !-------------------------------------------------------------
  if( present(stat) ) stat = 0

  nx = size(dat,1,kind=8)
  ny = size(dat,2,kind=8)
  !-------------------------------------------------------------
  ! Case: Missing value was specified.
  if( present(miss) )then
    found = .false.
    iy = 1_8
    loop_init: &
    do while( iy <= ny )
      ix = 1_8
      do while( ix <= nx )
        if( dat(ix,iy) /= miss )then
          vmin_ = dat(ix,iy)
          vmax_ = dat(ix,iy)
          vsum_ = dat(ix,iy)
          imin1_ = ix
          imin2_ = iy
          uloc1_ = ix
          uloc2_ = iy
          found = .true.
          exit loop_init
        endif
        ix = ix + 1_8
      enddo
      iy = iy + 1_8
    enddo &
    loop_init

    if( .not. found )then
      if( present(stat) )then
        stat = 1
        vmin_ = 0_2
        vmax_ = 0_2
        vsum_ = 0_8
        imin1_ = 0_8
        imin2_ = 0_8
        uloc1_ = 0_8
        uloc2_ = 0_8
      else
        call eerr('All elements have missing value.')
      endif
    else
      ix = ix + 1_8
      do while( ix <= nx )
        if( dat(ix,iy) /= miss ) call updateValues()
        ix = ix + 1_8
      enddo
      iy = iy + 1_8
      do while( iy <= ny )
        ix = 1_8
        do while( ix <= nx )
          if( dat(ix,iy) /= miss ) call updateValues()
          ix = ix + 1_8
        enddo
        iy = iy + 1_8
      enddo
    endif
  !-------------------------------------------------------------
  ! Case: Missing value was not specified.
  else
    vmin_ = dat(1,1)
    vmax_ = dat(1,1)
    vsum_ = dat(1,1)
    imin1_ = 1_8
    imin2_ = 1_8
    uloc1_ = 1_8
    uloc2_ = 1_8

    do iy = 1_8, ny
      do ix = 1_8, nx
        call updateValues()
      enddo
    enddo
  endif
  !-------------------------------------------------------------
  if( present(vmin) ) vmin = vmin_
  if( present(vmax) ) vmax = vmax_
  if( present(vsum) ) vsum = vsum_
  if( present(imin1) ) imin1 = imin1_
  if( present(imin2) ) imin2 = imin2_
  if( present(uloc1) ) uloc1 = uloc1_
  if( present(uloc2) ) uloc2 = uloc2_
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine updateValues()
  implicit none

  if( dat(ix,iy) < vmin_ )then
    vmin_ = dat(ix,iy)
    imin1_ = ix
    imin2_ = iy
  endif

  if( dat(ix,iy) > vmax_ )then
    vmax_ = dat(ix,iy)
    uloc1_ = ix
    uloc2_ = iy
  endif

  vsum_ = vsum_ + dat(ix,iy)
end subroutine updateValues
!---------------------------------------------------------------
end subroutine get_stats_int2_2d
!===============================================================
!
!===============================================================
subroutine get_stats_int4_2d(dat, vmin, vmax, vsum, imin1, uloc1, imin2, uloc2, miss, stat)
  use lib_base, only: is_nan
  implicit none
  integer, parameter :: byte = 4
  integer(byte), intent(in)            :: dat(:,:)
  integer(byte), intent(out), optional :: vmin, vmax
  integer(8)   , intent(out), optional :: vsum
  integer(8)   , intent(out), optional :: imin1, uloc1, imin2, uloc2
  integer(byte), intent(in) , optional :: miss
  integer      , intent(out), optional :: stat  ! status
  integer(byte) :: vmin_, vmax_
  integer(8)    :: vsum_
  integer(8)    :: imin1_, imin2_, uloc1_, uloc2_
  integer(8)    :: nx, ix, ny, iy
  logical       :: found

  call echo(CODE%BGN, 'get_stats__MP__get_stats_int4_2d', '-p')
  !-------------------------------------------------------------
  if( present(stat) ) stat = 0

  nx = size(dat,1,kind=8)
  ny = size(dat,2,kind=8)
  !-------------------------------------------------------------
  ! Case: Missing value was specified.
  if( present(miss) )then
    found = .false.
    iy = 1_8
    loop_init: &
    do while( iy <= ny )
      ix = 1_8
      do while( ix <= nx )
        if( dat(ix,iy) /= miss )then
          vmin_ = dat(ix,iy)
          vmax_ = dat(ix,iy)
          vsum_ = dat(ix,iy)
          imin1_ = ix
          imin2_ = iy
          uloc1_ = ix
          uloc2_ = iy
          found = .true.
          exit loop_init
        endif
        ix = ix + 1_8
      enddo
      iy = iy + 1_8
    enddo &
    loop_init

    if( .not. found )then
      if( present(stat) )then
        stat = 1
        vmin_ = 0_4
        vmax_ = 0_4
        vsum_ = 0_8
        imin1_ = 0_8
        imin2_ = 0_8
        uloc1_ = 0_8
        uloc2_ = 0_8
      else
        call eerr('All elements have missing value.')
      endif
    else
      ix = ix + 1_8
      do while( ix <= nx )
        if( dat(ix,iy) /= miss ) call updateValues()
        ix = ix + 1_8
      enddo
      iy = iy + 1_8
      do while( iy <= ny )
        ix = 1_8
        do while( ix <= nx )
          if( dat(ix,iy) /= miss ) call updateValues()
          ix = ix + 1_8
        enddo
        iy = iy + 1_8
      enddo
    endif
  !-------------------------------------------------------------
  ! Case: Missing value was not specified.
  else
    vmin_ = dat(1,1)
    vmax_ = dat(1,1)
    vsum_ = dat(1,1)
    imin1_ = 1_8
    imin2_ = 1_8
    uloc1_ = 1_8
    uloc2_ = 1_8

    do iy = 1_8, ny
      do ix = 1_8, nx
        call updateValues()
      enddo
    enddo
  endif
  !-------------------------------------------------------------
  if( present(vmin) ) vmin = vmin_
  if( present(vmax) ) vmax = vmax_
  if( present(vsum) ) vsum = vsum_
  if( present(imin1) ) imin1 = imin1_
  if( present(imin2) ) imin2 = imin2_
  if( present(uloc1) ) uloc1 = uloc1_
  if( present(uloc2) ) uloc2 = uloc2_
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine updateValues()
  implicit none

  if( dat(ix,iy) < vmin_ )then
    vmin_ = dat(ix,iy)
    imin1_ = ix
    imin2_ = iy
  endif

  if( dat(ix,iy) > vmax_ )then
    vmax_ = dat(ix,iy)
    uloc1_ = ix
    uloc2_ = iy
  endif

  vsum_ = vsum_ + dat(ix,iy)
end subroutine updateValues
!---------------------------------------------------------------
end subroutine get_stats_int4_2d
!===============================================================
!
!===============================================================
subroutine get_stats_int8_2d(dat, vmin, vmax, vsum, imin1, uloc1, imin2, uloc2, miss, stat)
  use lib_base, only: is_nan
  implicit none
  integer, parameter :: byte = 8
  integer(byte), intent(in)            :: dat(:,:)
  integer(byte), intent(out), optional :: vmin, vmax
  integer(8)   , intent(out), optional :: vsum
  integer(8)   , intent(out), optional :: imin1, uloc1, imin2, uloc2
  integer(byte), intent(in) , optional :: miss
  integer      , intent(out), optional :: stat  ! status
  integer(byte) :: vmin_, vmax_
  integer(8)    :: vsum_
  integer(8)    :: imin1_, imin2_, uloc1_, uloc2_
  integer(8)    :: nx, ix, ny, iy
  logical       :: found

  call echo(CODE%BGN, 'get_stats__MP__get_stats_int8_2d', '-p')
  !-------------------------------------------------------------
  if( present(stat) ) stat = 0

  nx = size(dat,1,kind=8)
  ny = size(dat,2,kind=8)
  !-------------------------------------------------------------
  ! Case: Missing value was specified.
  if( present(miss) )then
    found = .false.
    iy = 1_8
    loop_init: &
    do while( iy <= ny )
      ix = 1_8
      do while( ix <= nx )
        if( dat(ix,iy) /= miss )then
          vmin_ = dat(ix,iy)
          vmax_ = dat(ix,iy)
          vsum_ = dat(ix,iy)
          imin1_ = ix
          imin2_ = iy
          uloc1_ = ix
          uloc2_ = iy
          found = .true.
          exit loop_init
        endif
        ix = ix + 1_8
      enddo
      iy = iy + 1_8
    enddo &
    loop_init

    if( .not. found )then
      if( present(stat) )then
        stat = 1
        vmin_ = 0_8
        vmax_ = 0_8
        vsum_ = 0_8
        lloc1_ = 0_8
        lloc2_ = 0_8
        uloc1_ = 0_8
        uloc2_ = 0_8
      else
        call eerr('All elements have missing value.')
      endif
    else
      ix = ix + 1_8
      do while( ix <= nx )
        if( dat(ix,iy) /= miss ) call updateValues()
        ix = ix + 1_8
      enddo
      iy = iy + 1_8
      do while( iy <= ny )
        ix = 1_8
        do while( ix <= nx )
          if( dat(ix,iy) /= miss ) call updateValues()
          ix = ix + 1_8
        enddo
        iy = iy + 1_8
      enddo
    endif
  !-------------------------------------------------------------
  ! Case: Missing value was not specified.
  else
    vmin_ = dat(1,1)
    vmax_ = dat(1,1)
    vsum_ = dat(1,1)
    lloc1_ = 1_8
    lloc2_ = 1_8
    uloc1_ = 1_8
    uloc2_ = 1_8

    do iy = 1_8, ny
      do ix = 1_8, nx
        call updateValues()
      enddo
    enddo
  endif
  !-------------------------------------------------------------
  if( present(vmin) ) vmin = vmin_
  if( present(vmax) ) vmax = vmax_
  if( present(vsum) ) vsum = vsum_
  if( present(lloc1) ) lloc1 = lloc1_
  if( present(lloc2) ) lloc2 = lloc2_
  if( present(uloc1) ) uloc1 = uloc1_
  if( present(uloc2) ) uloc2 = uloc2_
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine updateValues()
  implicit none

  if( dat(ix,iy) < vmin_ )then
    vmin_ = dat(ix,iy)
    lloc1_ = ix
    lloc2_ = iy
  endif

  if( dat(ix,iy) > vmax_ )then
    vmax_ = dat(ix,iy)
    uloc1_ = ix
    uloc2_ = iy
  endif

  vsum_ = vsum_ + dat(ix,iy)
end subroutine updateValues
!---------------------------------------------------------------
end subroutine get_stats_int8_2d
!===============================================================
!
!===============================================================
subroutine get_stats_real_2d(dat, vmin, vmax, vsum, lloc1, lloc2, uloc1, uloc2, miss, stat)
  use lib_base, only: is_nan
  implicit none
  integer, parameter :: byte = 4
  real(byte), intent(in)            :: dat(:,:)
  real(byte), intent(out), optional :: vmin, vmax
  real(8)   , intent(out), optional :: vsum
  integer(8), intent(out), optional :: lloc1, uloc1, lloc2, uloc2
  real(byte), intent(in) , optional :: miss
  integer   , intent(out), optional :: stat  ! status
  real(byte) :: vmin_, vmax_
  real(8)    :: vsum_
  integer(8) :: lloc1_, lloc2_, uloc1_, uloc2_
  integer(8) :: nx, ix, ny, iy
  logical    :: found

  call echo(CODE%BGN, 'get_stats__MP__get_stats_real_2d', '-p')
  !-------------------------------------------------------------
  if( present(stat) ) stat = 0

  nx = size(dat,1,kind=8)
  ny = size(dat,2,kind=8)
  !-------------------------------------------------------------
  ! Case: Missing value was specified.
  if( present(miss) )then
    !-----------------------------------------------------------
    ! Case: $miss is NaN
    if( is_nan(miss) )then
      found = .false.
      iy = 1_8
      loop_init_nan: &
      do while( iy < ny )
        ix = 1_8
        do while( ix <= nx )
          if( .not. is_nan(dat(ix,iy)) )then
            vmin_ = dat(ix,iy)
            vmax_ = dat(ix,iy)
            vsum_ = dat(ix,iy)
            lloc1_ = ix
            lloc2_ = iy
            uloc1_ = ix
            uloc2_ = iy
            found = .true.
            exit loop_init_nan
          endif
          ix = ix + 1_8
        enddo
        iy = iy + 1_8
      enddo &
      loop_init_nan

      if( .not. found )then
        if( present(stat) )then
          stat = 1
          vmin_ = 0.0
          vmax_ = 0.0
          vsum_ = 0.d0
          lloc1_ = 0_8
          lloc2_ = 0_8
          uloc1_ = 0_8
          uloc2_ = 0_8
        else
          call eerr('All elements have missing value.')
        endif
      else
        ix = ix + 1_8
        do while( ix <= nx )
          if( .not. is_nan(dat(ix,iy)) ) call updateValues()
        enddo
        iy = iy + 1_8
        do while( iy <= ny )
          ix = 1_8
          do while( ix <= nx )
            if( .not. is_nan(dat(ix,iy)) ) call updateValues()
            ix = ix + 1_8
          enddo
          iy = iy + 1_8
        enddo
      endif
    !-----------------------------------------------------------
    ! Case: $miss is not NaN
    else
      found = .false.
      iy = 1_8
      loop_init_notnan: &
      do while( iy < ny )
        ix = 1_8
        do while( ix <= nx )
          if( dat(ix,iy) /= miss )then
            vmin_ = dat(ix,iy)
            vmax_ = dat(ix,iy)
            vsum_ = dat(ix,iy)
            lloc1_ = ix
            lloc2_ = iy
            uloc1_ = ix
            uloc2_ = iy
            found = .true.
            exit loop_init_notnan
          endif
          ix = ix + 1_8
        enddo
        iy = iy + 1_8
      enddo &
      loop_init_notnan

      if( .not. found )then
        if( present(stat) )then
          stat = 1
          vmin_ = 0.0
          vmax_ = 0.0
          vsum_ = 0.d0
          lloc1_ = 0_8
          lloc2_ = 0_8
          uloc1_ = 0_8
          uloc2_ = 0_8
        else
          call eerr('All elements have missing value.')
        endif
      else
        ix = ix + 1_8
        do while( ix <= nx )
          if( dat(ix,iy) /= miss ) call updateValues()
        enddo
        iy = iy + 1_8
        do while( iy <= ny )
          ix = 1_8
          do while( ix <= nx )
            if( dat(ix,iy) /= miss ) call updateValues()
            ix = ix + 1_8
          enddo
          iy = iy + 1_8
        enddo
      endif
    endif
  !-------------------------------------------------------------
  ! Case: Missing value was not specified.
  else
    vmin_ = dat(1,1)
    vmax_ = dat(1,1)
    vsum_ = dat(1,1)
    lloc1_ = 1_8
    lloc2_ = 1_8
    uloc1_ = 1_8
    uloc2_ = 1_8

    do iy = 1_8, ny
      do ix = 1_8, nx
        call updateValues()
      enddo
    enddo
  endif
  !-------------------------------------------------------------
  if( present(vmin) ) vmin = vmin_
  if( present(vmax) ) vmax = vmax_
  if( present(vsum) ) vsum = vsum_
  if( present(lloc1) ) lloc1 = lloc1_
  if( present(lloc2) ) lloc2 = lloc2_
  if( present(uloc1) ) uloc1 = uloc1_
  if( present(uloc2) ) uloc2 = uloc2_
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine updateValues()
  implicit none

  if( dat(ix,iy) < vmin_ )then
    vmin_ = dat(ix,iy)
    lloc1_ = ix
    lloc2_ = iy
  endif

  if( dat(ix,iy) > vmax_ )then
    vmax_ = dat(ix,iy)
    uloc1_ = ix
    uloc2_ = iy
  endif

  vsum_ = vsum_ + dat(ix,iy)
end subroutine updateValues
!---------------------------------------------------------------
end subroutine get_stats_real_2d
!===============================================================
!
!===============================================================
subroutine get_stats_dble_2d(dat, vmin, vmax, vsum, lloc1, lloc2, uloc1, uloc2, miss, stat)
  use lib_base, only: is_nan
  implicit none
  integer, parameter :: byte = 8
  real(byte), intent(in)            :: dat(:,:)
  real(byte), intent(out), optional :: vmin, vmax
  real(8)   , intent(out), optional :: vsum
  integer(8), intent(out), optional :: lloc1, uloc1, lloc2, uloc2
  real(byte), intent(in) , optional :: miss
  integer   , intent(out), optional :: stat  ! status
  real(byte) :: vmin_, vmax_
  real(8)    :: vsum_
  integer(8) :: lloc1_, lloc2_, uloc1_, uloc2_
  integer(8) :: nx, ix, ny, iy
  logical    :: found

  call echo(CODE%BGN, 'get_stats__MP__get_stats_dble_2d', '-p')
  !-------------------------------------------------------------
  if( present(stat) ) stat = 0

  nx = size(dat,1,kind=8)
  ny = size(dat,2,kind=8)
  !-------------------------------------------------------------
  ! Case: Missing value was specified.
  if( present(miss) )then
    !-----------------------------------------------------------
    ! Case: $miss is NaN
    if( is_nan(miss) )then
      found = .false.
      iy = 1_8
      loop_init_nan: &
      do while( iy < ny )
        ix = 1_8
        do while( ix <= nx )
          if( .not. is_nan(dat(ix,iy)) )then
            vmin_ = dat(ix,iy)
            vmax_ = dat(ix,iy)
            vsum_ = dat(ix,iy)
            lloc1_ = ix
            lloc2_ = iy
            uloc1_ = ix
            uloc2_ = iy
            found = .true.
            exit loop_init_nan
          endif
          ix = ix + 1_8
        enddo
        iy = iy + 1_8
      enddo &
      loop_init_nan

      if( .not. found )then
        if( present(stat) )then
          stat = 1
          vmin_ = 0.d0
          vmax_ = 0.d0
          vsum_ = 0.d0
          lloc1_ = 0_8
          lloc2_ = 0_8
          uloc1_ = 0_8
          uloc2_ = 0_8
        else
          call eerr('All elements have missing value.')
        endif
      else
        ix = ix + 1_8
        do while( ix <= nx )
          if( .not. is_nan(dat(ix,iy)) ) call updateValues()
        enddo
        iy = iy + 1_8
        do while( iy <= ny )
          ix = 1_8
          do while( ix <= nx )
            if( .not. is_nan(dat(ix,iy)) ) call updateValues()
            ix = ix + 1_8
          enddo
          iy = iy + 1_8
        enddo
      endif
    !-----------------------------------------------------------
    ! Case: $miss is not NaN
    else
      found = .false.
      iy = 1_8
      loop_init_notnan: &
      do while( iy < ny )
        ix = 1_8
        do while( ix <= nx )
          if( dat(ix,iy) /= miss )then
            vmin_ = dat(ix,iy)
            vmax_ = dat(ix,iy)
            vsum_ = dat(ix,iy)
            lloc1_ = ix
            lloc2_ = iy
            uloc1_ = ix
            uloc2_ = iy
            found = .true.
            exit loop_init_notnan
          endif
          ix = ix + 1_8
        enddo
        iy = iy + 1_8
      enddo &
      loop_init_notnan

      if( .not. found )then
        if( present(stat) )then
          stat = 1
          vmin_ = 0.d0
          vmax_ = 0.d0
          vsum_ = 0.d0
          lloc1_ = 0_8
          lloc2_ = 0_8
          uloc1_ = 0_8
          uloc2_ = 0_8
        else
          call eerr('All elements have missing value.')
        endif
      else
        ix = ix + 1_8
        do while( ix <= nx )
          if( dat(ix,iy) /= miss ) call updateValues()
        enddo
        iy = iy + 1_8
        do while( iy <= ny )
          ix = 1_8
          do while( ix <= nx )
            if( dat(ix,iy) /= miss ) call updateValues()
            ix = ix + 1_8
          enddo
          iy = iy + 1_8
        enddo
      endif
    endif
  !-------------------------------------------------------------
  ! Case: Missing value was not specified.
  else
    vmin_ = dat(1,1)
    vmax_ = dat(1,1)
    vsum_ = dat(1,1)
    lloc1_ = 1_8
    lloc2_ = 1_8
    uloc1_ = 1_8
    uloc2_ = 1_8

    do iy = 1_8, ny
      do ix = 1_8, nx
        call updateValues()
      enddo
    enddo
  endif
  !-------------------------------------------------------------
  if( present(vmin) ) vmin = vmin_
  if( present(vmax) ) vmax = vmax_
  if( present(vsum) ) vsum = vsum_
  if( present(lloc1) ) lloc1 = lloc1_
  if( present(lloc2) ) lloc2 = lloc2_
  if( present(uloc1) ) uloc1 = uloc1_
  if( present(uloc2) ) uloc2 = uloc2_
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine updateValues()
  implicit none

  if( dat(ix,iy) < vmin_ )then
    vmin_ = dat(ix,iy)
    lloc1_ = ix
    lloc2_ = iy
  endif

  if( dat(ix,iy) > vmax_ )then
    vmax_ = dat(ix,iy)
    uloc1_ = ix
    uloc2_ = iy
  endif

  vsum_ = vsum_ + dat(ix,iy)
end subroutine updateValues
!---------------------------------------------------------------
end subroutine get_stats_dble_2d
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
integer(4) function gcd_int4(x, y) result(i)
  implicit none
  integer(4), intent(in) :: x, y

  i = min(x,y)
  do while( mod(x,i) /= 0_4 .and. mod(y,i) /= 0_4 )
    i = i - 1_4
  enddo
end function gcd_int4
!===============================================================
!
!===============================================================
integer(8) function gcd_int8(x, y) result(i)
  implicit none
  integer(8), intent(in) :: x, y

  i = min(x,y)
  do while( mod(x,i) /= 0_8 .and. mod(y,i) /= 0_8 )
    i = i - 1_8
  enddo
end function gcd_int8
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
integer function which_is_western(lon1, lon2) result(stat)
  implicit none
  real(8), intent(in) :: lon1, lon2

  if( abs(lon1 - lon2) < rad_180deg )then
    if( lon1 < lon2 )then
      stat = 1
    elseif( lon1 > lon2 )then
      stat = 2
    else
      stat = 0
    endif
  else
    if( lon1 < lon2 )then
      stat = 2
    elseif( lon1 > lon2 )then
      stat = 1
    else
      stat = 0
    endif
  endif
end function which_is_western
!===============================================================
!
!===============================================================
real(8) function western(lon1, lon2)
  implicit none
  real(8), intent(in) :: lon1, lon2

  if( abs(lon1-lon2) < rad_180deg )then
    western = min(lon1, lon2)
  else
    western = max(lon1, lon2)
  endif
end function western
!===============================================================
!
!===============================================================
real(8) function eastern(lon1, lon2)
  implicit none
  real(8), intent(in) :: lon1, lon2

  if( abs(lon1-lon2) < rad_180deg )then
    eastern = max(lon1, lon2)
  else
    eastern = min(lon1, lon2)
  endif
end function eastern
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine conv_spherical_to_cartesian_deg_0d_miss_unspecified(lon, lat, x, y, z)
  implicit none
  real(8), intent(in)  :: lon, lat
  real(8), intent(out) :: x, y, z

  if( abs(lat) == 9.d1 )then
    x = 0.d0
    y = 0.d0
    z = sign(1.d0, lat)
  else
    x = cos(lat*d2r) * cos(lon*d2r)
    y = cos(lat*d2r) * sin(lon*d2r)
    z = sin(lat*d2r)
  endif
end subroutine conv_spherical_to_cartesian_deg_0d_miss_unspecified
!===============================================================
!
!===============================================================
subroutine conv_spherical_to_cartesian_deg_0d_miss_specified(lon, lat, miss, x, y, z)
  implicit none
  real(8), intent(in)  :: lon, lat
  real(8), intent(in)  :: miss
  real(8), intent(out) :: x, y, z

  if( lat == miss )then
    x = miss
    y = miss
    z = miss
  else
    call conv_spherical_to_cartesian_deg_0d_miss_unspecified(lon, lat, x, y, z)
  endif
end subroutine conv_spherical_to_cartesian_deg_0d_miss_specified
!===============================================================
!
!===============================================================
subroutine conv_spherical_to_cartesian_deg_1d_miss_unspecified(lon, lat, x, y, z)
  implicit none
  real(8), intent(in)  :: lon(:), lat(:)
  real(8), intent(out) :: x(:), y(:), z(:)
  integer :: i

  do i = 1, size(lon)
    call conv_spherical_to_cartesian_deg_0d_miss_unspecified(lon(i), lat(i), x(i), y(i), z(i))
  enddo
end subroutine conv_spherical_to_cartesian_deg_1d_miss_unspecified
!===============================================================
!
!===============================================================
subroutine conv_spherical_to_cartesian_deg_1d_miss_specified(lon, lat, miss, x, y, z)
  implicit none
  real(8), intent(in)  :: lon(:), lat(:)
  real(8), intent(in)  :: miss
  real(8), intent(out) :: x(:), y(:), z(:)
  integer :: i

  do i = 1, size(lon)
    if( lat(i) == miss )then
      x(i) = miss
      y(i) = miss
      z(i) = miss
    else
      call conv_spherical_to_cartesian_deg_0d_miss_unspecified(lon(i), lat(i), x(i), y(i), z(i))
    endif
  enddo
end subroutine conv_spherical_to_cartesian_deg_1d_miss_specified
!===============================================================
!
!===============================================================
subroutine conv_spherical_to_cartesian_rad_0d_miss_unspecified(lon, lat, x, y, z)
  implicit none
  real(8), intent(in)  :: lon, lat
  real(8), intent(out) :: x, y, z

  if( abs(lat) == rad_90deg )then
    x = 0.d0
    y = 0.d0
    z = sign(1.d0, lat)
  else
    x = cos(lat) * cos(lon)
    y = cos(lat) * sin(lon)
    z = sin(lat)
  endif
end subroutine conv_spherical_to_cartesian_rad_0d_miss_unspecified
!===============================================================
!
!===============================================================
subroutine conv_spherical_to_cartesian_rad_0d_miss_specified(lon, lat, miss, x, y, z)
  implicit none
  real(8), intent(in)  :: lon, lat
  real(8), intent(in)  :: miss
  real(8), intent(out) :: x, y, z

  if( lat == miss )then
    x = miss
    y = miss
    z = miss
  else
    call conv_spherical_to_cartesian_rad_0d_miss_unspecified(lon, lat, x, y, z)
  endif
end subroutine conv_spherical_to_cartesian_rad_0d_miss_specified
!===============================================================
!
!===============================================================
subroutine conv_spherical_to_cartesian_rad_1d_miss_unspecified(lon, lat, x, y, z)
  implicit none
  real(8), intent(in)  :: lon(:), lat(:)
  real(8), intent(out) :: x(:), y(:), z(:)
  integer :: i

  do i = 1, size(lon)
    call conv_spherical_to_cartesian_rad_0d_miss_unspecified(lon(i), lat(i), x(i), y(i), z(i))
  enddo
end subroutine conv_spherical_to_cartesian_rad_1d_miss_unspecified
!===============================================================
!
!===============================================================
subroutine conv_spherical_to_cartesian_rad_1d_miss_specified(lon, lat, miss, x, y, z)
  implicit none
  real(8), intent(in)  :: lon(:), lat(:)
  real(8), intent(in)  :: miss
  real(8), intent(out) :: x(:), y(:), z(:)
  integer :: i

  do i = 1, size(lon)
    if( lat(i) == miss )then
      x(i) = miss
      y(i) = miss
      z(i) = miss
    else
      call conv_spherical_to_cartesian_rad_0d_miss_unspecified(lon(i), lat(i), x(i), y(i), z(i))
    endif
  enddo
end subroutine conv_spherical_to_cartesian_rad_1d_miss_specified
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine conv_cartesian_to_spherical_deg_0d_miss_unspecified(x, y, z, lon, lat, miss_s)
  implicit none
  real(8), intent(in)  :: x, y, z
  real(8), intent(out) :: lon, lat
  real(8), intent(in)  :: miss_s  ! missing value of spherical

  if( x == 0.d0 .and. y == 0.d0 )then
    lat = sign(9.d1, z)
    lon = miss_s
  else
    lat = asin( z / sqrt(x**2+y**2+z**2) ) * r2d
    lon = atan2( y, x ) * r2d
    if( lon < 0.d0 ) lon = lon + 3.6d2
  endif
end subroutine conv_cartesian_to_spherical_deg_0d_miss_unspecified
!===============================================================
!
!===============================================================
subroutine conv_cartesian_to_spherical_deg_0d_miss_specified(x, y, z, lon, lat, miss_c, miss_s)
  implicit none
  real(8), intent(in)  :: x, y, z
  real(8), intent(out) :: lon, lat
  real(8), intent(in)  :: miss_c  ! missing value of cartesian
  real(8), intent(in)  :: miss_s  ! missing value of spherical

  if( x == miss_c )then
    lon = miss_s
    lat = miss_s
  else
    call conv_cartesian_to_spherical_deg_0d_miss_unspecified(x, y, z, lon, lat, miss_s)
  endif
end subroutine conv_cartesian_to_spherical_deg_0d_miss_specified
!===============================================================
!
!===============================================================
subroutine conv_cartesian_to_spherical_deg_1d_miss_unspecified(x, y, z, lon, lat, miss_s)
  implicit none
  real(8), intent(in)  :: x(:), y(:), z(:)
  real(8), intent(out) :: lon(:), lat(:)
  real(8), intent(in)  :: miss_s  ! missing value of spherical

  integer :: i

  do i = 1, size(x)
    call conv_cartesian_to_spherical_deg_0d_miss_unspecified(&
           x(i), y(i), z(i), lon(i), lat(i), miss_s)
  enddo
end subroutine conv_cartesian_to_spherical_deg_1d_miss_unspecified
!===============================================================
!
!===============================================================
subroutine conv_cartesian_to_spherical_deg_1d_miss_specified(x, y, z, lon, lat, miss_c, miss_s)
  implicit none
  real(8), intent(in)  :: x(:), y(:), z(:)
  real(8), intent(out) :: lon(:), lat(:)
  real(8), intent(in)  :: miss_c  ! missing value of cartesian
  real(8), intent(in)  :: miss_s  ! missing value of spherical
  integer :: i

  do i = 1, size(x)
    if( x(i) == miss_c )then
      lon(i) = miss_s
      lat(i) = miss_s
    else
      call conv_cartesian_to_spherical_deg_0d_miss_unspecified(&
               x(i), y(i), z(i), lon(i), lat(i), miss_s)
    endif
  enddo
end subroutine conv_cartesian_to_spherical_deg_1d_miss_specified
!===============================================================
!
!===============================================================
subroutine conv_cartesian_to_spherical_rad_0d_miss_unspecified(x, y, z, lon, lat, miss_s)
  implicit none
  real(8), intent(in)  :: x, y, z
  real(8), intent(out) :: lon, lat
  real(8), intent(in)  :: miss_s  ! missing value of spherical

  real(8) :: r

  r = sqrt(x**2 + y**2 + z**2)
  lon = atan2(y,x)
  lat = asin(z/r)

  if( x == 0.d0 .and. y == 0.d0 )then
    lat = sign(rad_90deg, z)
    lon = miss_s
  else
    lat = asin( z / sqrt(x**2+y**2+z**2) )
    lon = atan2( y, x )
    if( lon < rad_0deg ) lon = lon + rad_360deg
  endif
end subroutine conv_cartesian_to_spherical_rad_0d_miss_unspecified
!===============================================================
!
!===============================================================
subroutine conv_cartesian_to_spherical_rad_0d_miss_specified(x, y, z, lon, lat, miss_c, miss_s)
  implicit none
  real(8), intent(in)  :: x, y, z
  real(8), intent(out) :: lon, lat
  real(8), intent(in)  :: miss_c  ! missing value of cartesian
  real(8), intent(in)  :: miss_s  ! missing value of spherical

  if( x == miss_c )then
    lon = miss_s
    lat = miss_s
  else
    call conv_cartesian_to_spherical_rad_0d_miss_unspecified(x, y, z, lon, lat, miss_s)
  endif
end subroutine conv_cartesian_to_spherical_rad_0d_miss_specified
!===============================================================
!
!===============================================================
subroutine conv_cartesian_to_spherical_rad_1d_miss_unspecified(x, y, z, lon, lat, miss_s)
  implicit none
  real(8), intent(in)  :: x(:), y(:), z(:)
  real(8), intent(out) :: lon(:), lat(:)
  real(8), intent(in)  :: miss_s  ! missing value of spherical

  integer :: i

  do i = 1, size(x)
    call conv_cartesian_to_spherical_rad_0d_miss_unspecified(&
             x(i), y(i), z(i), lon(i), lat(i), miss_s)
  enddo
end subroutine conv_cartesian_to_spherical_rad_1d_miss_unspecified
!===============================================================
!
!===============================================================
subroutine conv_cartesian_to_spherical_rad_1d_miss_specified(x, y, z, lon, lat, miss_c, miss_s)
  implicit none
  real(8), intent(in)  :: x(:), y(:), z(:)
  real(8), intent(out) :: lon(:), lat(:)
  real(8), intent(in)  :: miss_c  ! missing value of cartesian
  real(8), intent(in)  :: miss_s  ! missing value of spherical

  integer :: i

  do i = 1, size(x)
    if( x(i) == miss_c )then
      lon(i) = miss_s
      lat(i) = miss_s
    else
      call conv_cartesian_to_spherical_rad_0d_miss_unspecified(&
               x(i), y(i), z(i), lon(i), lat(i), miss_s)
    endif
  enddo
end subroutine conv_cartesian_to_spherical_rad_1d_miss_specified
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
logical function included_in_triangle_cartesian(a1, a2, a3, px, py, pz) result(included)
  implicit none
  real(8), intent(in)  :: a1(3), a2(3), a3(3)
  real(8), intent(in)  :: px, py, pz
  real(8) :: inv(3,3), coef

  call calc_inv_matrix_3x3(reshape((/a1(1),a2(1),a3(1), &
                                     a1(2),a2(2),a3(2), &
                                     a1(3),a2(3),a3(3)/), (/3,3/)), &
                           inv, coef)

  if( coef == 0.d0 )then
    included = .false.
  else
    included = ( .not. &
        (inv(1,1)*px + inv(2,1)*py + inv(3,1)*pz < 0.d0 .or. &
         inv(1,2)*px + inv(2,2)*py + inv(3,2)*pz < 0.d0 .or. &
         inv(1,3)*px + inv(2,3)*py + inv(3,3)*pz < 0.d0) )
  endif
end function included_in_triangle_cartesian
!===============================================================
!
!===============================================================
logical function included_in_bbox_sphere(&
    west, east, south, north, lon0, plon, plat) &
    result(included)
  implicit none
  real(8), intent(in) :: west, east, south, north
  real(8), intent(in) :: plon, plat
  logical, intent(in) :: lon0

  if( plat < south .or. plat > north )then
    included = .false.
    return
  endif

  included = ( (lon0 .and. (west < plon .or. plon < east)) .or. &
               (.not. lon0 .and. (west < plon .and. plon < east)) )
end function included_in_bbox_sphere
!===============================================================
!
!===============================================================
logical function bbox_intersect(&
    ssouth, snorth, swest, seast, slon0, &
    tsouth, tnorth, twest, teast, tlon0) &
    result(res)
  implicit none
  real(8), intent(in) :: ssouth, snorth, swest, seast, &
                         tsouth, tnorth, twest, teast  ![rad]
  logical, intent(in) :: slon0, tlon0

  if( snorth <= tsouth .or. ssouth >= tnorth )then
    res = .false.
  elseif( .not. slon0 .and. .not. tlon0 )then
    res = ( .not. (swest >= teast .or. seast <= twest) )
  elseif( .not. tlon0 )then
    res = ( swest < teast .or. twest < seast )
  elseif( .not. slon0 )then
    res = ( twest < seast .or. swest < teast )
  else
    res = .true.
  endif
end function bbox_intersect
!===============================================================
!
!===============================================================
subroutine judge_inclusion_point_plane(sx, sy, txs, tys, stat)
  implicit none
  real(8), intent(in)  :: sx, sy
  real(8), intent(in)  :: txs(:), tys(:)
  integer, intent(out) :: stat

  integer :: tijmax, tij
  integer :: tij_this, tij_prev, tij_next
  real(8) :: tx1, ty1, tx2, ty2
  real(8) :: a, b
  real(8) :: py
  integer :: nLines_above, nLines_below

  call echo(CODE%BGN, 'judge_inclusion_point_plane', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  tijmax = size(txs)

  stat = STAT_INCLUSION_UNDEF
  nLines_above = 0
  nLines_below = 0

  tx2 = txs(tijmax)
  ty2 = tys(tijmax)

  do tij = 1, tijmax
    tx1 = tx2
    ty1 = ty2
    tx2 = txs(tij)
    ty2 = tys(tij)
    !-----------------------------------------------------------
    ! Skip or exit if line is vertical.
    !-----------------------------------------------------------
    if( tx1 == tx2 )then
      if( min(ty1,ty2) <= sy .and. sy <= max(ty1,ty2) )then
        stat = STAT_INCLUSION_ON_SIDE
        exit
      else
        cycle
      endif
    endif
    !-----------------------------------------------------------
    ! Skip if out of range.
    !-----------------------------------------------------------
    if( sx < min(tx1,tx2) .or. max(tx1,tx2) < sx )then
      cycle
    endif
    !-----------------------------------------------------------
    ! Skip if sx == max(tx1,tx2)
    !-----------------------------------------------------------
    if( sx == max(tx1,tx2) )then
      cycle
    !-----------------------------------------------------------
    ! Skip if sx == min(tx1,tx2) and min(tx1,tx2) is connected
    ! to two lines who have opposite directions.
    !-----------------------------------------------------------
    elseif( sx == min(tx1,tx2) )then
      if( tx1 < tx2 )then
        tij_this = tij - 1
        if( tij_this == 0 ) tij_this = tijmax
      else
        tij_this = tij
      endif

      tij_prev = tij_this
      do while( txs(tij_prev) == sx )
        tij_prev = tij_prev - 1
        if( tij_prev == 0 ) tij_prev = tijmax
      enddo

      tij_next = tij
      do while( txs(tij_next) == sx )
        tij_next = tij_next + 1
        if( tij_next == tijmax+1 ) tij_next = 1
      enddo

      if( txs(tij_prev) < sx .eqv. txs(tij_next) < sx )then
        cycle
      endif
    endif
    !-----------------------------------------------------------
    ! Judge if the line is above or below the point.
    !-----------------------------------------------------------
    ! ty = a*tx + b
    a = (ty2 - ty1) / (tx2 - tx1)
    b = -(ty2*tx1 - ty1*tx2) / (tx2 - tx1)

    ! y of intersection of vertical line x=sx and line t: x=a*x+b
    py = a*sx + b

    if( py > sy )then
      nLines_above = nLines_above + 1
    elseif( py < sy )then
      nLines_below = nLines_below + 1
    else
      stat = STAT_INCLUSION_ON_SIDE
      exit
    endif
  enddo  ! tij/
  !-------------------------------------------------------------
  ! Judge the status.
  !-------------------------------------------------------------
  selectcase( stat )
  case( STAT_INCLUSION_UNDEF )
    if( mod(nLines_above + nLines_below, 2) /= 0 )then
      call ewrn(str(msg_unexpected_condition())//&
              '\n  mod(nLines_above + nLines_below, 2) /= 0')
      stat = stat_inclusion_invalid
    else
      if( mod(nLines_below,2) == 0 )then
        stat = STAT_INCLUSION_OUTSIDE
      else
        stat = STAT_INCLUSION_INSIDE
      endif
    endif
  case( STAT_INCLUSION_ON_SIDE )
    continue
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  stat: '//str(stat))
  endselect
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine judge_inclusion_point_plane
!===============================================================
!
!===============================================================
subroutine judge_inclusion_polygon_plane(sxs, sys, txs, tys, stat)
  implicit none
  real(8), intent(in)  :: sxs(:), sys(:)
  real(8), intent(in)  :: txs(:), tys(:)
  integer, intent(out) :: stat

  integer :: sijmax, sij
  integer :: nPoints_inside, &
             nPoints_outside, &
             nPoints_on_side, &
             nPoints_invalid
  integer :: stat_point

  call echo(CODE%BGN, 'judge_inclusion_polygon_plane', '-p')
  !-------------------------------------------------------------
  ! Count points for each condition.
  !-------------------------------------------------------------
  sijmax = size(sxs)

  nPoints_inside  = 0
  nPoints_outside = 0
  nPoints_on_side = 0
  nPoints_invalid = 0

  do sij = 1, sijmax
    call judge_inclusion_point_plane(sxs(sij), sys(sij), txs, tys, stat_point)

    selectcase( stat_point )
    case( STAT_INCLUSION_INSIDE )
      nPoints_inside = nPoints_inside + 1
    case( STAT_INCLUSION_OUTSIDE )
      nPoints_outside = nPoints_outside + 1
    case( STAT_INCLUSION_ON_SIDE )
      nPoints_on_side = nPoints_on_side + 1
    case( STAT_INCLUSION_INVALID )
      nPoints_invalid = nPoints_invalid + 1
    case default
      call eerr(str(msg_invalid_value())//&
              '\n  stat: '//str(stat))
    endselect
  enddo
  !-------------------------------------------------------------
  ! Judge status of inclusion.
  !-------------------------------------------------------------
  if( nPoints_invalid > 0 )then
    call eerr(str(msg_unexpected_condition())//&
            '\n  nPoints        : '//str(sijmax)//&
            '\n  nPoints_invalid: '//str(nPoints_invalid,dgt(sijmax))//&
            '\n  nPoints_inside : '//str(nPoints_inside,dgt(sijmax))//&
            '\n  nPoints_outside: '//str(nPoints_outside,dgt(sijmax))//&
            '\n  nPoints_on_side: '//str(nPoints_on_side,dgt(sijmax)))
  else
    if( nPoints_inside > 0 .and. nPoints_outside == 0 )then
      stat = STAT_INCLUSION_INSIDE
    elseif( nPoints_inside == 0 .and. nPoints_outside > 0 )then
      stat = STAT_INCLUSION_OUTSIDE
    else
      call eerr(str(msg_unexpected_condition())//&
              '\n  nPoints_inside > 0 .and. nPoints_outside > 0'//&
              '\n  nPoints        : '//str(sijmax)//&
              '\n  nPoints_inside : '//str(nPoints_inside,dgt(sijmax))//&
              '\n  nPoints_outside: '//str(nPoints_outside,dgt(sijmax))//&
              '\n  nPoints_on_side: '//str(nPoints_on_side,dgt(sijmax)))
    endif
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine judge_inclusion_polygon_plane
!===============================================================
!
!===============================================================





!===============================================================
! Calc. coefs. of plane OAB (ax + by + cz = 0)
!===============================================================
subroutine calc_coefs_large_arc_spherical(lon1, lat1, lon2, lat2, a, b, c)
  implicit none
  real(8), intent(in)  :: lon1, lat1, lon2, lat2
  real(8), intent(out) :: a, b, c
  real(8) :: x1, y1, z1, x2, y2, z2

  x1 = cos(lat1) * cos(lon1)
  y1 = cos(lat1) * sin(lon1)
  z1 = sin(lat1)

  x2 = cos(lat2) * cos(lon2)
  y2 = cos(lat2) * sin(lon2)
  z2 = sin(lat2)

  call calc_cross_product(x1, y1, z1, x2, y2, z2, a, b, c)
end subroutine calc_coefs_large_arc_spherical
!===============================================================
!
!===============================================================
subroutine calc_coefs_large_arc_cartesian(x1, y1, z1, x2, y2, z2, a, b, c)
  implicit none
  real(8), intent(in)  :: x1, y1, z1, x2, y2, z2
  real(8), intent(out) :: a, b, c

  call calc_cross_product(x1, y1, z1, x2, y2, z2, a, b, c)
end subroutine calc_coefs_large_arc_cartesian
!===============================================================
!
!===============================================================
subroutine calc_lat_range_large_arc(&
    lon1, lat1, lon2, lat2, &
    south, north, convex, lontop, lattop, &
    a, b, c)
  implicit none
  real(8)   , intent(in)  :: lon1, lat1, lon2, lat2
  real(8)   , intent(out) :: south, north
  integer(1), intent(out) :: convex
  real(8)   , intent(out) :: lontop, lattop
  real(8)   , intent(in), optional :: a, b, c
  real(8) :: a_, b_, c_
  real(8) :: plon, qlon
  real(8) :: wlon, wlat, elon, elat
  logical :: is_convex

  call echo(CODE%BGN, 'calc_lat_range_large_arc', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  south = min(lat1, lat2)
  north = max(lat1, lat2)

  convex = arc_convex_no
  lontop = 0.d0
  lattop = 0.d0
  !-------------------------------------------------------------
  ! Case: Arc intersects with the equator
  if( lat1 == rad_0deg .or. lat2 == rad_0deg .or. sign(1.d0,lat1) /= sign(1.d0,lat2) )then
    continue
  !-------------------------------------------------------------
  ! Case: Arc is meridian
  elseif( lon1 == lon2 )then
    continue
  !-------------------------------------------------------------
  ! Case: Others
  else
    if( present(a) )then
      a_ = a
      b_ = b
      c_ = c
    else
      call calc_coefs_large_arc(lon1, lat1, lon2, lat2, a_, b_, c_)
    endif
    !-----------------------------------------------------------
    ! Case: Plane is vertical
    if( c_ == 0.d0 )then
      south = min(lat1, lat2)
      north = max(lat1, lat2)
    !-----------------------------------------------------------
    ! Case: Plane is not vertical
    else
      plon = atan2(b_, a_)
      if( plon < rad_0deg ) plon = plon + rad_360deg

      if( plon < rad_180deg )then
        qlon = plon + rad_180deg
      else
        qlon = plon - rad_180deg
      endif

      is_convex = .false.
      if( abs(lon1 - lon2) < rad_180deg )then
        if( min(lon1,lon2) < plon .and. plon < max(lon1,lon2) )then
          lontop = plon
          is_convex = .true.
        elseif( min(lon1,lon2) < qlon .and. qlon < max(lon1,lon2) )then
          lontop = qlon
          is_convex = .true.
        endif
      else
        if( max(lon1,lon2) < plon .or. plon < min(lon1,lon2) )then
          lontop = plon
          is_convex = .true.
        elseif( max(lon1,lon2) < qlon .or. qlon < min(lon1,lon2) )then
          lontop = qlon
          is_convex = .true.
        endif
      endif

      if( is_convex )then
        if( abs(lon1 - lon2) < rad_180deg )then
          if( lon1 < lon2 )then
            wlon = lon1
            wlat = lat1
            elon = lon2
            elat = lat2
          else
            wlon = lon2
            wlat = lat2
            elon = lon1
            elat = lat1
          endif
        else
          if( lon1 > lon2 )then
            wlon = lon1
            wlat = lat1
            elon = lon2
            elat = lat2
          else
            wlon = lon2
            wlat = lat2
            elon = lon1
            elat = lat1
          endif
        endif

        call calc_intersection_sphere_normal_meridian(wlon, wlat, elon, elat, lontop, lattop)
        !-------------------------------------------------------
        ! Northern hemisphere
        if( min(lat1,lat2) > rad_0deg )then
          convex = arc_convex_northward
          north = lattop
        !-------------------------------------------------------
        ! Southern hemisphere
        else
          convex = arc_convex_southward
          south = lattop
        endif
      endif
    endif
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine calc_lat_range_large_arc
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
real(8) function londiff_deg_0d(lon1, lon2) result(lon)
  implicit none
  real(8), intent(in) :: lon1, lon2  ![deg]

  lon = abs(lon1 - lon2)
  if( lon > 1.8d2 ) lon = 3.6d2 - lon
end function londiff_deg_0d
!===============================================================
!
!===============================================================
function londiff_deg_1d(lon1, lon2) result(lon)
  implicit none
  real(8), intent(in) :: lon1(:), lon2(:)  ![deg]
  real(8)             :: lon(size(lon1))

  integer :: i

  do i = 1, size(lon1)
    lon(i) = londiff_deg_0d(lon1(i), lon2(i))
  enddo
end function londiff_deg_1d
!===============================================================
!
!===============================================================
real(8) function londiff_rad_0d(lon1, lon2) result(lon)
  implicit none
  real(8), intent(in) :: lon1, lon2  ![rad]

  lon = abs(lon1 - lon2)
  if( lon > rad_180deg ) lon = rad_360deg - lon
end function londiff_rad_0d
!===============================================================
!
!===============================================================
function londiff_rad_1d(lon1, lon2) result(lon)
  implicit none
  real(8), intent(in) :: lon1(:), lon2(:)  ![rad]
  real(8)             :: lon(size(lon1))

  integer :: i

  do i = 1, size(lon1)
    lon(i) = londiff_rad_0d(lon1(i), lon2(i))
  enddo
end function londiff_rad_1d
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
integer function sign_lon(lon1, lon2) result(sgn)
  implicit none
  real(8), intent(in) :: lon1, lon2

  if( abs(lon1-lon2) < rad_180deg )then
    if( lon1 < lon2 )then
      sgn = 1
    else
      sgn = -1
    endif
  else
    if( lon1 > lon2 )then
      sgn = 1
    else
      sgn = -1
    endif
  endif
end function sign_lon
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
! Calc. longit. and latit. of the intersection.
! Intersection is confirmed.
!===============================================================
subroutine calc_intersection_sphere_normal_normal_confirmed(&
    sx1, sy1, sz1, sx2, sy2, sz2, sa, sb, sc, &
    tx1, ty1, tz1, tx2, ty2, tz2, ta, tb, tc, &
    west, east, plon, plat)
  implicit none
  real(8), intent(in) :: sx1, sy1, sz1, sx2, sy2, sz2
  real(8), intent(in) :: sa, sb, sc
  real(8), intent(in) :: tx1, ty1, tz1, tx2, ty2, tz2
  real(8), intent(in) :: ta, tb, tc
  real(8), intent(in) :: west, east
  real(8), intent(out) :: plon, plat

  real(8) :: dp_S1_IT, dp_S2_IT, dp_T1_IS, dp_T2_IS
  real(8) :: px, py, pz
  real(8) :: plon1, plat1, plon2, plat2
  real(8) :: diff_plon1_west, diff_plon1_east, diff_plon2_west, diff_plon2_east
  real(8) :: diff_min

  real(8), parameter :: miss_lonlat = -999.d0
  !-------------------------------------------------------------
  dp_S1_IT = sx1*ta + sy1*tb + sz1*tc  ! dot products
  dp_S2_IT = sx2*ta + sy2*tb + sz2*tc
  dp_T1_IS = tx1*sa + ty1*sb + tz1*sc
  dp_T2_IS = tx2*sa + ty2*sb + tz2*sc

!  call edbg('s1 ('//str((/sx1,sy1,sz1/),'es10.3',',')//')')
!  call edbg('s2 ('//str((/sx2,sy2,sz2/),'es10.3',',')//')')
!  call edbg('t1 ('//str((/tx1,ty1,tz1/),'es10.3',',')//')')
!  call edbg('t2 ('//str((/tx2,ty2,tz2/),'es10.3',',')//')')
!  call edbg('IS ('//str((/sa,sb,sc/),'es15.8',',')//')')
!  call edbg('IT ('//str((/ta,tb,tc/),'es15.8',',')//')')
!  call edbg('S1_IT '//str(dp_S1_IT,'es15.8'))
!  call edbg('S2_IT '//str(dp_S2_IT,'es15.8'))
!  call edbg('T1_IS '//str(dp_T1_IS,'es15.8'))
!  call edbg('T2_IS '//str(dp_T2_IS,'es15.8'))

  call calc_cross_product(sa, sb, sc, ta, tb, tc, px, py, pz)

  if( dp_S1_IT > 0.d0 .and. dp_S2_IT < 0.d0 .and. &
      dp_T1_IS < 0.d0 .and. dp_T2_IS > 0.d0 )then
    call conv_cartesian_to_spherical_rad(px, py, pz, plon, plat, miss_lonlat)
  elseif( dp_S1_IT < 0.d0 .and. dp_S2_IT > 0.d0 .and. &
          dp_T1_IS > 0.d0 .and. dp_T2_IS < 0.d0 )then
    call conv_cartesian_to_spherical_rad(-px, -py, -pz, plon, plat, miss_lonlat)
  else
    call conv_cartesian_to_spherical_rad(px, py, pz, plon1, plat1, miss_lonlat)

    if( plon1 > rad_180deg )then
      plon2 = plon1 - rad_180deg
    else
      plon2 = plon1 + rad_180deg
    endif
    plat2 = -plat1

    if( which_is_western(plon1,west) /= 1 .and. which_is_western(plon1,east) /= 2 )then
      plon = plon1
      plat = plat1
    elseif( which_is_western(plon2,west) /= 1 .and. which_is_western(plon2,east) /= 2 )then
      plon = plon2
      plat = plat2
    else
      diff_plon1_west = abs(londiff_rad(plon1,west))
      diff_plon1_east = abs(londiff_rad(plon1,east))
      diff_plon2_west = abs(londiff_rad(plon2,west))
      diff_plon2_east = abs(londiff_rad(plon2,east))
      diff_min = min(diff_plon1_west, diff_plon1_east, diff_plon2_west, diff_plon2_east)
      if( diff_min == diff_plon1_west )then
        plon = west
        plat = plat1
      elseif( diff_min == diff_plon1_east )then
        plon = east
        plat = plat1
      elseif( diff_min == diff_plon2_west )then
        plon = west
        plat = plat2
      elseif( diff_min == diff_plon2_east )then
        plon = east
        plat = plat2
      endif

      call edbg('Calculated intersection is out of range.')
    endif
  endif
  !-------------------------------------------------------------
end subroutine calc_intersection_sphere_normal_normal_confirmed
!===============================================================
! Calc. latit. of the intersection
!===============================================================
subroutine calc_intersection_sphere_normal_meridian_0d(&
    wlon, wlat, elon, elat, lon, lat)
  implicit none
  real(8), intent(in)  :: wlon, wlat, elon, elat
  real(8), intent(in)  :: lon
  real(8), intent(out) :: lat

  if( wlon < elon )then
    lat = atan( (tan(wlat)*sin(elon-lon) + tan(elat)*sin(lon-wlon)) / sin(elon-wlon) )
  else
    lat = atan( (tan(wlat)*sin(londiff_rad(elon,lon)) + tan(elat)*sin(londiff_rad(wlon,lon))) &
                  / sin(rad_360deg-wlon+elon) )
  endif
end subroutine calc_intersection_sphere_normal_meridian_0d
!===============================================================
!
!===============================================================
subroutine calc_intersection_sphere_normal_meridian_1d(&
    wlon, wlat, elon, elat, lon, lat)
  implicit none
  real(8), intent(in)  :: wlon, wlat, elon, elat
  real(8), intent(in)  :: lon(:)
  real(8), intent(out) :: lat(:)

  integer :: i, imax

  imax = size(lon)

  if( wlon < elon )then
    do i = 1, imax
      lat(i) = atan( (tan(wlat)*sin(elon-lon(i)) + tan(elat)*sin(lon(i)-wlon)) / sin(elon-wlon) )
    enddo
  else
    do i = 1, imax
      lat(i) = atan( (tan(wlat)*sin(londiff_rad(elon,lon(i))) &
                        + tan(elat)*sin(londiff_rad(wlon,lon(i)))) &
                       / sin(rad_360deg-wlon+elon) )
    enddo
  endif
end subroutine calc_intersection_sphere_normal_meridian_1d
!===============================================================
! Calc. longit. of the intersection.
! 交差することが事前に分かっている場合に使う。
! 交点が (wlon, elon) の範囲外になった場合、
! その値を wlon または elon に補正する。
!===============================================================
subroutine calc_intersection_sphere_normal_parallel3(&
    wlon, elon, a, b, c, sgn, z, plon)
  implicit none
  real(8), intent(in)  :: wlon, elon
  real(8), intent(in)  :: a, b, c
  integer, intent(in)  :: sgn
  real(8), intent(in)  :: z
  real(8), intent(out) :: plon

  real(8) :: d
  real(8) :: x_numer, y_numer

  d = sqrt(max(0.d0, (a**2+b**2) - (a**2+b**2+c**2) * z**2))

  x_numer = -a*c*z - sgn*b*d
  y_numer = -b*c*z + sgn*a*d

  plon = atan2(y_numer, x_numer)
  if( plon < rad_0deg ) plon = plon + rad_360deg

  if( which_is_western(plon,wlon) /= 2 )then
    plon = wlon
  elseif( which_is_western(plon,elon) /= 1 )then
    plon = elon
  endif
end subroutine calc_intersection_sphere_normal_parallel3
!===============================================================
! Calc. longit. of the intersection.
! 交差するかどうか事前に分かっていない場合に使う。
! 経度の範囲を指定する。
!===============================================================
subroutine calc_intersection_sphere_normal_parallel2(&
    wlon, elon, a, b, c, sgn, z, plon, stat)
  implicit none
  real(8), intent(in)  :: wlon, elon
  real(8), intent(in)  :: a, b, c
  integer, intent(in)  :: sgn
  real(8), intent(in)  :: z
  real(8), intent(out) :: plon
  integer, intent(out) :: stat

  real(8) :: cos_lat
  real(8) :: tan_wlat, tan_elat
  real(8) :: d
  real(8) :: x_numer, y_numer
  logical :: intersect
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  cos_lat = sqrt(1.d0 - z**2)  ! > 0

  tan_wlat = - (a*cos(wlon) + b*sin(wlon)) / c
  tan_elat = - (a*cos(elon) + b*sin(elon)) / c

  selectcase( sgn )
  case( 1 )
    ! tan_wlat < tan_lat .and. tan_lat < tan_elat
    intersect = tan_wlat*cos_lat < z .and. z < tan_elat*cos_lat
  case( -1 )
    ! tan_wlat > tan_lat .and. tan_lat > tan_elat
    intersect = tan_wlat*cos_lat > z .and. z > tan_elat*cos_lat
  case default
    call echo(CODE%BGN, 'calc_intersection_sphere_normal_parallel2')
    call eerr(str(msg_invalid_value())//&
            '\nsgn: '//str(sgn))
    call echo(CODE%RET)
  endselect

  if( intersect )then
    stat = STAT_INTERSECTION_YES

    d = sqrt(max(0.d0, (a**2+b**2) - (a**2+b**2+c**2) * z**2))

    x_numer = -a*c*z - sgn*b*d
    y_numer = -b*c*z + sgn*a*d

    plon = atan2(y_numer, x_numer)

    if( which_is_western(plon,wlon) /= 2 )then
      plon = wlon
    elseif( which_is_western(plon,elon) /= 1 )then
      plon = elon
    endif
  else
    stat = STAT_INTERSECTION_NO
    plon = 0.d0
  endif
end subroutine calc_intersection_sphere_normal_parallel2
!===============================================================
! Calc. longit. of the intersection.
! 交差するかどうか事前に分かっていない場合に使う。
! 経度の範囲を指定しない。
!===============================================================
subroutine calc_intersection_sphere_normal_parallel1(&
    ta, tb, tc, tsgn, sz, lon, stat)
  implicit none
  real(8), intent(in)  :: ta, tb, tc
  integer, intent(in)  :: tsgn
  real(8), intent(in)  :: sz
  real(8), intent(out) :: lon
  integer, intent(out) :: stat

  real(8) :: d2
  real(8) :: x_numer, y_numer
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  d2 = (ta**2 + tb**2) - (ta**2 + tb**2 + tc**2) * sz**2

  if( d2 > 0.d0 )then
    stat = STAT_INTERSECTION_YES

    x_numer = -ta*tc*sz - tsgn*tb*sqrt(d2)
    y_numer = -tb*tc*sz + tsgn*ta*sqrt(d2)

    lon = atan2(y_numer, x_numer)
  else
    stat = STAT_INTERSECTION_NO
    lon = 0.d0
  endif
  !-------------------------------------------------------------
end subroutine  calc_intersection_sphere_normal_parallel1
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
! Area of rectangular on the sphere
!===============================================================
function area_sphere_rect_lat0d(lat1, lat2) result(area)
  implicit none
  real(8), intent(in) :: lat1, lat2
  real(8)             :: area

  area = abs(sin(lat1) - sin(lat2))
end function area_sphere_rect_lat0d
!===============================================================
!
!===============================================================
function area_sphere_rect_lat1d(lat1, lat2) result(area)
  implicit none
  real(8), intent(in) :: lat1(:), lat2(:)
  real(8)             :: area(size(lat1))

  area(:) = abs(sin(lat1(:)) - sin(lat2(:)))
end function area_sphere_rect_lat1d
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
! Area of rectangular on the ellipsoid
!===============================================================
function area_ellips_rect_lat0d(lat1, lat2, e2) result(area)
  implicit none
  real(8), intent(in) :: lat1, lat2  ![rad]
  real(8), intent(in) :: e2   ! Square of the eccentricity
  real(8)             :: area
  real(8) :: e

  e = sqrt(e2)
  area = (1.d0-e2) * abs(s(lat1)-s(lat2))
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
function s(lat) result(res)
  implicit none
  real(8), intent(in) :: lat  ![rad]
  real(8)             :: res
  real(8) :: z

  z = sin(lat)
  res = 0.5d0 * z/(1.d0-e2*z*2) &
        + 0.25d0 / e * log(abs((1.d0+e*z)/(1.d0-e*z)))
end function
!---------------------------------------------------------------
end function area_ellips_rect_lat0d
!===============================================================
!
!===============================================================
function area_ellips_rect_lat1d(lat1, lat2, e2) result(area)
  implicit none
  real(8), intent(in) :: lat1(:), lat2(:)  ![rad]
  real(8), intent(in) :: e2   ! Square of the eccentricity
  real(8)             :: area(size(lat1))
  real(8) :: e

  e = sqrt(e2)
  area(:) = (1.d0-e2) * abs(s(lat1(:))-s(lat2(:)))
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
function s(lat) result(res)
  implicit none
  real(8), intent(in) :: lat(:)  ![rad]
  real(8)             :: res(size(lat))
  real(8) :: z(size(lat))

  z(:) = sin(lat(:))
  res(:) = 0.5d0 * z(:)/(1.d0-e2*z(:)**2) &
           + 0.25d0 / e * log(abs((1.d0+e*z(:))/(1.d0-e*z(:))))
end function
!---------------------------------------------------------------
end function area_ellips_rect_lat1d
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
real(8) function area_sphere_polarrect(lon, sin_lat, sgn_pole) result(area)
  implicit none
  real(8), intent(in) :: lon, sin_lat
  integer, intent(in) :: sgn_pole

  area = abs(sgn_pole - sin_lat) * lon
end function area_sphere_polarrect
!===============================================================
!
!===============================================================
!
!
!
!
!
!
!===============================================================
!
!===============================================================
real(8) function area_sphere_tri_south_bottom(lat1, lat2, lon) result(area)
  implicit none
  real(8), intent(in) :: lat1, lat2
  real(8), intent(in) :: lon

  real(8) :: south, north

  if( lat1 == lat2 )then
    area = 0.d0
  else
    south = min(lat1,lat2)
    north = max(lat1,lat2)
    area = ((cos(north) - cos(south)) / (south - north) - sin(south)) * lon
  endif
end function area_sphere_tri_south_bottom
!===============================================================
!
!===============================================================
real(8) function area_sphere_tri_north_bottom(lat1, lat2, lon) result(area)
  implicit none
  real(8), intent(in) :: lat1, lat2
  real(8), intent(in) :: lon

  real(8) :: south, north

  if( lat1 == lat2 )then
    area = 0.d0
  else
    south = min(lat1,lat2)
    north = max(lat1,lat2)
    area = ((cos(north) - cos(south)) / (north - south) + sin(north)) * lon
  endif
end function area_sphere_tri_north_bottom
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
real(8) function area_sphere_pl_tri_6_spherical(&
    lon, lat1, lat2, sgn_pole) result(area)
  implicit none
  real(8), intent(in) :: lon, lat1, lat2
  integer, intent(in) :: sgn_pole

  real(8) :: cos1, cos2, sin1, sin2
  real(8) :: cos_lon, sin_lon
  real(8) :: eta2
  real(8) :: chi

  cos1 = cos(lat1)
  cos2 = cos(lat2)
  sin1 = sin(lat1)
  sin2 = sin(lat2)
  cos_lon = cos(lon)
  sin_lon = sin(lon)

  if( lon == 0.d0 )then
    area = 0.d0
  else
!    eta2 = ((cos2*sin1+cos1*sin2)*sin(lon*0.5d0))**2 &
!           + ((cos2*sin1-cos1*sin2)*cos(lon*0.5d0))**2 &
!           + (cos1*cos2*sin_lon)**2
    eta2 = (sin(lat1+lat2)*sin(lon*0.5d0))**2 &
           + (sin(lat2-lat1)*cos(lon*0.5d0))**2 &
           + (cos1*cos2*sin_lon)**2
    if( eta2 == 0.d0 )then
      area = 0.d0
    else
      call edbg('(1) sin1*sin2         '//str(sin1*sin2,'es24.17'))
      call edbg('(2) cos1*cos2*cos_lon '//str(cos1*cos2*cos_lon,'es24.17'))
      call edbg('(3) 1-((1)+(2))       '//str(1.d0-(sin1*sin2+cos1*cos2*cos_lon),'es24.17'))
      call edbg('(4) 1-(1)-(2)         '//str(1.d0-sin1*sin2-cos1*cos2*cos_lon,'es24.17'))
      call edbg('1-(1)                 '//str(1.d0-sin1*sin2,'es24.17'))
      call edbg('1-(1)+(1)             '//str(1.d0-sin1*sin2+sin1*sin2,'es24.17'))
      call edbg('(1)+(2)               '//str(sin1*sin2+cos1*cos2*cos_lon,'es24.17'))
      call edbg('eta2                  '//str(eta2,'es24.17'))
      call edbg('sin_lon               '//str(sin_lon,'es24.17'))
      call edbg('(3)*sin_lon/eta2      '//str((1.d0-(sin1*sin2+cos1*cos2*cos_lon))*sin_lon/eta2,'es24.17'))
      call edbg('(4)*sin_lon/eta2      '//str((1.d0-sin1*sin2-cos1*cos2*cos_lon)*sin_lon/eta2,'es24.17'))
      call edbg('area north to pole    '//str(sgn_pole*(sgn_pole - max(sin1,sin2))*lon,'es24.17'))
      call edbg('area south to pole    '//str(sgn_pole*(sgn_pole - min(sin1,sin2))*lon,'es24.17'))
      call edbg('area triangle         '//str(lon*((cos2-cos1)/(lat1-lat2)-sin1),'es24.17'))
      chi = sgn_pole * (sin1+sin2) * (1.d0 - (sin1*sin2 + cos1*cos2*cos_lon)) * sin_lon / eta2
      chi = min(1.d0,max(-1.d0,chi))
      area = lon - asin(chi)
      call edbg('area = lon-asin(chi)  '//str(area,'es24.17'))
    endif
  endif
end function area_sphere_pl_tri_6_spherical
!===============================================================
!
!===============================================================
real(8) function area_sphere_pl_tri_5_spherical(&
    lon, lat1, lat2, sgn_pole) result(area)
  implicit none
  real(8), intent(in) :: lon, lat1, lat2
  integer, intent(in) :: sgn_pole

  real(8) :: cos1, cos2, sin1, sin2
  real(8) :: cos_lon, sin_lon
  real(8) :: eta2, eta_inv
  real(8) :: cosA1, cosA2

  cos1 = cos(lat1)
  cos2 = cos(lat2)
  sin1 = sin(lat1)
  sin2 = sin(lat2)
  cos_lon = cos(lon)
  sin_lon = sin(lon)

  eta2 = ((cos2*sin1+cos1*sin2)*sin(lon*0.5d0))**2 &
         + ((cos2*sin1-cos1*sin2)*cos(lon*0.5d0))**2 &
         + (cos1*cos2*sin_lon)**2

  if( eta2 == 0.d0 )then
    area = 0.d0
  else
    eta_inv = 1.d0 / sqrt(eta2)

    cosA1 = min(1.d0,max(-1.d0,sgn_pole * eta_inv * (cos1*sin2 - cos2*sin1*cos_lon)))
    cosA2 = min(1.d0,max(-1.d0,sgn_pole * eta_inv * (cos2*sin1 - cos1*sin2*cos_lon)))
    area = lon + acos(cosA1) + acos(cosA2) - pi
    call edbg('lon '//str(lon,'es20.13'))
    call edbg('eta2 '//str(eta2,'es20.13'))
    call edbg('cos1*sin2 - cos2*sin1*cos_lon '//str(cos1*sin2 - cos2*sin1*cos_lon,'es20.13'))
    call edbg('cos2*sin1 - cos1*sin2*cos_lon '//str(cos2*sin1 - cos1*sin2*cos_lon,'es20.13'))
    call edbg('cosA1 '//str(sgn_pole * eta_inv * (cos1*sin2 - cos2*sin1*cos_lon),'es20.13'))
    call edbg('cosA2 '//str(sgn_pole * eta_inv * (cos2*sin1 - cos1*sin2*cos_lon),'es20.13'))
    call edbg('A1 '//str(acos(cosA1),'es20.13')//' rad '//str(acos(cosA1)*r2d,'es20.13')//' deg')
    call edbg('A2 '//str(acos(cosA2),'es20.13')//' rad '//str(acos(cosA2)*r2d,'es20.13')//' deg')
    call edbg('A1 + A2 '//str(acos(cosA1)+acos(cosA2),'es20.13'))
  endif
end function area_sphere_pl_tri_5_spherical
!===============================================================
!
!===============================================================
real(8) function area_sphere_polartri4_spherical(&
    lon1, lon2, a, b, c, sgn_pole) result(area)
  implicit none
  real(8), intent(in) :: lon1, lon2
  real(8), intent(in) :: a, b, c
  integer, intent(in) :: sgn_pole

  real(8) :: eta
  real(8) :: cosA1_, cosA2_, cosA1, cosA2

  eta = sqrt(a**2 + b**2 + c**2)

  cosA1_ =  sgn_pole * (a*sin(lon1) - b*cos(lon1)) / eta
  cosA2_ = -sgn_pole * (a*sin(lon2) - b*cos(lon2)) / eta
  cosA1 = min(max(cosA1_,-1.d0),1.d0)
  cosA2 = min(max(cosA2_,-1.d0),1.d0)

  area = londiff_rad(lon1,lon2) + acos(cosA1) + acos(cosA2) - pi
end function area_sphere_polartri4_spherical
!===============================================================
!
!===============================================================
real(8) function area_sphere_polartri4_cartesian(&
    x1, y1, x2, y2, a, b, c, sgn_pole) result(area)
  implicit none
  real(8), intent(in) :: x1, y1, x2, y2
  real(8), intent(in) :: a, b, c
  integer, intent(in) :: sgn_pole

  real(8) :: lon1, lon2
  real(8) :: eta
  real(8) :: cosA1_, cosA2_, cosA1, cosA2

  lon1 = atan2(y1, x1)
  lon2 = atan2(y2, x2)
  eta = sqrt(a**2 + b**2 + c**2)

  cosA1_ =  sgn_pole * (a*sin(lon1) - b*cos(lon1)) / eta
  cosA2_ = -sgn_pole * (a*sin(lon2) - b*cos(lon2)) / eta
  cosA1 = min(max(cosA1_,-1.d0),1.d0)
  cosA2 = min(max(cosA2_,-1.d0),1.d0)

  area = londiff_rad(lon1,lon2) + acos(cosA1) + acos(cosA2) - pi
end function area_sphere_polartri4_cartesian
!===============================================================
! 多角形とラスターとの重複部分の面積の計算に用いる。
!===============================================================
real(8) function area_sphere_polartri3(&
    lon, lon1, lat1, lon2, lat2, &
    a, b, c, sgn_pole) result(area)
  implicit none
  real(8), intent(in) :: lon
  real(8), intent(in) :: lon1, lat1, lon2, lat2
  real(8), intent(in) :: a, b, c
  integer, intent(in) :: sgn_pole

  real(8) :: eta
  real(8) :: sinA1_, sinA2_, cosA1_, cosA2_
  real(8) :: sinA1, sinA2, cosA1, cosA2
  real(8) :: A1, A2
  real(8) :: x1, y1, z1, x2, y2, z2
  real(8) :: d1, d2
  real(8) :: e1, e2
  real(8) :: chi
  real(8) :: a_, b_, c_, eta_

  if( lon == 0.d0 )then
    area = 0.d0
    return
  endif

  eta = a**2 + b**2 + c**2

  call edbg('p1 ('//str((/lon1,lat1/)*r2d,'f20.15',', ')//')')
  call edbg('p2 ('//str((/lon2,lat2/)*r2d,'f20.15',', ')//')')

  x1 = cos(lat1) * cos(lon1)
  y1 = cos(lat1) * sin(lon1)
  z1 = sin(lat1)
  x2 = cos(lat2) * cos(lon2)
  y2 = cos(lat2) * sin(lon2)
  z2 = sin(lat2)
!  call edbg('a*x1 + b*y1 + c*z1 '//str(a*x1+b*y1+c*z1,'es20.13'))
!  call edbg('a*x2 + b*y2 + c*z2 '//str(a*x2+b*y2+c*z2,'es20.13'))

!if( .false. )then
  cosA1_ =  sgn_pole * (a*sin(lon1) - b*cos(lon1)) / sqrt(eta)
  cosA2_ = -sgn_pole * (a*sin(lon2) - b*cos(lon2)) / sqrt(eta)
  sinA1_ = abs(c) / (cos(lat1) * sqrt(eta))
  sinA2_ = abs(c) / (cos(lat2) * sqrt(eta))

  cosA1 = min(1.d0,max(-1.d0,cosA1_))
  cosA2 = min(1.d0,max(-1.d0,cosA2_))
  sinA1 = min(1.d0,max(-1.d0,sinA1_))
  sinA2 = min(1.d0,max(-1.d0,sinA2_))

!if( .false. )then
  call edbg('lon '//str(lon*r2d,'f20.15'))
  call edbg('sinA1 '//str(sinA1_)//' A1 '//str(asin(sinA1),'f20.15')//' rad '//str(asin(sinA1)*r2d,'f20.15')//' deg')
  call edbg('cosA1 '//str(cosA1_)//' A1 '//str(acos(cosA1),'f20.15')//' rad '//str(acos(cosA1)*r2d,'f20.15')//' deg')
  call edbg('sinA2 '//str(sinA2_)//' A2 '//str(asin(sinA2),'f20.15')//' rad '//str(asin(sinA2)*r2d,'f20.15')//' deg')
  call edbg('cosA2 '//str(cosA2_)//' A2 '//str(acos(cosA2),'f20.15')//' rad '//str(acos(cosA2)*r2d,'f20.15')//' deg')
  call edbg('sinA1**2+cosA1**2 '//str(sinA1_**2+cosA1_**2,'es20.13'))
  call edbg('sinA2**2+cosA2**2 '//str(sinA2_**2+cosA2_**2,'es20.13'))
!endif

!if( .false. )then
  call calc_cross_product(x1, y1, z1, x2, y2, z2, a_, b_, c_)
  eta_ = a_**2 + b_**2 + c_**2
  cosA1_ = sgn_pole * (cos(lat1)*sin(lat2) - cos(lat2)*sin(lat1)*cos(lon)) / sqrt(eta_)
  cosA2_ = sgn_pole * (cos(lat2)*sin(lat1) - cos(lat1)*sin(lat2)*cos(lon)) / sqrt(eta_)
  sinA1_ = cos(lat2)*sin(lon) / sqrt(eta_)
  sinA2_ = cos(lat1)*sin(lon) / sqrt(eta_)

  cosA1 = min(1.d0,max(-1.d0,cosA1_))
  cosA2 = min(1.d0,max(-1.d0,cosA2_))
  sinA1 = min(1.d0,max(-1.d0,sinA1_))
  sinA2 = min(1.d0,max(-1.d0,sinA2_))

!if( .false. )then
  call edbg('lon '//str(lon*r2d,'f20.15'))
  call edbg('sinA1 '//str(sinA1_)//' A1 '//str(asin(sinA1),'f20.15')//' rad '//str(asin(sinA1)*r2d,'f20.15')//' deg')
  call edbg('cosA1 '//str(cosA1_)//' A1 '//str(acos(cosA1),'f20.15')//' rad '//str(acos(cosA1)*r2d,'f20.15')//' deg')
  call edbg('sinA2 '//str(sinA2_)//' A2 '//str(asin(sinA2),'f20.15')//' rad '//str(asin(sinA2)*r2d,'f20.15')//' deg')
  call edbg('cosA2 '//str(cosA2_)//' A2 '//str(acos(cosA2),'f20.15')//' rad '//str(acos(cosA2)*r2d,'f20.15')//' deg')
  call edbg('sinA1**2+cosA1**2 '//str(sinA1_**2+cosA1_**2,'es20.13'))
  call edbg('sinA2**2+cosA2**2 '//str(sinA2_**2+cosA2_**2,'es20.13'))
  call edbg('sin(A1+A2) = sinA1cosA2 + cosA1sinA2 '//str(sinA1*cosA2 + cosA1*sinA2,'es20.13'))
  call edbg('pi-(A1+A2) '//str(asin(min(1.d0,max(-1.d0,sinA1*cosA2+cosA1*sinA2)))*r2d,'f20.15'))
!endif
  !-------------------------------------------------------------
  ! Directly calc. from cosA1 and cosA2
  !-------------------------------------------------------------
!if( .false. )then
  A1 = acos(cosA1)
  A2 = acos(cosA2)
  area = lon + A1 + A2 - pi

  call edbg('route 2')
  call edbg('  A1 = acos(cosA1) '//str(A1,'es20.13'))
  call edbg('  A2 = acos(cosA2) '//str(A2,'es20.13'))
  call edbg('  A1 + A2 '//str(A1+A2,'es20.13'))
  call edbg('  area = lon + A1 + A2 - pi '//str(lon+A1+A2-pi,'es20.13'))
!endif
  !-------------------------------------------------------------
  ! Calc. from d1, d2
  !-------------------------------------------------------------
if( .false. )then
  d1 = (a*sin(lon1) - b*cos(lon1)) / cos(lat2)
  d2 = (a*sin(lon2) - b*cos(lon2)) / cos(lat1)
  chi = abs((d1-d2) * c) / eta
  area = lon - asin(chi)

  call edbg('route 3')
  call edbg('  chi = |(d1-d2)*c| / eta '//str(chi,'es20.13'))
  call edbg('  asin(chi) '//str(asin(chi),'es20.13'))
  call edbg('  lon '//str(lon,'es20.13'))
  call edbg('  area = lon - asin(chi) '//str(lon-asin(chi),'es20.13'))
endif
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
if( .false. )then
  chi = abs(sin(lon) * (cos(lat1)*(a*sin(lon1)-b*cos(lon1)) - cos(lat2)*(a*sin(lon2)-b*cos(lon2))) / eta)
  !area = lon - asin(chi)

  call edbg('route 4')
  call edbg('  chi '//str(chi,'es20.13'))
  call edbg('  area = lon - asin(chi) '//str(lon-asin(chi),'es20.13'))
endif
  !-------------------------------------------------------------
  ! Calc. from e1, e2
  !-------------------------------------------------------------
if( .false. )then
  x1 = cos(lat1) * cos(lon1)
  y1 = cos(lat1) * sin(lon1)
  x2 = cos(lat2) * cos(lon2)
  y2 = cos(lat2) * sin(lon2)
  e1 = a*(y1-y2)
  e2 = b*(x1-x2)
  chi = abs((e1-e2) * c / (cos(lat1)*cos(lat2)*eta))

  call edbg('route 5')
  call edbg('  chi = |(e1-e2)*c / (cos(lat1)*cos(lat2)*eta)| '//str(chi,'es20.13'))
  call edbg('  asin(chi) '//str(asin(chi),'es20.13'))
  call edbg('  area = lon - asin(chi) '//str(lon-asin(chi),'es20.13'))
endif
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
end function area_sphere_polartri3
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
real(8) function area_sphere_polartri1_spherical( &
    lon, lat1, lat2, sgn_pole) result(area)
  implicit none
  real(8), intent(in) :: lon, lat1, lat2
  integer, intent(in) :: sgn_pole

  real(8) :: cos1, cos2, sin1, sin2
  real(8) :: cos_lon
  real(8) :: sin_lon
  real(8) :: sz2_A1xA2  ! |A1 x A2|^2
  real(8) :: sz_A1xA2  ! |A1 x A2|
  real(8) :: numer_cosA1, numer_cosA2
  real(8) :: A1, A2
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  cos1 = cos(lat1)
  cos2 = cos(lat2)
  sin1 = sin(lat1)
  sin2 = sin(lat2)
  cos_lon = cos(lon)
  sin_lon = sin(lon)

  sz2_A1xA2 = ((cos1*sin2 + cos2*sin1) * sin(lon*0.5d0))**2 &
              + ((cos1*sin2 - cos2*sin1) * cos(lon*0.5d0))**2 &
              + (cos1*cos2*sin_lon)**2

  sz_A1xA2 = sqrt(sz2_A1xA2)

  if( sz_A1xA2 == 0.d0 )then
    area = lon
  else
    numer_cosA1 = cos1*sin2 - cos2*sin1*cos_lon
    numer_cosA2 = cos2*sin1 - cos1*sin2*cos_lon

    A1 = acos(min(max(sgn_pole * numer_cosA1 / sz_A1xA2,-1.d0),1.d0))
    A2 = acos(min(max(sgn_pole * numer_cosA2 / sz_A1xA2,-1.d0),1.d0))

    area = A1 + A2 + lon - pi
  endif
  !-------------------------------------------------------------
end function area_sphere_polartri1_spherical
!===============================================================
!
!===============================================================
real(8) function area_sphere_polartri1_cartesian( &
    x1, y1, z1, x2, y2, z2, sgn_pole) result(area)
  implicit none
  real(8), intent(in) :: x1, y1, z1, x2, y2, z2
  integer, intent(in) :: sgn_pole

  real(8) :: xx, yy
  real(8) :: sz2_A1xP, &  ! |A1 x P|^2
             sz2_A2xP, &  ! |A2 x P|^2
             sz2_A1xA2    ! |A1 x A2|^2
  real(8) :: A1, A2, P
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  xx = y1*z2 - y2*z1
  yy = x1*z2 - x2*z1

  sz2_A1xP = x1*x1 + y1*y1
  sz2_A2xP = x2*x2 + y2*y2
  sz2_A1xA2 = (y1*z2 - z1*y2) * (y1*z2 - z1*y2) &
                + (z1*x2 - x1*z2) * (z1*x2 - x1*z2) &
                + (x1*y2 - y1*x2) * (x1*y2 - y1*x2)

  A1 = acos( sgn_pole * (y1*xx + x1*yy) / sqrt(sz2_A1xP * sz2_A1xA2))
  A2 = acos(-sgn_pole * (y2*xx + x2*yy) / sqrt(sz2_A2xP * sz2_A1xA2))
  P  = acos((y1*y2 + x1*x2) / sqrt(sz2_A1xP * sz2_A2xP))

  area = A1 + A2 + P - pi
  !-------------------------------------------------------------
end function area_sphere_polartri1_cartesian
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
! 2本の経線と1本の大円弧で囲まれた領域のうち小さい方の面積を求める。
! ・大円弧は単調
! ・大円弧は赤道を跨がない
! ・大円弧が北半球にあるとき、得られる面積は北半球側の領域
!===============================================================
real(8) function area_sphere_polartri2(&
    lon, gamma, eta, d1, d2, zeta1, zeta2) result(area)
  implicit none
  real(8), intent(in) :: lon
  real(8), intent(in) :: gamma, eta
  real(8), intent(in) :: d1, d2, zeta1, zeta2

  area = lon - asin(abs((d1 - d2) * gamma) * (zeta1 * zeta2 / eta))
  call edbg('pi-(A1+A2) '//str(asin(abs((d1-d2)*gamma)*(zeta1*zeta2/eta))*r2d,'es20.13')//' deg')
  call edbg('theta      '//str(lon*r2d,'es20.13')//' deg')
  call edbg('area       '//str(area,'es20.13'))
end function area_sphere_polartri2
!===============================================================
!
!===============================================================
subroutine calc_vals_for_polartri_from_longit( &
    alpha, beta, gamma, cos_lon, sin_lon, &
    tan_lat, zeta, d)
  implicit none
  real(8), intent(in) :: alpha, beta, gamma
  real(8), intent(in) :: cos_lon, sin_lon
  real(8), intent(out) :: tan_lat, d, zeta

  tan_lat = -(alpha*cos_lon + beta*sin_lon) / gamma
  zeta = sqrt(1.d0 + tan_lat*tan_lat)
  d = abs(beta*cos_lon - alpha*sin_lon) / zeta
end subroutine calc_vals_for_polartri_from_longit
!===============================================================
!
!===============================================================
subroutine calc_vals_for_polartri_from_latit( &
    alpha, beta, gamma, kappa, eta, sgn_lat, sin_lat, cos_lat, &
    d, lon, zeta)
  implicit none
  real(8), intent(in)  :: alpha, beta, gamma
  real(8), intent(in)  :: kappa, eta
  real(8), intent(in)  :: sgn_lat
  real(8), intent(in)  :: sin_lat, cos_lat
  real(8), intent(out) :: d, lon, zeta

  d = sqrt(max(kappa - eta * sin_lat*sin_lat, 0.d0))
  lon = atan2(-beta*gamma*sin_lat + sgn_lat*alpha*d, &
              -alpha*gamma*sin_lat - sgn_lat*beta*d)
  if( lon < rad_0deg ) lon = lon + rad_360deg
  zeta = 1.d0 / cos_lat
end subroutine calc_vals_for_polartri_from_latit
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
function area_sphere_tri(lonA, latA, lonB, latB, lonC, latC) result(area)
  implicit none
  real(8), intent(in)  :: lonA, latA, lonB, latB, lonC, latC  ![rad]
  real(8)              :: area
  real(8) :: OA(3), OB(3), OC(3)
  real(8) :: OAxOB(3), OBxOC(3), OCxOA(3)
  real(8) :: cosA_denom2, cosB_denom2, cosC_denom2
  real(8) :: cosA_numer, cosB_numer, cosC_numer
  real(8) :: cosA, cosB, cosC

  OA(:) = (/cos(latA)*cos(lonA), cos(latA)*sin(lonA), sin(latA)/)
  OB(:) = (/cos(latB)*cos(lonB), cos(latB)*sin(lonB), sin(latB)/)
  OC(:) = (/cos(latC)*cos(lonC), cos(latC)*sin(lonC), sin(latC)/)

  ! cos
  call calc_cross_product(OA, OB, OAxOB)
  call calc_cross_product(OB, OC, OBxOC)
  call calc_cross_product(OC, OA, OCxOA)

  cosA_denom2 = sum(OCxOA(:)**2) * sum(OAxOB(:)**2)
  cosB_denom2 = sum(OAxOB(:)**2) * sum(OBxOC(:)**2)
  cosC_denom2 = sum(OBxOC(:)**2) * sum(OCxOA(:)**2)
  if( cosA_denom2 == 0.d0 .or. cosB_denom2 == 0.d0 .or. cosC_denom2 == 0.d0 )then
    area = 0.d0
    return
  endif

  cosA_numer = sum(-OCxOA(:)*OAxOB(:))
  cosB_numer = sum(-OAxOB(:)*OBxOC(:))
  cosC_numer = sum(-OBxOC(:)*OCxOA(:))

  cosA = cosA_numer / sqrt(cosA_denom2)
  cosB = cosB_numer / sqrt(cosB_denom2)
  cosC = cosC_numer / sqrt(cosC_denom2)

  area = max(0.d0, acos(cosA) + acos(cosB) + acos(cosC) - pi)
end function area_sphere_tri
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
real(8) function area_sphere_polygon_spherical(typ, lon, lat, arctyp, a, b, c) result(area)
  implicit none
  integer(1), intent(in) :: typ
  real(8)   , intent(in) :: lon(:), lat(:)
  integer(1), intent(in) :: arctyp(:)
  real(8)   , intent(in) :: a(:), b(:), c(:)

  integer :: nmax, n1, n2
  integer :: sgn_pole

  call echo(CODE%BGN, 'area_sphere_polygon_spherical', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  sgn_pole = 1

  selectcase( typ )
  case( polygon_position_polar )
    if( lat(1) < rad_0deg )then
      sgn_pole = -1
    endif
  case( polygon_position_normal, &
        polygon_position_lon0 )
    if( any(lat(:) == -rad_90deg) )then
      sgn_pole = -1
    endif
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  typ: '//str(typ))
  endselect
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  nmax = size(lon)

  n1 = nmax
  n2 = 1

  area = 0.d0

  do while( n2 <= nmax )
    selectcase( arctyp(n1) )
    case( ARC_TYPE_NORMAL )
      area = area + area_sphere_polartri4(lon(n1), lon(n2), a(n1), b(n1), c(n1), sgn_pole)&
                      * sign_lon(lon(n1),lon(n2)) * sgn_pole
    case( ARC_TYPE_PARALLEL )
      area = area + area_sphere_polarrect(londiff_rad(lon(n1),lon(n2)), sin(lat(n1)), sgn_pole)&
                      * sign_lon(lon(n1),lon(n2)) * sgn_pole
    case( ARC_TYPE_MERIDIAN )
      continue
    case default
      call eerr(str(msg_invalid_value())//&
              '\n  arctyp('//str(n1)//'): '//str(arctyp(n1)))
    endselect

    n1 = n2
    n2 = n2 + 1
  enddo
  !-------------------------------------------------------------
  call echo(CODE%RET)
end function area_sphere_polygon_spherical
!===============================================================
!
!===============================================================
real(8) function area_sphere_polygon_cartesian(typ, x, y, z, arctyp, a, b, c) result(area)
  implicit none
  integer(1), intent(in) :: typ
  real(8)   , intent(in) :: x(:), y(:), z(:)
  integer(1), intent(in) :: arctyp(:)
  real(8)   , intent(in) :: a(:), b(:), c(:)

  integer :: nmax, n1, n2
  real(8) :: lon1, lon2
  integer :: sgn_pole

  call echo(CODE%BGN, 'area_sphere_polygon_cartesian', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  sgn_pole = 1

  selectcase( typ )
  case( polygon_position_polar )
    if( z(1) < 0.d0 )then
      sgn_pole = -1
    endif
  case( polygon_position_normal, &
        polygon_position_lon0 )
    if( any(z(:) == -1.d0) )then
      sgn_pole = -1
    endif
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  typ: '//str(typ))
  endselect
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  nmax = size(x)

  n1 = nmax
  n2 = 1

  lon1 = atan2(y(n1), x(n1))
  lon2 = atan2(y(n2), x(n2))

  area = 0.d0

  do while( n2 <= nmax )

    selectcase( arctyp(n1) )
    case( ARC_TYPE_NORMAL )
      area = area + area_sphere_polartri4(lon1, lon2, a(n1), b(n1), c(n1), sgn_pole)&
                      * sign_lon(lon1,lon2) * sgn_pole
    case( ARC_TYPE_PARALLEL )
      area = area + area_sphere_polarrect(londiff_rad(lon1,lon2), z(n1), sgn_pole)&
                      * sign_lon(lon1,lon2) * sgn_pole
    case( ARC_TYPE_MERIDIAN )
      continue
    case default
      call eerr(str(msg_invalid_value())//&
              '\n  arctyp('//str(n1)//'): '//str(arctyp(n1)))
    endselect

    n1 = n2
    n2 = n2 + 1

    lon1 = lon2
    lon2 = atan2(y(n2), x(n2))
  enddo
  !-------------------------------------------------------------
  call echo(CODE%RET)
end function area_sphere_polygon_cartesian
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
real(8) function area_sphere_parallel_to_parallel(&
    slon1, slon2, slat, tlon1, tlon2, tlat) result(area)
  implicit none
  real(8), intent(in) :: slon1, slon2, slat
  real(8), intent(in) :: tlon1, tlon2, tlat

  integer :: s_id_west, t_id_west
  integer :: id_west, id_east
  integer :: sgn
  integer :: s_dir_lon, t_dir_lon
  real(8) :: west, east
  real(8) :: lon

  call echo(CODE%BGN, 'area_sphere_parallel_to_parallel', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  area = 0.d0

  if( slat >= tlat )then
    call echo(CODE%RET)
    return
  endif

  call calc_range_longit_shared(&
         slon1, slon2, tlon1, tlon2, &
         s_id_west, t_id_west, id_west, id_east, west, east)

  if( id_west == 0 )then
    call echo(CODE%RET)
    return
  endif

  selectcase( s_id_west )
  case( 1 )
    s_dir_lon = 1
  case( 2 )
    s_dir_lon = -1
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  s_id_west: '//str(s_id_west))
  endselect

  selectcase( t_id_west )
  case( 1 )
    t_dir_lon = 1
  case( 2 )
    t_dir_lon = -1
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  t_id_west: '//str(t_id_west))
  endselect

  sgn = -1 * s_dir_lon * t_dir_lon
  lon = londiff_rad(west, east)
  area = area_sphere_rect(slat, tlat) * lon * sgn
  !-------------------------------------------------------------
  call echo(CODE%RET)
end function area_sphere_parallel_to_parallel
!===============================================================
!
!===============================================================
real(8) function area_sphere_normal_to_parallel(&
    slon1, slat1, slon2, slat2, &
    sa, sb, sc, sconvex, slontop, slattop, &
    tlon1, tlon2, tlat) result(area)
  implicit none
  real(8)   , intent(in) :: slon1, slat1, slon2, slat2
  real(8)   , intent(in) :: sa, sb, sc
  integer(1), intent(in) :: sconvex
  real(8)   , intent(in) :: slontop, slattop
  real(8)   , intent(in) :: tlon1, tlon2, tlat

  integer(1) :: sconvex_this
  integer    :: s_dir_lon, t_dir_lon
  integer    :: sgn
  integer    :: s_id_west, t_id_west, id_west, id_east
  real(8)    :: west, east
  real(8)    :: plon, plon1, plon2
  real(8)    :: lon
  real(8)    :: slat_west, slat_east
  real(8)    :: sarea, tarea
  integer, parameter :: sgn_pole = 1

  call echo(CODE%BGN, 'area_sphere_normal_to_parallel', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  area = 0.d0
  !-------------------------------------------------------------
  ! Return if s is above t.
  !-------------------------------------------------------------
  selectcase( sconvex )
  case( arc_convex_no, &
        arc_convex_northward )
    if( min(slat1,slat2) >= tlat )then
      call echo(CODE%RET)
      return
    endif
  case( arc_convex_southward )
    if( slattop >= tlat )then
      call echo(CODE%RET)
      return
    endif
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  sconvex: '//str(sconvex))
  endselect
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  call calc_range_longit_shared(&
         slon1, slon2, tlon1, tlon2, &
         s_id_west, t_id_west, id_west, id_east, west, east)

  if( id_west == 0 )then
    call echo(CODE%RET)
    return
  endif

  selectcase( id_west )
  case( 1 )
    slat_west = slat1
  case( 2 )
    slat_west = slat2
  case( 3, 4 )
    selectcase( s_id_west )
    case( 1 )
      call calc_intersection_sphere_normal_meridian(slon1, slat1, slon2, slat2, west, slat_west)
    case( 2 )
      call calc_intersection_sphere_normal_meridian(slon2, slat2, slon1, slat1, west, slat_west)
    case default
      call eerr(str(msg_invalid_value())//&
              '\n  s_id_west: '//str(s_id_west))
    endselect
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  id_west: '//str(id_west))
  endselect

  selectcase( id_east )
  case( 1 )
    slat_east = slat1
  case( 2 )
    slat_east = slat2
  case( 3, 4 )
    selectcase( s_id_west )
    case( 1 )
      call calc_intersection_sphere_normal_meridian(slon1, slat1, slon2, slat2, east, slat_east)
    case( 2 )
      call calc_intersection_sphere_normal_meridian(slon2, slat2, slon1, slat1, east, slat_east)
    case default
      call eerr(str(msg_invalid_value())//&
              '\n  s_id_west: '//str(s_id_west))
    endselect
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  id_east: '//str(id_east))
  endselect
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  selectcase( sconvex )
  case( arc_convex_no )
    sconvex_this = sconvex
  case( arc_convex_northward, &
        arc_convex_southward )
    if( which_is_western(west, slontop) == 1 .and. &
        which_is_western(east, slontop) == 2 )then
      sconvex_this = sconvex
    else
      sconvex_this = arc_convex_no
    endif
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  sconvex: '//str(sconvex))
  endselect

  s_dir_lon = dir_lon_from_id(s_id_west)
  t_dir_lon = dir_lon_from_id(t_id_west)
  sgn = -1 * s_dir_lon * t_dir_lon
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  selectcase( sconvex_this )
  !-------------------------------------------------------------
  ! Case: Not convex.
  case( arc_convex_no )
    !-----------------------------------------------------------
    ! Case: s is below t.
    if( max(slat_west, slat_east) <= tlat )then
      lon = londiff_rad(west, east)
      sarea = area_sphere_polartri1(lon, slat_west, slat_east, sgn_pole)
      tarea = area_sphere_polarrect(lon, sin(tlat), sgn_pole)
      area = (sarea - tarea) * sgn
    !-----------------------------------------------------------
    ! Case: s is below t at the western side of the intersection.
    !       s is upward at the intersection if it is west to east.
    elseif( slat_west < tlat .and. slat_east > tlat )then
      call calc_intersection_sphere_normal_parallel3(&
               west, east, sa, sb, sc, +s_dir_lon, sin(tlat), plon)
      lon = londiff_rad(west, plon)
      sarea = area_sphere_polartri1(lon, slat_west, tlat, sgn_pole)
      tarea = area_sphere_polarrect(lon, sin(tlat), sgn_pole)
      area = (sarea - tarea) * sgn
    !-----------------------------------------------------------
    ! Case: s is below t at the eastern side of the intersection.
    !       s is downward at the intersection if it is west to east.
    elseif( slat_west > tlat .and. slat_east < tlat )then
      call calc_intersection_sphere_normal_parallel3(&
               west, east, sa, sb, sc, -s_dir_lon, sin(tlat), plon)
      lon = londiff_rad(east, plon)
      sarea = area_sphere_polartri1(lon, slat_east, tlat, sgn_pole)
      tarea = area_sphere_polarrect(lon, sin(tlat), sgn_pole)
      area = (sarea - tarea) * sgn
    !-----------------------------------------------------------
    ! Case: s is above t..
    elseif( min(slat_west, slat_east) >= tlat )then
      continue
    !-----------------------------------------------------------
    ! Case: ERROR
    else
      call eerr(str(msg_unexpected_condition())//&
              '\n  Relation of s and t did not match any case.')
    endif
  !-------------------------------------------------------------
  ! Case: Convex northward.
  case( arc_convex_northward )
    !-----------------------------------------------------------
    ! Case: s is below t.
    if( slattop <= tlat )then
      lon = londiff_rad(west, east)
      sarea = area_sphere_polartri1(lon, slat_west, slat_east, sgn_pole)
      tarea = area_sphere_polarrect(lon, sin(tlat), sgn_pole)
      area = (sarea - tarea) * sgn
    !-----------------------------------------------------------
    ! Case: s is above t.
    elseif( slat_west >= tlat .and. slat_east >= tlat )then
      continue
    !-----------------------------------------------------------
    ! Case: s is below at the western side of the intersection.
    !       s is upward at the intersection if it is west to east.
    elseif( slat_west < tlat .and. slat_east >= tlat )then
      call calc_intersection_sphere_normal_parallel3(&
               west, east, sa, sb, sc, +s_dir_lon, sin(tlat), plon)
      lon = londiff_rad(plon, west)
      sarea = area_sphere_polartri1(lon, slat_west, tlat, sgn_pole)
      tarea = area_sphere_polarrect(lon, sin(tlat), sgn_pole)
      area = (sarea - tarea) * sgn
    !-----------------------------------------------------------
    ! Case: s is below at the eastern side of the intersection.
    !       s is downward at the intersection if it is west to east.
    elseif( slat_west >= tlat .and. slat_east < tlat )then
      call calc_intersection_sphere_normal_parallel3(&
               west, east, sa, sb, sc, -s_dir_lon, sin(tlat), plon)
      lon = londiff_rad(plon, east)
      sarea = area_sphere_polartri1(lon, slat_east, tlat, sgn_pole)
      tarea = area_sphere_polarrect(lon, sin(tlat), sgn_pole)
      area = (sarea - tarea) * sgn
    !-----------------------------------------------------------
    ! Case: s and t have two intersections.
    elseif( slat_west > tlat .and. slat_east > tlat )then
      call calc_intersection_sphere_normal_parallel3(&
               west, east, sa, sb, sc, +s_dir_lon, sin(tlat), plon)
      lon = londiff_rad(west, plon)
      sarea = area_sphere_polartri1(lon, slat_west, tlat, sgn_pole)
      tarea = area_sphere_polarrect(lon, sin(tlat), sgn_pole)
      area = area + (sarea - tarea)

      call calc_intersection_sphere_normal_parallel3(&
               west, east, sa, sb, sc, -s_dir_lon, sin(tlat), plon)
      lon = londiff_rad(east, plon)
      sarea = area_sphere_polartri1(lon, slat_east, tlat, sgn_pole)
      tarea = area_sphere_polarrect(lon, sin(tlat), sgn_pole)
      area = area + (sarea - tarea)

      area = area * sgn
    !-----------------------------------------------------------
    ! Case: ERROR
    else
      call eerr(str(msg_unexpected_condition())//&
              '\n  Relation of s and t did not match any case.')
    endif
  !-------------------------------------------------------------
  ! Case: Convex southward.
  case( arc_convex_southward )
    !-----------------------------------------------------------
    ! Case: s is above t.
    if( slattop >= tlat )then
      continue
    !-----------------------------------------------------------
    ! Case: s is below t.
    elseif( max(slat_west, slat_east) <= tlat )then
      lon = londiff_rad(west, east)
      sarea = area_sphere_polartri1(lon, slat_west, slat_east, sgn_pole)
      tarea = area_sphere_polarrect(lon, sin(tlat), sgn_pole)
      area = (sarea - tarea) * sgn
    !-----------------------------------------------------------
    ! Case: s is below t at the western side of the intersection.
    !       s is upward at the intersection if it is west to east.
    elseif( slat_west <= tlat .and. slat_east > tlat )then
      call calc_intersection_sphere_normal_parallel3(&
               west, east, sa, sb, sc, +s_dir_lon, sin(tlat), plon)
      lon = londiff_rad(plon, west)
      sarea = area_sphere_polartri1(lon, slat_west, tlat, sgn_pole)
      tarea = area_sphere_polarrect(lon, sin(tlat), sgn_pole)
      area = (sarea - tarea) * sgn
    !-----------------------------------------------------------
    ! Case: s is above t at the eastern side of the intersection.
    !       s is downward at the intersection if it is west to east.
    elseif( slat_west > tlat .and. slat_east <= tlat )then
      call calc_intersection_sphere_normal_parallel3(&
               west, east, sa, sb, sc, -s_dir_lon, sin(tlat), plon)
      lon = londiff_rad(plon, east)
      sarea = area_sphere_polartri1(lon, slat_east, tlat, sgn_pole)
      tarea = area_sphere_polarrect(lon, sin(tlat), sgn_pole)
      area = (sarea - tarea) * sgn
    !-----------------------------------------------------------
    ! Case: s and t have two intersections.
    !       s is downward at the western intersection if it is west to east.
    elseif( slat_west > tlat .and. slat_east > tlat )then
      call calc_intersection_sphere_normal_parallel3(&
               west, east, sa, sb, sc, -s_dir_lon, sin(tlat), plon1)
      call calc_intersection_sphere_normal_parallel3(&
               west, east, sa, sb, sc, +s_dir_lon, sin(tlat), plon2)
      lon = londiff_rad(plon1, plon2)
      sarea = area_sphere_polartri1(lon, tlat, tlat, sgn_pole)
      tarea = area_sphere_polarrect(lon, sin(tlat), sgn_pole)
      area = (sarea - tarea) * sgn
    !-----------------------------------------------------------
    ! Case: ERROR
    else
      call eerr(str(msg_unexpected_condition())//&
              '\n  Relation of s and t did not match any case.')
    endif
  !-------------------------------------------------------------
  ! Case: ERROR
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  sconvex_this: '//str(sconvex_this))
  endselect
  !-------------------------------------------------------------
  call echo(CODE%RET)
end function area_sphere_normal_to_parallel
!===============================================================
!
!===============================================================
real(8) function area_sphere_parallel_to_normal(&
    slon1, slon2, slat, &
    tlon1, tlat1, tlon2, tlat2, &
    ta, tb, tc, tconvex, tlontop, tlattop) result(area)
  implicit none
  real(8)   , intent(in) :: slon1, slon2, slat
  real(8)   , intent(in) :: tlon1, tlat1, tlon2, tlat2
  real(8)   , intent(in) :: ta, tb, tc
  integer(1), intent(in) :: tconvex
  real(8)   , intent(in) :: tlontop, tlattop

  integer(1) :: tconvex_this
  integer    :: s_dir_lon, t_dir_lon
  integer    :: sgn
  integer    :: s_id_west, t_id_west, id_west, id_east
  real(8)    :: west, east
  real(8)    :: plon, plon1, plon2
  real(8)    :: lon
  real(8)    :: tlat_west, tlat_east
  real(8)    :: sarea, tarea
  integer, parameter :: sgn_pole = 1.d0

  call echo(CODE%BGN, 'area_sphere_parallel_to_normal', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  area = 0.d0
  !-------------------------------------------------------------
  ! Return if t is below s.
  !-------------------------------------------------------------
  selectcase( tconvex )
  case( arc_convex_no, &
        arc_convex_southward )
    if( max(tlat1,tlat2) <= slat )then
      call echo(CODE%RET)
      return
    endif
  case( arc_convex_northward )
    if( tlattop <= slat )then
      call echo(CODE%RET)
      return
    endif
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  tconvex: '//str(tconvex))
  endselect
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  call calc_range_longit_shared(&
         slon1, slon2, tlon1, tlon2, &
         s_id_west, t_id_west, id_west, id_east, west, east)

  if( id_west == 0 )then
    call echo(CODE%RET)
    return
  endif

  selectcase( id_west )
  case( 1, 2 )
    selectcase( t_id_west )
    case( 1 )
      call calc_intersection_sphere_normal_meridian(tlon1, tlat1, tlon2, tlat2, west, tlat_west)
    case( 2 )
      call calc_intersection_sphere_normal_meridian(tlon2, tlat2, tlon1, tlat1, west, tlat_west)
    case default
      call eerr(str(msg_invalid_value())//&
              '\n  t_id_west: '//str(t_id_west))
    endselect
  case( 3 )
    tlat_west = tlat1
  case( 4 )
    tlat_west = tlat2
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  id_west: '//str(id_west))
  endselect

  selectcase( id_east )
  case( 1, 2 )
    selectcase( t_id_west )
    case( 1 )
      call calc_intersection_sphere_normal_meridian(tlon1, tlat1, tlon2, tlat2, east, tlat_east)
    case( 2 )
      call calc_intersection_sphere_normal_meridian(tlon2, tlat2, tlon1, tlat1, east, tlat_east)
    case default
      call eerr(str(msg_invalid_value())//&
              '\n  t_id_west: '//str(t_id_west))
    endselect
  case( 3 )
    tlat_east = tlat1
  case( 4 )
    tlat_east = tlat2
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  id_east: '//str(id_east))
  endselect

  !call edbg('  west: ('//str((/west,tlat_west/)*r2d,'f12.7',', ')//')')
  !call edbg('  east: ('//str((/east,tlat_east/)*r2d,'f12.7',', ')//')')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  selectcase( tconvex )
  case( arc_convex_no )
    tconvex_this = tconvex
  case( arc_convex_northward, &
        arc_convex_southward )
    if( which_is_western(west, tlontop) == 1 .and. &
        which_is_western(east, tlontop) == 2 )then
      tconvex_this = tconvex
    else
      tconvex_this = arc_convex_no
    endif
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  tconvex: '//str(tconvex))
  endselect

  s_dir_lon = dir_lon_from_id(s_id_west)
  t_dir_lon = dir_lon_from_id(t_id_west)
  sgn = -1 * s_dir_lon * t_dir_lon
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  selectcase( tconvex_this )
  !-------------------------------------------------------------
  ! Case: Not convex.
  case( arc_convex_no )
    !-----------------------------------------------------------
    ! Case: t is above s.
    if( min(tlat_west, tlat_east) >= slat )then
      lon = londiff_rad(west, east)
      sarea = area_sphere_polarrect(lon, sin(slat), sgn_pole)
      tarea = area_sphere_polartri1(lon, tlat_west, tlat_east, sgn_pole)
      area = (sarea - tarea) * sgn
    !-----------------------------------------------------------
    ! Case: t is below s.
    elseif( max(tlat_west, tlat_east) <= slat )then
      continue
    !-----------------------------------------------------------
    ! Case: t is above s at the western side of the intersection.
    !       t is downward at the intersection if it is west to east.
    elseif( tlat_west > slat .and. tlat_east < slat )then
      call calc_intersection_sphere_normal_parallel3(&
               west, east, ta, tb, tc, -t_dir_lon, sin(slat), plon)
      lon = londiff_rad(west, plon)
      sarea = area_sphere_polarrect(lon, sin(slat), sgn_pole)
      tarea = area_sphere_polartri1(lon, tlat_west, slat, sgn_pole)
      area = (sarea - tarea) * sgn

      !call edbg('t is above s at the western side of the intersection.')
      !call edbg('direction at the intersection: '//str(t_dir_lon*-1))
      !call edbg('plon: '//str(plon*r2d,'f12.7'))
      !call edbg('sarea: '//str(sarea))
      !call edbg('tarea: '//str(tarea))
    !-----------------------------------------------------------
    ! Case: t is above s at the eastern side of the intersection.
    !       t is upward at the intersection if it is west to east.
    elseif( tlat_west < slat .and. tlat_east > slat )then
      call calc_intersection_sphere_normal_parallel3(&
               west, east, ta, tb, tc, +t_dir_lon, sin(slat), plon)
      lon = londiff_rad(east, plon)
      sarea = area_sphere_polarrect(lon, sin(slat), sgn_pole)
      tarea = area_sphere_polartri1(lon, tlat_east, slat, sgn_pole)
      area = (sarea - tarea) * sgn
    !-----------------------------------------------------------
    ! Case: ERROR
    else
      call eerr(str(msg_unexpected_condition())//&
              '\n  Relation of s and t did not match any case.')
    endif
  !-------------------------------------------------------------
  ! Case: Convex northward.
  case( arc_convex_northward )
    !-----------------------------------------------------------
    ! Case: t is above s.
    if( tlat_west >= slat .and. tlat_east >= slat )then
      lon = londiff_rad(west, east)
      sarea = area_sphere_polarrect(lon, sin(slat), sgn_pole)
      tarea = area_sphere_polartri1(lon, tlat_west, tlat_east, sgn_pole)
      area = (sarea - tarea) * sgn
    !-----------------------------------------------------------
    ! Case: t is above at the western side of the intersection.
    !       t is downward at the intersection if it is west to east.
    elseif( tlat_west >= slat .and. tlat_east < slat )then
      call calc_intersection_sphere_normal_parallel3(&
               west, east, ta, tb, tc, -t_dir_lon, sin(slat), plon)
      lon = londiff_rad(plon, west)
      sarea = area_sphere_polarrect(lon, sin(slat), sgn_pole)
      tarea = area_sphere_polartri1(lon, tlat_west, slat, sgn_pole)
      area = (sarea - tarea) * sgn
    !-----------------------------------------------------------
    ! Case: t is above at the eastern side of the intersection.
    !       t is upward at the intersection if it is west to east.
    elseif( tlat_west < slat .and. tlat_east >= slat )then
      call calc_intersection_sphere_normal_parallel3(&
               west, east, ta, tb, tc, +t_dir_lon, sin(slat), plon)
      lon = londiff_rad(plon, east)
      sarea = area_sphere_polarrect(lon, sin(slat), sgn_pole)
      tarea = area_sphere_polartri1(lon, tlat_east, slat, sgn_pole)
      area = (sarea - tarea) * sgn
    !-----------------------------------------------------------
    ! Case: t and s have two intersections.
    elseif( tlat_west < slat .and. tlat_east < slat )then
      call calc_intersection_sphere_normal_parallel3(&
               west, east, ta, tb, tc, +1, sin(slat), plon1)
      call calc_intersection_sphere_normal_parallel3(&
               west, east, ta, tb, tc, -1, sin(slat), plon2)
      lon = londiff_rad(plon1, plon2)
      sarea = area_sphere_polarrect(lon, sin(slat), sgn_pole)
      tarea = area_sphere_polartri1(lon, slat, slat, sgn_pole)
      area = (sarea - tarea) * sgn
    !-----------------------------------------------------------
    ! Case: ERROR
    else
      call eerr(str(msg_unexpected_condition())//&
              '\n  Relation of s and t did not match any case.')
    endif
  !-------------------------------------------------------------
  ! Case: Convex southward.
  case( arc_convex_southward )
    !-----------------------------------------------------------
    ! Case: t is above s.
    if( tlattop >= slat )then
      lon = londiff_rad(west, east)
      sarea = area_sphere_polarrect(lon, sin(slat), sgn_pole)
      tarea = area_sphere_polartri1(lon, tlat_west, tlat_east, sgn_pole)
      area = (sarea - tarea) * sgn
    !-----------------------------------------------------------
    ! Case: t is below s.
    elseif( tlat_west <= slat .and. tlat_east <= slat )then
      continue
    !-----------------------------------------------------------
    ! Case: t is above s at the western side of the intersection.
    !       t is downward at the intersection if it is west to east.
    elseif( tlat_west > slat .and. tlat_east <= slat )then
      call calc_intersection_sphere_normal_parallel3(&
               west, east, ta, tb, tc, -t_dir_lon, sin(slat), plon)
      lon = londiff_rad(plon, west)
      sarea = area_sphere_polarrect(lon, sin(slat), sgn_pole)
      tarea = area_sphere_polartri1(lon, tlat_west, slat, sgn_pole)
      area = (sarea - tarea) * sgn
    !-----------------------------------------------------------
    ! Case: t is above s at the eastern side of the intersection.
    !       t is upward at the intersection if it is west to east.
    elseif( tlat_west <= slat .and. tlat_east > slat )then
      call calc_intersection_sphere_normal_parallel3(&
               west, east, ta, tb, tc, +t_dir_lon, sin(slat), plon)
      lon = londiff_rad(plon, east)
      sarea = area_sphere_polarrect(lon, sin(slat), sgn_pole)
      tarea = area_sphere_polartri1(lon, tlat_east, slat, sgn_pole)
      area = (sarea - tarea) * sgn
    !-----------------------------------------------------------
    ! Case: t and s have two intersections.
    !       t is downward at the western intersection if it is west to east.
    elseif( tlat_west > slat .and. tlat_east > slat )then
      call calc_intersection_sphere_normal_parallel3(&
               west, east, ta, tb, tc, -t_dir_lon, sin(slat), plon)
      lon = londiff_rad(west, plon)
      sarea = area_sphere_polarrect(lon, sin(slat), sgn_pole)
      tarea = area_sphere_polartri1(lon, tlat_west, slat, sgn_pole)
      area = area + (sarea - tarea)

      call calc_intersection_sphere_normal_parallel3(&
               west, east, ta, tb, tc, +t_dir_lon, sin(slat), plon)
      lon = londiff_rad(east, plon)
      sarea = area_sphere_polarrect(lon, sin(slat), sgn_pole)
      tarea = area_sphere_polartri1(lon, tlat_east, slat, sgn_pole)
      area = area + (sarea - tarea)

      area = area * sgn
    !-----------------------------------------------------------
    ! Case: ERROR
    else
      call eerr(str(msg_unexpected_condition())//&
              '\n  Relation of s and t did not match any case.')
    endif
  !-------------------------------------------------------------
  ! Case: ERROR
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  tconvex_this: '//str(tconvex_this))
  endselect
  !-------------------------------------------------------------
  call echo(CODE%RET)
end function area_sphere_parallel_to_normal
!===============================================================
! Calc. area above s and below t.
!===============================================================
real(8) function area_sphere_normal_to_normal(&
    sx1, sy1, sz1, sx2, sy2, sz2, &
    slon1, slat1, slon2, slat2, &
    sa, sb, sc, &
    tx1, ty1, tz1, tx2, ty2, tz2, &
    tlon1, tlat1, tlon2, tlat2, &
    ta, tb, tc, &
    tconvex, tlattop) result(area)
  implicit none
  real(8)   , intent(in) :: sx1, sy1, sz1, sx2, sy2, sz2
  real(8)   , intent(in) :: tx1, ty1, tz1, tx2, ty2, tz2
  real(8)   , intent(in) :: slon1, slat1, slon2, slat2
  real(8)   , intent(in) :: tlon1, tlat1, tlon2, tlat2
  real(8)   , intent(in) :: sa, sb, sc
  real(8)   , intent(in) :: ta, tb, tc
  integer(1), intent(in) :: tconvex
  real(8)   , intent(in) :: tlattop

  real(8) :: west, east
  real(8) :: lon
  integer :: s_id_west, t_id_west
  integer :: id_west, id_east
  real(8) :: slat_west, slat_east
  real(8) :: tlat_west, tlat_east
  integer :: s_dir_lon, t_dir_lon
  integer :: sgn
  real(8) :: plon, plat
  real(8) :: sarea, tarea
  integer, parameter :: sgn_pole = 1.d0

  call echo(CODE%BGN, 'area_sphere_normal_to_normal', '-a')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  area = 0.d0
  !-------------------------------------------------------------
  ! Return if t is below s.
  !-------------------------------------------------------------
  selectcase( tconvex )
  case( arc_convex_no )
    if( max(tlat1,tlat2) <= min(slat1,slat2) )then
      call echo(CODE%RET)
      return
    endif
  case( arc_convex_northward )
    if( tlattop <= min(slat1,slat2) )then
      call echo(CODE%RET)
      return
    endif
  case( arc_convex_southward )
    continue
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  tconvex: '//str(tconvex))
  endselect
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  call calc_range_longit_shared(&
        slon1, slon2, tlon1, tlon2, &
        s_id_west, t_id_west, id_west, id_east, west, east)

!  call edbg('s ('//str((/slon1,slat1/)*r2d,'f12.7',', ')//&
!          ') - ('//str((/slon2,slat2/)*r2d,'f12.7',', ')//')')
!  call edbg('t ('//str((/tlon1,tlat1/)*r2d,'f12.7',', ')//&
!          ') - ('//str((/tlon2,tlat2/)*r2d,'f12.7',', ')//')')
!  call edbg('id_west '//str(id_west)//' id_east '//str(id_east)//&
!            ' s_id_west '//str(s_id_west)//' t_id_west '//str(t_id_west))

  if( id_west == 0 )then
    call echo(CODE%RET)
    return
  endif

  selectcase( id_west )
  case( 1, 2 )
    if( id_west == 1 )then
      slat_west = slat1
    else
      slat_west = slat2
    endif

    selectcase( t_id_west )
    case( 1 )  ! tlon1 < tlon2
      call calc_intersection_sphere_normal_meridian(tlon1, tlat1, tlon2, tlat2, west, tlat_west)
    case( 2 )  ! tlon2 < tlon1
      call calc_intersection_sphere_normal_meridian(tlon2, tlat2, tlon1, tlat1, west, tlat_west)
    case default
      call eerr(str(msg_invalid_value())//&
              '\n  t_id_west: '//str(t_id_west))
    endselect
  case( 3, 4 )
    if( id_west == 3 )then
      tlat_west = tlat1
    else
      tlat_west = tlat2
    endif

    selectcase( s_id_west )
    case( 1 )  ! slon1 < slon2
      call calc_intersection_sphere_normal_meridian(slon1, slat1, slon2, slat2, west, slat_west)
    case( 2 )  ! slon2 < slon1
      call calc_intersection_sphere_normal_meridian(slon2, slat2, slon1, slat1, west, slat_west)
    case default
      call eerr(str(msg_invalid_value())//&
              '\n  s_id_west: '//str(s_id_west))
    endselect
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  id_west: '//str(id_west))
  endselect

  selectcase( id_east )
  case( 1, 2 )
    if( id_east == 1 )then
      slat_east = slat1
    else
      slat_east = slat2
    endif

    selectcase( t_id_west )
    case( 1 )  ! tlon1 < tlon2
      call calc_intersection_sphere_normal_meridian(tlon1, tlat1, tlon2, tlat2, east, tlat_east)
    case( 2 )  ! tlon2 < tlon1
      call calc_intersection_sphere_normal_meridian(tlon2, tlat2, tlon1, tlat1, east, tlat_east)
    case default
      call eerr(str(msg_invalid_value())//&
              '\n  t_id_west: '//str(t_id_west))
    endselect
  case( 3, 4 )
    if( id_east == 3 )then
      tlat_east = tlat1
    else
      tlat_east = tlat2
    endif

    selectcase( s_id_west )
    case( 1 )  ! slon1 < slon2
      call calc_intersection_sphere_normal_meridian(slon1, slat1, slon2, slat2, east, slat_east)
    case( 2 )  ! slon2 < slon1
      call calc_intersection_sphere_normal_meridian(slon2, slat2, slon1, slat1, east, slat_east)
    case default
      call eerr(str(msg_invalid_value())//&
              '\n  s_id_west: '//str(s_id_west))
    endselect
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  id_east: '//str(id_east))
  endselect

  s_dir_lon = dir_lon_from_id(s_id_west)
  t_dir_lon = dir_lon_from_id(t_id_west)
  sgn = -1 * s_dir_lon * t_dir_lon

  call edbg('west: '//str(west*r2d,'f12.7')//' east: '//str(east*r2d,'f12.7'))
  call edbg('slat: '//str((/slat_west,slat_east/)*r2d,'f12.7',', '))
  call edbg('tlat: '//str((/tlat_west,tlat_east/)*r2d,'f12.7',', '))
  call edbg('dir_lon s: '//str(s_dir_lon)//' t: '//str(t_dir_lon))
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  ! Case: t is above s.
  if( tlat_west >= slat_west .and. tlat_east >= slat_east )then
    lon = londiff_rad(west, east)
    sarea = area_sphere_polartri1(lon, slat_west, slat_east, sgn_pole)
    tarea = area_sphere_polartri1(lon, tlat_west, tlat_east, sgn_pole)
    area = (sarea - tarea) * sgn

    call edbg('Case: t is above s'//&
            '\n  lon  '//str(lon*r2d,'f20.15')//&
            '\n  sarea '//str(sarea,'es20.13')//&
            '\n  tarea '//str(tarea,'es20.13')//&
            '\n  area  '//str((sarea-tarea)*sgn,'es20.13'))

    sarea = area_sphere_polarrect(lon, sin(slat_west), sgn_pole)
!    call edbg('  sarea_rect '//str(sarea,'es20.13'))
  !-------------------------------------------------------------
  ! Case: t is above s at the western side of the intersection.
  elseif( tlat_west > slat_west .and. tlat_east <= slat_east )then
    call calc_intersection_sphere_normal_normal(&
           sx1, sy1, sz1, sx2, sy2, sz2, sa, sb, sc, &
           tx1, ty1, tz1, tx2, ty2, tz2, ta, tb, tc, &
           west, east, plon, plat)
    lon = londiff_rad(west, plon)
    sarea = area_sphere_polartri1(lon, slat_west, plat, sgn_pole)
    tarea = area_sphere_polartri1(lon, tlat_west, plat, sgn_pole)
    area = (sarea - tarea) * sgn

!    call edbg('Case: t is above s at the western side of the intersection'//&
!            '\n  plon '//str(plon*r2d,'f20.15')//&
!            '\n  lon  '//str(lon*r2d,'f20.15')//&
!            '\n  sarea '//str(sarea,'es20.13')//&
!            '\n  tarea '//str(tarea,'es20.13')//&
!            '\n  area  '//str((sarea-tarea)*sgn,'es20.13'))
  !-------------------------------------------------------------
  ! Case: t is above s at the eastern side of the intersection.
  elseif( tlat_west <= slat_west .and. tlat_east > slat_east )then
    call calc_intersection_sphere_normal_normal(&
           sx1, sy1, sz1, sx2, sy2, sz2, sa, sb, sc, &
           tx1, ty1, tz1, tx2, ty2, tz2, ta, tb, tc, &
           west, east, plon, plat)
    lon = londiff_rad(east, plon)
    sarea = area_sphere_polartri1(lon, slat_east, plat, sgn_pole)
    tarea = area_sphere_polartri1(lon, tlat_east, plat, sgn_pole)
    area = (sarea - tarea) * sgn

!    call edbg('Case: t is above s at the eastern side of the intersection'//&
!            '\n  plon '//str(plon*r2d,'f20.15')//&
!            '\n  lon  '//str(lon*r2d,'f20.15')//&
!            '\n  sarea '//str(sarea,'es20.13')//&
!            '\n  tarea '//str(tarea,'es20.13')//&
!            '\n  area  '//str((sarea-tarea)*sgn,'es20.13'))
  !-------------------------------------------------------------
  ! Case: t is below s.
  elseif( tlat_west <= slat_west .and. tlat_east <= slat_east )then
    continue

!    call edbg('Case: t is below s')
  !-------------------------------------------------------------
  ! Case: ERROR
  else
    call eerr(str(msg_unexpected_condition()))
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end function area_sphere_normal_to_normal
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
real(8) function area_sphere_intersection_polygon_polygon(&
    styp, sx, sy, sz, slon, slat, sarctyp, sa, sb, sc, &
    sconvex, slontop, slattop, &
    ttyp, tx, ty, tz, tlon, tlat, tarctyp, ta, tb, tc, &
    tconvex, tlontop, tlattop) result(area)
  implicit none
  integer(1), intent(in) :: styp
  real(8)   , intent(in) :: sx(:), sy(:), sz(:)
  real(8)   , intent(in) :: slon(:), slat(:)
  integer(1), intent(in) :: sarctyp(:)
  real(8)   , intent(in) :: sa(:), sb(:), sc(:)
  integer(1), intent(in) :: sconvex(:)
  real(8)   , intent(in) :: slontop(:), slattop(:)
  integer(1), intent(in) :: ttyp
  real(8)   , intent(in) :: tx(:), ty(:), tz(:)
  real(8)   , intent(in) :: tlon(:), tlat(:)
  integer(1), intent(in) :: tarctyp(:)
  real(8)   , intent(in) :: ta(:), tb(:), tc(:)
  integer(1), intent(in) :: tconvex(:)
  real(8)   , intent(in) :: tlontop(:), tlattop(:)

  integer :: snmax, sn, sn_
  integer :: tnmax, tn, tn_
  integer :: sgn_pole
  real(8) :: area_add

  call echo(CODE%BGN, 'area_sphere_intersection_polygon_polygon', '-a')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  area = 0.d0

  tnmax = size(tlon)
  snmax = size(slon)

  tn_ = tnmax
  tn  = 1
  do while( tn <= tnmax )

    call edbg('t('//str(tn_)//') type '//str(tarctyp(tn_)))
    call edbg('  ('//str((/tlon(tn_),tlat(tn_)/)*r2d,'f12.7',', ')//&
            ') - ('//str((/tlon(tn) ,tlat(tn) /)*r2d,'f12.7',', ')//')')

    selectcase( tarctyp(tn_) )
    !-----------------------------------------------------------
    ! Case: t is normal.
    case( ARC_TYPE_NORMAL )
      sn_ = snmax
      sn  = 1
      do while( sn <= snmax )

        call edbg('s('//str(sn_)//') type '//str(sarctyp(sn_)))
        call edbg('  ('//str((/slon(sn_),slat(sn_)/)*r2d,'f12.7',', ')//&
                ') - ('//str((/slon(sn) ,slat(sn) /)*r2d,'f12.7',', ')//')')

        selectcase( sarctyp(sn_) )
        !-------------------------------------------------------
        ! Case: s is normal.
        case( ARC_TYPE_NORMAL )
          area_add = area_sphere_normal_to_normal(&
                       sx(sn_), sy(sn_), sz(sn_), sx(sn), sy(sn), sz(sn), &
                       slon(sn_), slat(sn_), slon(sn), slat(sn), &
                       sa(sn_), sb(sn_), sc(sn_), &
                       tx(tn_), ty(tn_), tz(tn_), tx(tn), ty(tn), tz(tn), &
                       tlon(tn_), tlat(tn_), tlon(tn), tlat(tn), &
                       ta(tn_), tb(tn_), tc(tn_), &
                       tconvex(tn_), tlattop(tn_))
          !area_add = area_sphere_normal_to_normal(&
          !             (/sx(sn_),sy(sn_),sz(sn_)/), (/sx(sn),sy(sn),sz(sn)/), &
          !             slon(sn_), slat(sn_), slon(sn), slat(sn), &
          !             (/tx(tn_),ty(tn_),tz(tn_)/), (/tx(tn),ty(tn),tz(tn)/), &
          !             tlon(tn_), tlat(tn_), tlon(tn), tlat(tn), &
          !             tconvex(tn_), tlattop(tn_))
        !-------------------------------------------------------
        ! Case: s is parallel.
        case( ARC_TYPE_PARALLEL )
          area_add = area_sphere_parallel_to_normal(&
                       slon(sn_), slon(sn_), slat(sn_), &
                       tlon(tn_), tlat(tn_), tlon(tn), tlat(tn), &
                       ta(tn_), tb(tn_), tc(tn_), &
                       tconvex(tn_), tlontop(tn_), tlattop(tn_))
        !-------------------------------------------------------
        ! Case: s is meridian.
        case( ARC_TYPE_MERIDIAN )
          area_add = 0.d0
        !-------------------------------------------------------
        ! Case: ERROR
        case default
          call eerr(str(msg_invalid_value())//&
                  '\n  sarctyp(sn_): '//str(sarctyp(sn_)))
        endselect

!call edbg('area: '//str(area)//' + '//str(area_add))
        area = area + area_add

        sn_ = sn
        sn  = sn + 1
      enddo  ! sn/
    !-----------------------------------------------------------
    ! Case: t is parallel.
    case( ARC_TYPE_PARALLEL )
      sn_ = snmax
      sn  = 1
      do while( sn <= snmax )
        selectcase( sarctyp(sn_) )
        !-------------------------------------------------------
        ! Case: s is normal.
        case( ARC_TYPE_NORMAL )
          area_add = area_sphere_normal_to_parallel(&
                       slon(sn_), slat(sn_), slon(sn), slat(sn), &
                       sa(sn_), sb(sn_), sc(sn_), &
                       sconvex(sn_), slontop(sn_), slattop(sn_), &
                       tlon(tn_), tlon(tn), tlat(tn_))
        !-------------------------------------------------------
        ! Case: s is parallel.
        case( ARC_TYPE_PARALLEL )
          area_add = area_sphere_parallel_to_parallel(&
                       slon(sn_), slon(sn), slat(sn_), &
                       tlon(tn_), tlon(tn), tlat(tn_))
        !-------------------------------------------------------
        ! Case: s is meridian.
        case( ARC_TYPE_MERIDIAN )
          area = 0.d0
        !-------------------------------------------------------
        ! Case: ERROR
        case default
          call eerr(str(msg_invalid_value())//&
                  '\n  sarctyp(sn_): '//str(sarctyp(sn_)))
        endselect

        area = area + area_add

        sn_ = sn
        sn  = sn + 1
      enddo  ! sn/
    !-----------------------------------------------------------
    ! Case: t is meridian.
    case( ARC_TYPE_MERIDIAN )
      continue
    !-----------------------------------------------------------
    ! Case: ERROR
    case default
      call eerr(str(msg_invalid_value())//&
              '\n  tarctyp(tn_): '//str(tarctyp(tn_)))
    endselect

    tn_ = tn
    tn  = tn + 1
  enddo  ! tn/
  !-------------------------------------------------------------
  ! For case that s includes south pole
  !-------------------------------------------------------------
  if( styp == polygon_position_polar .and. slat(1) < rad_0deg )then
    call edbg('s includes south pole.')

    sgn_pole = -1

    tn_ = tnmax
    tn  = 1
    do while( tn <= tnmax )
      selectcase( tarctyp(tn_) )
      !---------------------------------------------------------
      ! Case: t is normal.
      case( ARC_TYPE_NORMAL )
        area_add = area_sphere_polartri4(&
                       tlon(tn_), tlon(tn), &
                       ta(tn_), tb(tn_), tc(tn_), sgn_pole)

        if( which_is_western(tlon(tn_),tlon(tn)) == 1 )then
          area_add = -area_add
        endif
      !---------------------------------------------------------
      ! Case: t is parallel.
      case( ARC_TYPE_PARALLEL )
        area_add = area_sphere_polarrect(&
                     londiff_rad(tlon(tn_), tlon(tn)), sin(tlat(tn_)), sgn_pole)

        if( which_is_western(tlon(tn_),tlon(tn)) == 1 )then
          area_add = -area_add
        endif
      !---------------------------------------------------------
      ! Case: t is meridian.
      case( ARC_TYPE_MERIDIAN )
        area_add = 0.d0
      !---------------------------------------------------------
      ! Case: ERROR
      case default
        call eerr(str(msg_invalid_value())//&
                '\n  tarctyp(tn_): '//str(tarctyp(tn_)))
      endselect

      call edbg('t('//str(tn_)//') type '//str(tarctyp(tn_)))
      call edbg('  ('//str((/tlon(tn_),tlat(tn_)/)*r2d,'f12.7',', ')//&
              ') - ('//str((/tlon(tn) ,tlat(tn) /)*r2d,'f12.7',', ')//')')
      call edbg('area '//str(area_add,'es20.13'))

      area = area + area_add

      tn_ = tn
      tn  = tn + 1
    enddo  ! tn/
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end function area_sphere_intersection_polygon_polygon
!===============================================================
!
!===============================================================
real(8) function area_sphere_intersection_lattice_polygon(&
    swest, seast, ssouth, snorth, &
    ttyp, tlon, tlat, tarctyp, ta, tb, tc, &
    tconvex, tlontop, tlattop) result(area)
  implicit none
  real(8)   , intent(in) :: swest, seast, ssouth, snorth
  integer(1), intent(in) :: ttyp
  real(8)   , intent(in) :: tlon(:), tlat(:)  ! (tnmax)
  integer(1), intent(in) :: tarctyp(:)
  real(8)   , intent(in) :: ta(:), tb(:), tc(:)
  integer(1), intent(in) :: tconvex(:)
  real(8)   , intent(in) :: tlontop(:), tlattop(:)

  integer :: tnmax, tn, tn_
  real(8) :: area_south, area_north
  logical :: calc_for_north

  call echo(CODE%BGN, 'area_sphere_intersection_lattice_polygon', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  tnmax = size(tlon)

  area = 0.d0

  tn_ = tnmax
  tn  = 1
  do while( tn <= tnmax )
    selectcase( tarctyp(tn_) )
    !-----------------------------------------------------------
    ! Case: Normal
    case( ARC_TYPE_NORMAL )
      !call edbg('tn: '//str(tn)//&
      !          ' ('//str((/tlon(tn_),tlat(tn_)/)*r2d,'f12.7',', ')//')'//&
      !        ' - ('//str((/tlon(tn ),tlat(tn )/)*r2d,'f12.7',', ')//')')
      !call edbg('  convex: '//str(tconvex(tn_))//&
      !          ' ('//str((/tlontop(tn_),tlattop(tn_)/)*r2d,'f12.7')//')')

      area_south = area_sphere_parallel_to_normal(&
        swest, seast, ssouth, &
        tlon(tn_), tlat(tn_), tlon(tn), tlat(tn), &
        ta(tn_), tb(tn_), tc(tn_), &
        tconvex(tn_), tlontop(tn_), tlattop(tn_))

      area_north = area_sphere_parallel_to_normal(&
        seast, swest, snorth, &
        tlon(tn_), tlat(tn_), tlon(tn), tlat(tn), &
        ta(tn_), tb(tn_), tc(tn_), &
        tconvex(tn_), tlontop(tn_), tlattop(tn_))

      !call edbg('  area south: '//str(area_south))
      !call edbg('       north: '//str(area_north))
      !call edbg('       add  : '//str(area_south+area_north))

      area = area + area_south + area_north
    !-----------------------------------------------------------
    ! Case: Parallel
    case( ARC_TYPE_PARALLEL )
      area_south = area_sphere_parallel_to_parallel(&
                     swest, seast, ssouth, &
                     tlon(tn_), tlon(tn), tlat(tn_))
      area_north = area_sphere_parallel_to_parallel(&
                     seast, swest, snorth, &
                     tlon(tn_), tlon(tn), tlat(tn_))

      !call edbg('tn: '//str(tn)//&
      !          ' ('//str((/tlon(tn_),tlat(tn_)/)*r2d,'f12.7',', ')//')'//&
      !        ' - ('//str((/tlon(tn ),tlat(tn )/)*r2d,'f12.7',', ')//')')
      !call edbg('s lon '//str((/seast,swest/)*r2d,'f12.7',' - '))
      !call edbg('  south '//str(ssouth*r2d,'f12.7'))
      !call edbg('  north '//str(snorth*r2d,'f12.7'))
      !call edbg('area south '//str(area_south))
      !call edbg('     north '//str(area_north))
      !call edbg('     add '//str(area_south+area_north))

      area = area + area_south + area_north
    !-----------------------------------------------------------
    ! Case: Meridian
    case( ARC_TYPE_MERIDIAN )
      continue
    !-----------------------------------------------------------
    ! Case: ERROR
    case default
      call eerr(str(msg_invalid_value())//&
              '\n  tarctyp(tn_): '//str(tarctyp(tn_))//&
              '\n  tn_: '//str(tn_))
    endselect

    tn_ = tn
    tn  = tn + 1
  enddo  ! tn/
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  calc_for_north = .false.
  if( ttyp == polygon_position_polar .and. tlat(1) > 0.d0 )then
    !call edbg('t includes north pole.')
    calc_for_north = .true.
  elseif( any(tlat(:) == rad_90deg) )then
    !call edbg('t has pole on the edge.')
    calc_for_north = .true.
  endif

  if( calc_for_north )then
    tn_ = tnmax
    tn  = 1
    do while( tn <= tnmax )
      !call edbg('tn: '//str(tn)//&
      !          ' ('//str((/tlon(tn ),rad_90deg/)*r2d,'f12.7',', ')//')'//&
      !        ' - ('//str((/tlon(tn_),rad_90deg/)*r2d,'f12.7',', ')//')')

      if( tarctyp(tn_) /= arc_type_meridian )then
        area_south = area_sphere_parallel_to_parallel(&
          swest, seast, ssouth, &
          tlon(tn), tlon(tn_), rad_90deg)

        area_north = area_sphere_parallel_to_parallel(&
          seast, swest, snorth, &
          tlon(tn), tlon(tn_), rad_90deg)

        !call edbg('  area south: '//str(area_south))
        !call edbg('       north: '//str(area_north))
        !call edbg('       add  : '//str(area_south+area_north))

        area = area + area_south + area_north
      endif

      tn_ = tn
      tn  = tn + 1
    enddo
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end function area_sphere_intersection_lattice_polygon
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine calc_range_longit_shared(&
    slon1, slon2, tlon1, tlon2, &
    s_id_west, t_id_west, id_west, id_east, west, east)
  real(8), intent(in)  :: slon1, slon2
  real(8), intent(in)  :: tlon1, tlon2
  integer, intent(out) :: s_id_west, t_id_west
  integer, intent(out) :: id_west, id_east
  real(8), intent(out) :: west, east

  integer :: stwest_id_west, steast_id_west
  integer :: seast_twest_id_west, teast_swest_id_west
  real(8) :: swest, seast, twest, teast

  call echo(CODE%BGN, 'calc_range_longit_shared', '-a')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  id_west = 0
  id_east = 0
  west = 0.d0
  east = 0.d0
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  s_id_west = which_is_western(slon1, slon2)
  t_id_west = which_is_western(tlon1, tlon2)

  !call edbg('s_id_west: '//str(s_id_west))
  !call edbg('t_id_west: '//str(t_id_west))

  selectcase( s_id_west )
  case( 1 )
    swest = slon1
    seast = slon2
  case( 2 )
    swest = slon2
    seast = slon1
  case default
    call eerr(str(msg_invalid_value())//&
           '\n  s_id_west: '//str(s_id_west))
  endselect

  selectcase( t_id_west )
  case( 1 )
    twest = tlon1
    teast = tlon2
  case( 2 )
    twest = tlon2
    teast = tlon1
  case default
    call eerr(str(msg_invalid_value())//&
           '\n  t_id_west: '//str(t_id_west))
  endselect

  !call edbg('s west '//str(swest*r2d,'f12.7')//' east '//str(seast*r2d,'f12.7'))
  !call edbg('t west '//str(twest*r2d,'f12.7')//' east '//str(teast*r2d,'f12.7'))
  !-------------------------------------------------------------
  ! Return if seast <= twest.
  !-------------------------------------------------------------
  seast_twest_id_west = which_is_western(seast, twest)

  selectcase( seast_twest_id_west )
  case( 0, 1 )
    call echo(CODE%RET)
    return
  case( 2 )
    continue
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  seast_twest_id_west: '//str(seast_twest_id_west))
  endselect
  !-------------------------------------------------------------
  ! Return if teast <= swest.
  !-------------------------------------------------------------
  teast_swest_id_west = which_is_western(teast, swest)

  selectcase( teast_swest_id_west )
  case( 0, 1 )
    call echo(CODE%RET)
    return
  case( 2 )
    continue
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  teast_swest_id_west: '//str(teast_swest_id_west))
  endselect
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  stwest_id_west = which_is_western(swest, twest)
  steast_id_west = which_is_western(seast, teast)

  selectcase( stwest_id_west )
  !-------------------------------------------------------------
  ! Case: swest <= twest
  case( 0, 1 )
    west = twest

    selectcase( t_id_west )
    case( 1 )
      id_west = 3  ! west == twest == tlon1
    case( 2 )
      id_west = 4  ! west == twest == tlon2
    case default
      call eerr(str(msg_invalid_value())//&
             '\n  t_id_west: '//str(t_id_west))
    endselect
  !-------------------------------------------------------------
  ! Case: twest < swest
  case( 2 )
    west = swest

    selectcase( s_id_west )
    case( 1 )
      id_west = 1  ! west == swest == slon1
    case( 2 )
      id_west = 2  ! west == swest == slon2
    case default
      call eerr(str(msg_invalid_value())//&
             '\n  s_id_west: '//str(s_id_west))
    endselect
  !-------------------------------------------------------------
  ! Case: ERROR
  case default
    call eerr(str(msg_invalid_value())//&
           '\n  stwest_id_west: '//str(stwest_id_west))
  endselect

  selectcase( steast_id_west )
  !-------------------------------------------------------------
  ! Case: seast <= teast
  case( 0, 1 )
    east = seast

    selectcase( s_id_west )
    case( 1 )
      id_east = 2  ! east == seast == slon2
    case( 2 )
      id_east = 1  ! east == seast == slon1
    case default
      call eerr(str(msg_invalid_value())//&
             '\n  s_id_west: '//str(s_id_west))
    endselect
  case( 2 )
    east = teast

    selectcase( t_id_west )
    case( 1 )
      id_east = 4  ! east == teast == tlon2
    case( 2 )
      id_east = 3  ! east == teast == tlon1
    case default
      call eerr(str(msg_invalid_value())//&
             '\n  t_id_west: '//str(t_id_west))
    endselect
  case default
    call eerr(str(msg_invalid_value())//&
           '\n  steast_id_west: '//str(steast_id_west))
  endselect
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine calc_range_longit_shared
!===============================================================
!
!===============================================================
integer function dir_lon_from_id(id_west) result(res)
  implicit none
  integer, intent(in) :: id_west
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  call echo(CODE%BGN, 'dir_lon_from_id', '-p')
  !-------------------------------------------------------------
  selectcase( id_west )
  !-------------------------------------------------------------
  ! Case: lon1 < lon2
  case( 1 )
    res = 1
  !-------------------------------------------------------------
  ! Case: lon1 > lon2
  case( 2 )
    res = -1
  !-------------------------------------------------------------
  ! Case: ERROR (unexpected condition)
  case( 0 )
    call eerr(str(msg_unexpected_condition())//&
            '\n  id_west: '//str(id_west))
  !-------------------------------------------------------------
  ! Case: ERROR (invalid value)
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  id_west: '//str(id_west))
  endselect
  !-------------------------------------------------------------
  call echo(CODE%RET)
end function dir_lon_from_id
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
real(8) function dist_sphere_0d(lon1, lat1, lon2, lat2) result(dist)
  implicit none
  real(8), intent(in) :: lon1, lat1, lon2, lat2  ![rad]

  dist = acos(min(1.d0, max(-1.d0, &
              sin(lat1)*sin(lat2) + cos(lat1)*cos(lat2)*cos(londiff_rad(lon1, lon2))))) * 2.d0
end function dist_sphere_0d
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine calc_inv_matrix_3x3(a, b, coef)
  implicit none
  real(8), intent(in)  :: a(3,3)
  real(8), intent(out) :: b(3,3)
  real(8), intent(out) :: coef

  coef =  a(1,1)*a(2,2)*a(3,3) + a(2,1)*a(3,2)*a(1,3) + a(3,1)*a(1,2)*a(2,3) &
          - (a(3,1)*a(2,2)*a(1,3) + a(2,1)*a(1,2)*a(3,3) + a(1,1)*a(3,2)*a(2,3))
  b(1,1) =   a(2,2)*a(3,3) - a(3,2)*a(2,3)
  b(1,2) = - a(1,2)*a(3,3) + a(3,2)*a(1,3)
  b(1,3) =   a(1,2)*a(2,3) - a(2,2)*a(1,3)
  b(2,1) = - a(2,1)*a(3,3) + a(3,1)*a(2,3)
  b(2,2) =   a(1,1)*a(3,3) - a(3,1)*a(1,3)
  b(2,3) = - a(1,1)*a(2,3) + a(2,1)*a(1,3)
  b(3,1) =   a(2,1)*a(3,2) - a(3,1)*a(2,2)
  b(3,2) = - a(1,1)*a(3,2) + a(3,1)*a(1,2)
  b(3,3) =   a(1,1)*a(2,2) - a(2,1)*a(1,2)
end subroutine calc_inv_matrix_3x3
!===============================================================
!
!===============================================================
subroutine calc_cross_product_array(a, b, c)
  implicit none
  real(8), intent(in)  :: a(:), b(:)
  real(8), intent(out) :: c(:)

  c(1) = a(2)*b(3) - a(3)*b(2)
  c(2) = a(3)*b(1) - a(1)*b(3)
  c(3) = a(1)*b(2) - a(2)*b(1)
end subroutine calc_cross_product_array
!===============================================================
!
!===============================================================
subroutine calc_cross_product_components(a1, a2, a3, b1, b2, b3, c1, c2, c3)
  implicit none
  real(8), intent(in)  :: a1, a2, a3, b1, b2, b3
  real(8), intent(out) :: c1, c2, c3

  c1 = a2*b3 - a3*b2
  c2 = a3*b1 - a1*b3
  c3 = a1*b2 - a2*b1
end subroutine calc_cross_product_components
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine swap_int4(a, b)
  implicit none
  integer(4), intent(inout) :: a, b
  integer(4) :: c

  c = a
  a = b
  b = c
end subroutine swap_int4
!===============================================================
!
!===============================================================
end module lib_math
